# Rex - a Smash Interpreter/Development Program
# Version 1.3 -- 7/4/05
#
# Copyright 2004-2005 Andrew Walters
#
#
#   This program is free software; you may redistribute it or modify
#   it under the terms of version 2 of the GNU General Public License
#   as published by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program (license.txt); if not, the full text can
#   be found near the end of this file.
#

""" Version history/changes:
	1.0 (12/02/04) Initial (limited) release for testing.
	1.1 (12/03/04)
		-- Fixed the incorrect behavior of h: lines in "game.txt".
		-- Allowed blank entries to be posted to Hall of Fame fields.
		-- Allowed customized strings in Hall of Fame fields to be blank.
		-- Fixed Hall of Fame sorting so that blank entries are always
			sorted last unless explicitly defined to be sorted earlier
			by using a blank string in a custom-defined field.
		-- Made the area/action text-hiding feature persist from turn
			to turn, until the user switches it to a different setting.
		-- Fixed the buggy behavior related to whitespace-only lines in Smash code.
		-- Added proper exception handling when examining items, if no
			(e)objectactions.sma is located.
		-- Improved handling of errors encountered in trying to initialize
			a Smash game, with no game previously in progress.
		-- Added a new command to output as HTML the Game Text page, plus
			the inventory table, in a single file, roughly as it would
			appear on _Adventure Games Live_.
	1.2 (12/17/04) Intended for first public release.
		-- Changed the .ico/.gif setting lines, as well as loading of the
			config.agc file, to look in Rex's directory instead of the
			current working directory
		-- Added the option of narrating area description text as well as
			action description text in the narration feature.
		-- Added the option of logging all parsed lines (for debugging Smash code)
		-- Implemented the config.agc configuration file at the level of
			Smash game directories as well as keeping the main ("global")
			config.agc for use upon startup and as a template for
			creating new configuration files in game directories.
		-- Added support for new Smash operators added since the last version:
			A:, p:, and P:
		-- Fixed a few minor bugs with GUI operation, unary operator parsing,
			and illegal setting of state variables from repeatpre/post files.
		-- Added support for new Smash use of "g" commands in "repeatpre.sma"
		-- Updated the Rex documentation for the changes since version 1.0.
	1.3 (7/04/05)
		-- Fixed a bug in which area text was narrated (if that option was enabled)
			when a loss or parse error state was encountered.
		-- Fixed a bug with posting "win" data with default Hall of Fame columns.
		-- Fixed a bug in which {}, {'}, and the like were not evaluated within
			[] for embedding images in description text.
		-- Fixed a bug introduced in 1.2 with editing of state variables.
		-- Added a new menu command for narrating the current state's area
			description text without making another move.
		-- Added a bit of red text showing the total inventory size, displayed in
			the inventory panel
		-- Allowed "-" as a valid state variable tag character for parsing whole files
			only, as global variable tags were already treated
		-- Fixed a bug with false flagging of invalid tag characters for regular/enumerated items and
			location jumps with { } constructions that can occur in parsing whole files
		-- Fixed a bug with logging changes to rewinding/fast-forwarding all timers
		-- Added the capability of reading in files of any/mixed newlines style
		-- Allowed invalid characters to be overlooked and only flagged as "possible" errors when
			they occur in parsing whole files, as they may result from { } constructions with
			contents only initialized properly during normal play
		
	"""

""" Possible improvements to the overall program structure, for developers:
	1. Having buttons for inventory use appear in the inventory panel,
		now that that window doesn't have to refresh on every move.
	2. Implementing backup states "last_state", "last2_state", etc.
		differently to speed up program operation for games with rather
		large state structures: instead of having an entirely new
		state for each move, just have a list/stack/other structure
		of strings representing those changes (need the "before"
		and "after" values, e.g. "v money [1, 2]" to signify that
		global variable "money" changed from 1 to 2.  Then when you back
		up, these strings could be used to "undo" changes to the
		game state.  Problem, though, is if a move involves loading
		a state, resetting the game, or setting a new game.  Perhaps
		a hybrid model would work.  This could be more trouble than it's
		worth, though.
	"""


import sys, os, time, string, array, re, copy
from wxPython.wx import *
import wx.html
import wxPython.lib.wxpTag
from wxPython.lib.dialogs import wxScrolledMessageDialog


#-------------------INTERPRETER CLASSES/FUNCTIONS----------------------

class Game_State:
	""" Stores everything that needs to be written into a save file. """

	def __init__(self):
		self.directory = ''
		
		self.wl_state = 0
		""" wl_state:
			0 = normal
			1 = after losing move
			2 = after winning (and game ending) move
			3 = after winning move from which player can continue
			-1 = an error (options are not displayed) """
		
		self.moves = 0
		self.location = ''
		self.act_text = ''
		self.area_text = ''
		self.gvars = {}
		self.svars = {}
		self.const_options = []
		self.move_options = []
		self.act_options = []
		self.visible_opt_count = 0
		self.enum_inv = {}
		self.reg_inv = {}
		self.up_timers = {}
		self.down_timers = {}
		self.act_image = ''
		self.area_image = ''
		self.status_lines = {}
		self.status_list = []
		self.random_seed = 1788195708
		
	def load(self, other):
		""" Copies the Game State "other" into "self".  Note the
			deepcopy method for copying structures. """

		self.directory = other.directory
		self.wl_state = other.wl_state
		self.moves = other.moves
		self.location = other.location
		self.act_text = other.act_text
		self.area_text = other.area_text
		self.gvars = copy.deepcopy(other.gvars)
		self.svars = copy.deepcopy(other.svars)
		self.const_options = copy.deepcopy(other.const_options)
		self.move_options = copy.deepcopy(other.move_options)
		self.act_options = copy.deepcopy(other.act_options)
		self.visible_opt_count = other.visible_opt_count
		self.enum_inv = copy.deepcopy(other.enum_inv)
		self.reg_inv = copy.deepcopy(other.reg_inv)
		self.up_timers = copy.deepcopy(other.up_timers)
		self.down_timers = copy.deepcopy(other.down_timers)
		self.act_image = other.act_image
		self.area_image = other.area_image
		self.status_lines = copy.deepcopy(other.status_lines)
		self.status_list = copy.deepcopy(other.status_list)
		self.random_seed = other.random_seed

class Option:
	""" Each Game_State instance has three lists of these Options
		corresponding to the three Smash option types. """

	def __init__(self, type, loc, pos, tag, opt_text):
		self.type = type
		self.text = opt_text
		self.location = loc
		self.position = pos
		self.tag = tag
		self.visible = 1	# only zero for hidden constant options
		
class ParsingException(Exception):
	pass


#-------------- GLOBAL FUNCTIONS --------------

def insort(a, x, low=0, high=None):
	""" Inserts item x in list a and keeps it sorted
		(assuming a is sorted to begin with).

		If x is already in a, inserts it to the
		right of the rightmost x.  Optional args
		low (default 0) and high (default len(a))
		bound the slice of a to be searched. """

	if high is None:
		high = len(a)
	while (low < high):
		mid = (low+high)/2
		if (x < a[mid]):
			high = mid
		else:
			low = mid+1
	a.insert(low, x)


def CountWords(line):
	""" Counts the words in the string passed in.  This function is
		applied to text description lines one at a time.  Because
		these lines are not appended before being passed in, { }
		spanning lines aren't processed first.

		This function attempts to properly exclude "words" which
		are really in HTML tags, such as the "p" in "<p>" or the
		"table" in "</table>".  It doesn't work right if there
		are spaces or multiple words in such tags.

		Some possible improvements to make:
			1. Only feeding in text strings to this function
				after they've been properly fed through
				eval_string and eval_string_imagetags.
				In this case, text lines would have to
				be appended in blocks to avoid problems
				with { } etc. spanning multiple lines
			2. Simplifying the algorithm so that it doesn't
				use the confusing "inword" variable. """

	global word_pieces
	global wordcount_dictionary
	
	word_count = 0
	inword = 0
	cur_word = []
	whole_word = ''
	line += ' '
	for index in range(0, len(line)):
		char = line[index]
		if ((char == '<') and (inword == 0)):
			inword = 3
		elif ((char == '/') and (inword == 3)):
			pass
		elif ((char == '>') and (inword > 2)):
			inword = 0
		elif (word_pieces.has_key(char)):
			if (inword < 3):
				if (inword == 0):
					inword = 1
				if (word_pieces[char] == 2):
					inword = 2
				cur_word.append(char)
			else:
				inword = 4
		elif (inword > 0):
			if (inword == 2):
				whole_word = ''.join(cur_word).lower().strip("'")
				word_number = wordcount_dictionary.get(whole_word, 0)
				wordcount_dictionary[whole_word] = word_number + 1
				word_count += 1
			if (inword < 4):
				inword = 0
				cur_word = []
				if (char == '<'):
					inword = 3
	return (word_count)

def UpdateHoFData():

	global game_title, credits, difficulty, synopsis_short, synopsis_long, hof_fields

	credit_lines = []
	difficulty_temp = -1
	synopsis_short_lines = []
	synopsis_long_lines = []
	hof_temp = []
	hof_empty_line = 0

	gtfile = open(os.path.join(cur_state.directory, 'game.txt'), 'rU')
	game_title_temp = gtfile.readline().strip()
	if (game_title_temp == ''):
		WriteText('*** WARNING ***\nThe game has not been given a title in "game.txt" -- data not imported.')
		raise ParsingException, "No title"
	while 1:
		gtline = gtfile.readline().strip()
		if (gtline == ''):
			test = gtfile.read(1)		# test whether this blank is at the end of the file
			if (test == ''):
				break
			else:
				gtfile.seek(-1,1)
				continue
		start = gtline[0:2]
		rest = gtline[2:].strip()
		if (start == "c:"):
			credit_lines.append(rest)
		elif (start == "d:"):
			try:
				difficulty_temp = int(rest)
				if (difficulty_temp < 0):
					raise ParsingException, gtline
			except ValueError:
				raise ParsingException, gtline
		elif (start == "s:"):
			synopsis_short_lines.append(rest)
			synopsis_long_lines.append(rest)
		elif (start == "S:"):
			synopsis_long_lines.append(rest)
		elif (start == "h:"):
			field_pieces = rest.split("|")
			field_name = field_pieces[0]
			if (field_name == ''):
				if (len(field_pieces) < 2):
					hof_empty_line = 1		  # the default columns are enabled
					continue
			field_type = 0
			temp_field_custom = []
			field_custom = []
			if (len(field_pieces) > 2):
				temp_field_custom = field_pieces[1:]
				for piece in temp_field_custom:
					if (not piece in field_custom):
						field_custom.append(piece)
				field_type = 4
			elif (len(field_pieces) == 2):
				if (field_pieces[1] == "n<"):
					field_type = 0
				elif (field_pieces[1] == "n>"):
					field_type = 1
				elif (field_pieces[1] == "a<"):
					field_type = 2
				elif (field_pieces[1] == "a>"):
					field_type = 3
				else:
					raise ParsingException, gtline
			hof_temp.append([field_name, field_type, field_custom])
		else:
			raise ParsingException, gtline
	gtfile.close()
	game_title = game_title_temp
	credits = "<br>".join(credit_lines)
	difficulty = difficulty_temp
	synopsis_short = " ".join(synopsis_short_lines)
	synopsis_long = " ".join(synopsis_long_lines)
	hof_fields = []
	if ((hof_temp != []) or (hof_empty_line)):
		hof_fields = [["Name", 3, []], ["Moves", 1, []]]
		hof_fields.extend(hof_temp)
		hof_fields.append(["Date", 0, []])

def process_option(choice):
	""" Executes after one of the option buttons is clicked. """

	global act_list, moved_back, just_jumped, cur_parse_file, parse_context
	global S1, S2, S3, S4, S5

	
	# initialize system variables for the turn
	S1 = 0
	S2 = 0
	S3 = 0
	S4 = 0
	S5 = encode(cur_state.location)

	# find out in which file/position to find the option corresponding to the user's selection
	const_count = cur_state.visible_opt_count
	move_count = len(cur_state.move_options)
	if (choice > const_count + move_count):
		cmenu = 'action'
		selection = choice - (const_count + move_count) - 1
		S3 = 42
	elif (choice > const_count):
		cmenu = 'move'
		selection = choice - (const_count) - 1
		S3 = 94
	else:
		cmenu = 'const'
		selection = choice - 1
		S3 = 64
	
	# figure out location/spot to jump to, then find out the tag
	if (cmenu == 'const'):
		i = -1
		j = -1
		for opt in cur_state.const_options:
			i += 1
			if (opt.visible == 1):
				j += 1
			if (j == selection):
				selection = i
				break
		WriteText('@ ' + str(cur_state.moves + 1) + ' ' + cur_state.const_options[selection].text)
		if (narrate == 1):
			NarText('---(' + str(cur_state.moves + 1) + ' / ' + cur_state.location + ')  @ ' + cur_state.const_options[selection].text)
		if (narrate == 2):
			NarText('<b>(' + str(cur_state.moves + 1) + ' / ' + cur_state.location + ')  @ ' + cur_state.const_options[selection].text + '</b><br>')
		jump_spot = cur_state.const_options[selection].position
		jump_loc = cur_state.const_options[selection].location
		jump_tag = cur_state.const_options[selection].tag
		jump_type = '@ '
	elif (cmenu == 'move'):
		WriteText('^ ' + str(cur_state.moves + 1) + ' ' + cur_state.move_options[selection].text)
		if (narrate == 1):
			NarText('---(' + str(cur_state.moves + 1) + ' / ' + cur_state.location + ')  ^ ' + cur_state.move_options[selection].text)
		if (narrate == 2):
			NarText('<b>(' + str(cur_state.moves + 1) + ' / ' + cur_state.location + ')  ^ ' + cur_state.move_options[selection].text + '</b><br>')
		jump_spot = cur_state.move_options[selection].position
		jump_loc = cur_state.move_options[selection].location
		jump_tag = cur_state.move_options[selection].tag
		jump_type = '^ '
	else:
		WriteText('* ' + str(cur_state.moves + 1) + ' ' + cur_state.act_options[selection].text)
		if (narrate == 1):
			NarText('---(' + str(cur_state.moves + 1) + ' / ' + cur_state.location + ')  * ' + cur_state.act_options[selection].text)
		if (narrate == 2):
			NarText('<b>(' + str(cur_state.moves + 1) + ' / ' + cur_state.location + ')  * ' + cur_state.act_options[selection].text + '</b><br>')
		jump_spot = cur_state.act_options[selection].position
		jump_loc = cur_state.act_options[selection].location
		jump_tag = cur_state.act_options[selection].tag
		jump_type = '* '
	S4 = encode(jump_tag)
		
	""" code order:
		1. remove timers <= 0 and advance the remaining timers
		2. parse the code for the option chosen
		3. parse "repeatpre.sma"
		4. parse the new location's code
		5. parse "repeatpost.sma" """
	
	""" Process the code for the selected option, first checking at
		the "remembered" position, then rechecking the file if
		the option is not located there.  If it has to recheck
		the "copts.sma" file for a moved constant option, it
		updates the position of that option.  If it can't find
		the location at all, there's an error. """
	
	try:
		f3 = open(os.path.join(cur_state.directory, jump_loc + '.sma'), 'rU')
	except IOError:
		warn('Unable to start processing option supposedly in file "' + jump_loc + '"\nOption tag: "' + jump_tag + '"\nLocation file with this option not found.\n')
		cur_state.wl_state = -1
	else:
		try:
			f3.seek(jump_spot)
			atlinestart = (not jump_spot)
			while (not atlinestart):
				f3.seek(-1,1)
				ch = f3.read(1)
				if (ch != '\t'):
					atlinestart = 1
				else:
					f3.seek(-1,1)
			found_line = 2
			opt_depth = get_depth(f3)
			optline = f3.readline().strip()
			l = len(jump_tag)
			if (len(optline) < l+3):
				found_line = 1
			elif (optline[:(l+3)] != jump_type + jump_tag + ' '):
				found_line = 1
			if (found_line == 1):
				""" It failed to find option where it should be, so
					it now jumps back to the start of the location
					and searches for the moved option. """
				
				WriteText('Relocating option...')
				
				f3.seek(0)
				pos = 0
				success = 0
				optline = f3.readline().strip()
				if (len(optline) > l+2):
					if (optline[:(l+3)] == jump_type + jump_tag + ' '):
						success = 1
				while (not success):
					pos = f3.tell()
					opt_depth = get_depth(f3)
					optline = f3.readline().strip()
					if (optline == ''):
						test = f3.read(1)	# test whether this blank is at the end of the file
						if (test == ''):
							found_line = 0
							break
						else:
							f3.seek(-1,1)
					elif (len(optline) > l+2):
						if (optline[:(l+3)] == jump_type + jump_tag + ' '):
							success = 1
				if (success):
					if (jump_type == "@ "):
						""" If this is a constant option, update its position (use
							the value of 'pos' we end up with). """
	
						cur_state.const_options[selection].position = pos
					
			if (not found_line):
				warn('Unable to start processing option supposedly in file "' + jump_loc + '"\nOption tag: "' + jump_tag + '"\nLocation found, but no option with this tag located.\n')
				cur_state.wl_state = -1
			else:
				""" Back up the current state in last state,
					the last state into last2 state, etc. """
	
				last10_state.load(last9_state)
				last9_state.load(last8_state)
				last8_state.load(last7_state)
				last7_state.load(last6_state)
				last6_state.load(last5_state)
				last5_state.load(last4_state)
				last4_state.load(last3_state)
				last3_state.load(last2_state)
				last2_state.load(last_state)
				last_state.load(cur_state)
				
				cur_state.wl_state = 0		# in case it *was* 3, coming off a "win, but continue" situation
				cur_state.moves += 1
				
				init_genrand(jump_tag)
				adjust_timers()
				
				# dump old move/act menus/graphics, since they no longer are necessary, also reset act_text
				cur_state.act_text = ''
				act_list = []
				cur_state.move_options = []
				cur_state.act_options = []
				cur_state.act_image = ''
				cur_state.area_image = ''
				
				# determine if a jump is required, then process location code, regenerating options
				# if loss/win-permanently command encountered at any step, don't move on to next step
				
				cur_parse_file = jump_loc
				parse_context = 1
				parseblock(f3, jump_loc, opt_depth + 1)
				if ((cur_state.wl_state == 0) or (cur_state.wl_state == 3)):
					try:
						f_pre = open(os.path.join(cur_state.directory, 'repeatpre.sma'), 'rU')
						parse_context = 3
						cur_parse_file = 'repeatpre'
						parseblock(f_pre, 'repeatpre')
						f_pre.close()
					except IOError:
						pass
				parse_context = 0
				if ((cur_state.wl_state == 0) or (cur_state.wl_state == 3)):
					S5 = encode(next_jump)
					try:
						newloc_parse()
					except IOError:
						warn('Unable to open file "' + os.path.join(cur_state.directory, next_jump) + '.sma"\n')
						cur_state.wl_state = -1
						just_jumped = 0
				if ((cur_state.wl_state == 0) or (cur_state.wl_state == 3)):
					try:
						f_post = open(os.path.join(cur_state.directory, 'repeatpost.sma'), 'rU')
						parse_context = 4
						cur_parse_file = 'repeatpost'
						parseblock(f_post, 'repeatpost')
						f_post.close()
					except IOError:
						pass
				parse_context = 0
		except ParsingException, message:
			warn('PARSING ERROR:\n' + str(message))
			cur_state.wl_state = -1
			just_jumped = 0
			parse_context = 0
		f3.close()
	

def adjust_timers():
	""" every move, advance all the timers in the right direction """

	for t_key in cur_state.up_timers.keys():
		if (cur_state.up_timers[t_key] > 0):
			cur_state.up_timers[t_key] += 1
			if (not timermasked.has_key(t_key)):
				WriteText('\tt+ (' + t_key + ') increased to ' + str(cur_state.up_timers[t_key]))
		else:
			del cur_state.up_timers[t_key]
	for t_key in cur_state.down_timers.keys():
		if (cur_state.down_timers[t_key] > 0):
			cur_state.down_timers[t_key] -= 1
			if (not timermasked.has_key(t_key)):
				WriteText('\tt- (' + t_key + ') decreased to ' + str(cur_state.down_timers[t_key]))
		else:
			del cur_state.down_timers[t_key]
	
def newloc_parse():
	""" begin parsing a .sma file from the top """

	global next_jump, just_jumped, cur_parse_file, parse_context

	if (cur_state.location != next_jump):
		cur_state.location = next_jump
		WriteText('\t\tJumping to location "' + next_jump + '"')
		
	f2 = open(os.path.join(cur_state.directory, next_jump + '.sma'),'rU')
	if (just_jumped == 1):
		just_jumped = 2
	cur_parse_file = next_jump
	parse_context = 0	# perhaps not necessary, but just to be safe
	parseblock(f2, next_jump)
	f2.close()
	just_jumped = 0

def get_depth(f):
	""" If we're at the end of the file, -1 is returned.  Otherwise, if
		the line has just whitespace, we skip to the next line and
		get *its* depth, repeating as needed.  If/when we hit a line with
		non-whitespace characters, the "depth" returned is the number of
		tab characters immediately after that point in the file (it's
		assumed we're at the start of a line, going in. """
	global stats_compute

	while 1:
		depth = 0		# at this point we're positioned at the start of a line
		still_tabs = 1		# 1 if we have read no non-tabs so far on the line
		h = f.read(1)
		while (h != ''):
			if (h == '#'):
				if (stats_compute):
					com_count['#'] += 1
				f.readline()
				break
			elif ((h == '\n') or (h == '\r')):	# the line had only tabs and spaces, so we go to next line
				f.seek(-1, 1)
				f.readline()
				break		# skip to the next line, and start get_depth over
			elif ((h == '\t') and (still_tabs)):	# we're still adding to indentation if have only had tabs
				depth += 1
			elif (h == ' '):
				still_tabs = 0	# tabs no longer in a streak, so no more indentation incrementing
			elif (h != '\t'):
				f.seek(-1, 1)
				return depth
			h = f.read(1)
		if (h == ''):
			return -1			# we hit this if we reach the end of the file


	while (h == '\t'):
		depth += 1
		h = f.read(1)
	if (h == '#'):
		""" A bit of a hack, but setting the depth this high for a commented
			line virtually ensures it'll beat out the indentation of
			any other line and not signal the end of a code block. """
		depth = 1000000	
	f.seek(-1, 1)	# set the file position back by one because we just read the first non-tab
	return depth

def warn(message):
	error_text = '\n***************  WARNING ***************\n\n'
	error_text += message
	error_text += 'Parsing was stopped at that point.\n'
	WriteText(error_text)

def parseblock(f, parse_loc, block_depth=0):
	""" parses a block with file handle "f" and name "parse_loc", at
		depth "block_depth".  A whole .sma file is counted as a block at depth 0.
		The block is over when the depth of the next line is
		measured to be less than the block depth. At the end of
		the file, a depth of -1 being measured thus ends parsing properly. """

	global next_jump
	global com_count
	global act_list, area_list, block_text_lines, log_lines
	global stats_compute
			
	""" prec_cond = 1 if the last command at the given indentation
		level in the block was a conditional, so we can "link" to
		it with a "C" command. """
	prec_cond = 0

	""" parsestatus =
		0: normally
		1: skipping over the lines of a block which shouldn't be
			parsed yet (e.g. option code blocks) -- except
			blocks of failed conditionals (see 4)
		2: in the middle of an action text block, so the parsed
			line gets interpreted as literal text to
			append as long as the indentation is exactly
			1 more than the block depth
		3: same as 2, but for area text blocks
		4: skipping over the lines of a failed conditional's
			block (so we know a "C" is appropriate if it
			comes up when we return to the main block's
			indentation level).
		5: just coming out of a loop's block, so we need to
			skip back to the loop's head (then set
			parsestatus to 0) """
	parsestatus = 0
	pos = f.tell()
	next_depth = get_depth(f)

	""" We break out of the while loop when we hit a line indented less
		than the given level of indentation *unless* we're
		coming out of a Smash "L" loop (parsestatus == 5), because
		then we want to return to the "L" line and possibly run
		the loop again. """
	line = 'line'
	while ((next_depth >= block_depth) or (parsestatus == 5) or (line == '')):
		if (parsestatus == 5):
			f.seek(loop_pos)
			pos = loop_pos
			next_depth = get_depth(f)
			parsestatus = 0
		line = f.readline().strip()
		#WriteText('line = ' + line)
		if (not len(line)):
			WriteText('blank line at position ' + str(pos))
			test = f.read(1)
			if (test == ''):
				break
			f.seek(-1,1)
			pos = f.tell()
			WriteText('after, pos ' + str(pos))
			next_depth = get_depth(f)
			pos = f.tell()
			WriteText('finally, pos ' + str(pos))
			continue
		if (line[0] == '#'):
			WriteText('# at position ' + str(pos))
			pos = f.tell()
			WriteText('after, pos ' + str(pos))
			next_depth = get_depth(f)
			pos = f.tell()
			WriteText('finally, pos ' + str(pos))
			if (stats_compute):
				com_count['#'] += 1
			continue
		if (parsestatus == 2):
			if (next_depth == block_depth + 1):
				pos = f.tell()
				next_depth = get_depth(f)
				if (stats_compute):
					spaces = line.count(' ')
					wordcount = CountWords(line)
					com_count['words'] += wordcount
					com_count['wordbytes'] += (len(line) + 1)
				else:
					block_text_lines.append(" " + line)
					if (log_lines):
						WriteText('>>>>> \t' + line)
				continue
			elif (next_depth == block_depth):
				""" end of description text block, so return to normal
					parsing status after appending the text lines
					that were being collected. """

				act_entry = "".join(block_text_lines)
				expression = act_entry
				act_entry = eval_string_imagetags(expression)
				act_list.append(act_entry)
				block_text_lines = []
				parsestatus = 0
			else:
				raise ParsingException, 'Over-indented line in action description block.\nLine: "' + line + '"\n'
		if (parsestatus == 3):
			if (next_depth == block_depth + 1):
				pos = f.tell()
				next_depth = get_depth(f)
				if (stats_compute):
					spaces = line.count(' ')
					wordcount = CountWords(line)
					com_count['words'] += wordcount
					com_count['wordbytes'] += (len(line) + 1)
				else:
					block_text_lines.append(" " + line)
					if (log_lines):
						WriteText('>>>>> \t' + line)
				continue
			elif (next_depth == block_depth):
				""" end of description text block, so return to normal
					parsing status after appending the text lines
					that were being collected. """

				area_entry = "".join(block_text_lines)
				expression = area_entry
				area_entry = eval_string_imagetags(expression)
				area_list.append(area_entry)
				block_text_lines = []
				parsestatus = 0
			else:
				raise ParsingException, 'Over-indented line in area description block.\nLine: "' + line + '"\n'
		if ((parsestatus == 0) and (next_depth > block_depth)):
			raise ParsingException, 'Over-indented Smash command.\nLine: "' + line + '"\n'
		if ((parsestatus == 1) or (parsestatus == 4)):
			""" Note that we actually don't skip lines for these parsestatus
				values if we're computing statistics, because *all* lines
				get parsed then. """

			if ((next_depth == block_depth) or (stats_compute)):
				if (parsestatus == 1):
					parsestatus = 0
			else:
				pos = f.tell()
				next_depth = get_depth(f)
				continue
		if ((next_depth == block_depth) and (line[0] == 'C')):
			if (not prec_cond):	# no preceding 'c' conditional to link it to
				raise ParsingException, '"C" conditional parsed with no linked precedent "c" conditional.\nLine: "' + line + '"\n'
			if (parsestatus != 4):	# previous conditional on that level passed, so we skip this block
				pos = f.tell()
				if (stats_compute):	# unless computing statistics -- parse the block after all
					com_count['C'] += 1
					parsestatus = 0
					parseblock(f, parse_loc, next_depth+1)
				else:
					parsestatus = 1
				next_depth = get_depth(f)
				continue
		# unlink the previously seen conditional if a non-conditional comes in the way here
		if (log_lines and not stats_compute):
			WriteText('>>>>> ' + line)
		prec_cond = 0
		if ((line[0] == 'c') or (line[0] == 'C')):
			""" This conditional should be parsed (in any case,
				if 'c', and because there's a failed
				precedent, if 'C', or we'd have left the
				loop at the "continue" earlier). """

			prec_cond = 1
			parsestatus = 0
		if (parsestatus == 4):
			parsestatus = 0
		if (parsestatus == 0):
			if (line[0] == 'L'):
				loop_pos = pos	# remember the loop head position for later, so we can jump back quickly
			parsestatus = parseline(f, next_depth, pos, line, parse_loc)
			pos = f.tell()
			next_depth = get_depth(f)
	if (next_depth > 0):
		f.seek(-next_depth, 1)	# return the file position to the start of the line
	if (not stats_compute):
		if (parsestatus == 2):
			act_entry = "".join(block_text_lines)
			expression = act_entry
			act_entry = eval_string_imagetags(expression)
			act_list.append(act_entry)
			block_text_lines = []
		if (parsestatus == 3):
			area_entry = "".join(block_text_lines)
			expression = area_entry
			area_entry = eval_string_imagetags(expression)
			area_list.append(area_entry)
			block_text_lines = []
	return

def context_error_text(command_context):
	""" Supplies the last bit of text for a parsing error based on
		finding a command in the wrong context. """

	global parse_context

	error_text = '  This command is'
	if (command_context == 0):
		error_text += ' not allowed in code parsed in (or in a function called from)'
		if (parse_context == 1):
			error_text += ' an option code block.'
		if (parse_context == 2):
			error_text += ' the function code block for invoking an inventory item.'
	if (command_context == 1):
		error_text += ' only allowed in option code blocks and (e)objectactions functions called directly from invoking an inventory item.'
	if (command_context == 2):
		error_text += ' only allowed in function code blocks.'
	if (command_context == 3):
		error_text += ' not allowed inside a code block.'
	if (command_context == 4):
		error_text += ' only allowed in option code blocks, (e)objectactions functions called directly from invoking an inventory item, and "repeatpre.sma" code executed every turn.'
	return error_text
	
def parseline(f, next_depth, pos, line, parse_loc, edit=0):
	""" parses a single line of Smash code (usually called from
		parse_block, but also when using the editor (when
		edit=1) to put in a Smash line directly, and *indirectly* when
		examining an inventory item, when the constructed
		line "f (e)objectactions.item_name" is passed in. """

	global next_jump, just_jumped, stats_compute, com_count, cur_parse_file
	global act_list, area_list, block_text_lines, tag_regex, tag_regex2, opdict
	global gvarmasked, svarmasked, robjmasked, eobjmasked, timermasked, parse_context
	global expression, S1, S2, inv_change, hof_fields, playername, new_hallfile
	global hof_track, hall_filename

	try:
		ch = line[0]
	except IndexError:
		return 0	# blank lines are okay, but skip them
	if ((parse_loc == 'copts') and (next_depth == 0) and (ch != '@')):
		raise ParsingException, 'Only constant option definitions are allowed in "copts.sma".\nLine: "' + line + '"\n'
	raw_line = line[1:]
	strip_line = raw_line.strip()
	if (ch.upper() == 'C' and not edit):
		if (not len(strip_line) and (ch == 'c')):
			raise ParsingException, 'No expression supplied to "c" command.\nLine: "' + line + '"\n'
		if (not len(strip_line)):
			condit = 1
		elif ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
			raise ParsingException, 'Space or tab expected after start of "' + ch + '" command.\nLine: "' + line + '"\n'
		else:	
			expression = strip_line
			condit_pre = eval_string(expression)
			expression = condit_pre
			condit = eval_exp(expression)
		if (stats_compute):
			com_count[ch] += 1
		if (condit or stats_compute):
			parseblock(f, parse_loc, next_depth+1)
			return 0
		else:
			return 4
	elif (ch == ',' or ch == ';'):
		if (parse_context in [1, 2]):
			error_text = 'Command "' + ch + '" in the wrong context.' + context_error_text(0) + '\nLine: "' + line + '"\n'
			raise ParsingException, error_text
		if (len(strip_line)):
			if ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
				raise ParsingException, 'Space, tab, or end of line expected after start of "' + ch + '" command.\nLine: "' + line + '"\n'
		if (stats_compute):
			com_count[ch] += 1
			wordcount = CountWords(strip_line)
			com_count['words'] += wordcount
			com_count['wordbytes'] += (len(strip_line))
			if (ch == ','):
				com_count['wordbytes'] += 1
		else:
			if (ch == ','):
				block_text_lines.append(" " + strip_line)
			else:
				block_text_lines.append(strip_line)
		return 3
	elif (ch == '.'):
		if (parse_context in [1, 2]):
			error_text = 'Command "." in the wrong context.' + context_error_text(0) + '\nLine: "' + line + '"\n'
			raise ParsingException, error_text
		if (len(strip_line)):
			if ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
				raise ParsingException, 'Space, tab, or end of line expected after start of "." command.\nLine: "' + line + '"\n'
		if (stats_compute):
			com_count['.'] += 1
			wordcount = CountWords(strip_line)
			com_count['words'] += wordcount
			com_count['wordbytes'] += (len(strip_line))
		else:
			area_list = []
			block_text_lines = [strip_line]
		return 3
	elif (ch == 'p'):
		if (len(strip_line)):
			if ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
				raise ParsingException, 'Space, tab, or end of line expected after start of "p" command.\nLine: "' + line + '"\n'
		if (stats_compute):
			com_count['p'] += 1
			wordcount = CountWords(strip_line)
			com_count['words'] += wordcount
			com_count['wordbytes'] += (len(strip_line))
		else:
			act_list = []
			block_text_lines = [strip_line]
		return 2
	elif (ch == 'P' or ch == 'A'):
		if (len(strip_line)):
			if ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
				raise ParsingException, 'Space, tab, or end of line expected after start of "' + ch + '" command.\nLine: "' + line + '"\n'
		if (stats_compute):
			com_count[ch] += 1
			wordcount = CountWords(strip_line)
			com_count['words'] += wordcount
			com_count['wordbytes'] += (len(strip_line))
			if (ch == 'P'):
				com_count['wordbytes'] += 1
		else:
			if (ch == 'P'):
				block_text_lines.append(" " + strip_line)
			else:
				block_text_lines.append(strip_line)
		return 2
	elif (ch == 'f'):
		if not len(strip_line):
			raise ParsingException, 'No tag-expression supplied to "f" command.\nLine: "' + line + '"\n'
		if ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
			raise ParsingException, 'Space or tab expected after start of "f" command.\nLine: "' + line + '"\n'
		expression = strip_line
		funcline = eval_string(expression)
		func_pieces = funcline.split(".")
		if (len(func_pieces) > 2):
			raise ParsingException, 'Too many "." separators in tag of "f" command.\nLine: "' + line + '"\n'
		if (len(func_pieces) == 2):
			if (not len(func_pieces[0]) or not len(func_pieces[1])):
				raise ParsingException, 'Empty string to one side of "." separator in tag of "f" command.\nLine: "' + line + '"\n'
			func_loc = func_pieces[0]
			func_name = func_pieces[1]
		else:
			func_loc = parse_loc
			func_name = func_pieces[0]
		if (stats_compute):
			com_count['f'] += 1
			return 0
		else:
			if (not tag_regex3.match(func_loc)):
				raise ParsingException, 'Uppercase character in location tag "' + func_loc + '" of "f" command.\n'
			try:
				f4 = open(os.path.join(cur_state.directory, func_loc + '.sma'), 'rU')
				f4.seek(0)
				find_line = f4.readline()
				if (find_line == ''):
					f4.close()
					raise ParsingException, 'Function "' + func_name + '" not found in file "' + func_loc + '".\nLocation found, but no function with this tag located.\n'
				find_line = find_line.strip()
				while (find_line != ("~ " + func_name)):
					find_line = f4.readline()
					if (find_line == ''):
						f4.close()
						raise ParsingException, 'Function "' + func_name + '" not found in file "' + func_loc + '".\nLocation found, but no function with this tag located.\n'
					find_line = find_line.strip()
				S2 += 1
				S1 = 0
				cur_parse_file = func_loc
				parseblock(f4, func_loc, 1)
				cur_parse_file = parse_loc
				f4.close()
				S2 -= 1
				return 0
			except IOError:
				raise ParsingException, 'Function "' + func_name + '" called, but source file "' + func_loc + '" not found.\n'
	elif (ch == 'v'):
		if (not len(strip_line)):
			raise ParsingException, 'No arguments supplied to "v" command.\nLine: "' + line + '"\n'
		if ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
			raise ParsingException, 'Space, tab, or end of line expected after start of "v" command.\nLine: "' + line + '"\n'
		expression = line
		v_str = eval_string(expression)
		v_pieces = v_str.split()
		if (len(v_pieces) < 4):
			raise ParsingException, 'Not enough arguments supplied to "v" command.\nLine: "' + line + '"\n'
		v_var = v_pieces[1]
		if not (('{' in strip_line) and ('{' not in v_var)):
			if (stats_compute):
				""" A negative sign can creep into a variable tag during stats_compute but
					still be a legal Smash tag if it never actually turns up negative
					during gameplay.  So we allow "-" with tag_regex2 if doing stats
					computation. """
				if (not tag_regex2.match(v_var)):
					raise ParsingException, 'Illegal character in variable tag of "v" command.\nLine: "' + line + '"\n'
			else:
				if (not tag_regex.match(v_var)):
					raise ParsingException, 'Illegal character in variable tag of "v" command.\nLine: "' + line + '"\n'
		if (v_var.isdigit()):
			""" We need to support the old AGLL syntax of setting variables with
				just numbers -- a leading "v" was added only for checking
				the variable's value. """
			v_var = 'v' + v_var
		v_op = v_pieces[2]
		if (not opdict.has_key(v_op)):
			raise ParsingException, 'Illegal operator in "v" command.\nLine: "' + line + '"\n'
		if (not stats_compute):
			if ((cur_state.svars.has_key(v_var)) or (cur_state.up_timers.has_key(v_var)) or (cur_state.down_timers.has_key(v_var))):
				raise ParsingException, 'Illegal setting of global variable already in the namespace in another form.\nLine: "' + line + '"\n'
		v_exp = " ".join(v_pieces[3:])
		if (v_op != '='):
			v_exp = " ".join([v_var, v_op, v_exp])
			v_op = "="
		expression = v_exp
		v_value = eval_exp(expression)
		if (stats_compute):
			com_count['v'] += 1
		else:
			if (not gvarmasked.has_key(v_var)):
				WriteText('\tv (' + v_var + ') = ' + str(v_value))
			if (v_value):
				cur_state.gvars[v_var] = v_value
			elif (cur_state.gvars.has_key(v_var)):
				del cur_state.gvars[v_var]
		return 0
	elif (ch == 'g'):
		if ((not parse_context in [1, 2, 3]) and (parse_loc != 'start')):
			if (not edit and not stats_compute):
				error_text = 'Command "g" in the wrong context.' + context_error_text(4) + '\nLine: "' + line + '"\n'
				raise ParsingException, error_text
		if not len(strip_line):
			raise ParsingException, 'No tag-expression supplied to "g" command.\nLine: "' + line + '"\n'
		if ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
			raise ParsingException, 'Space or tab expected after start of "g" command.\nLine: "' + line + '"\n'
		expression = strip_line
		loc_jump = eval_string(expression)
		if (not tag_regex3.match(loc_jump)):
			if not ((stats_compute) and ('{' in strip_line) and ('{' not in loc_jump)):
				if (tag_regex.match(loc_jump)):
					raise ParsingException, 'Uppercase character in location tag of "g" command.\nLine: "' + line + '"\n'
				else:
					raise ParsingException, 'Illegal character in location tag of "g" command.\nLine: "' + line + '"\n'
		if (stats_compute):
			com_count['g'] += 1
		try:
			f2 = open(os.path.join(cur_state.directory, loc_jump + '.sma'), 'rU')
			f2.close()
		except IOError:
			""" The following is to get around the possibility that during statistics
				computation, a line like "g loc{var}" may have {var} evaluate
				to something unattainable during normal play of the game, which
				might cause an error to be flagged improperly. """
			if ((not stats_compute) or (loc_jump == expression)):
				raise ParsingException, 'Location file pertaining to jump-to location of "g" command not found.\nLine: "' + line + '"\n'
		if (not stats_compute):
			next_jump = loc_jump
			just_jumped = 1
		return 0
	elif (ch == '^' and not edit):
		if (parse_context in [1, 2]):
			error_text = 'Command "^" in the wrong context.' + context_error_text(0) + '\nLine: "' + line + '"\n'
			raise ParsingException, error_text
		if (not len(strip_line)):
			raise ParsingException, 'No arguments supplied to "^" command.\nLine: "' + line + '"\n'
		if ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
			raise ParsingException, 'Space or tab expected after start of "^" command.\nLine: "' + line + '"\n'
		move_pieces = line.split()
		move_tag = move_pieces[1]
		if (not tag_regex.match(move_tag)):
			raise ParsingException, 'Illegal character in tag of "^" command.\nLine: "' + line + '"\n'
		tag_repeat = 0
		for option in cur_state.move_options:
			if (option.tag == move_tag):
				tag_repeat = 1
		if (tag_repeat):
			raise ParsingException, 'Tag of "^" command duplicates previously registered "^" command.\nLine: "' + line + '"\n'
		expression = " ".join(move_pieces[2:])
		move_text = eval_string(expression)
		cur_state.move_options.append(Option('^', parse_loc, pos, move_tag, move_text))
		if (stats_compute):
			com_count['^'] += 1
			parse_context = 1
			parseblock(f, parse_loc, next_depth+1)
			parse_context = 0
		return 1
	elif (ch == '*' and not edit):
		if (parse_context in [1, 2]):
			error_text = 'Command "*" in the wrong context.' + context_error_text(0) + '\nLine: "' + line + '"\n'
			raise ParsingException, error_text
		if (not len(strip_line)):
			raise ParsingException, 'No arguments supplied to "*" command.\nLine: "' + line + '"\n'
		if ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
			raise ParsingException, 'Space or tab expected after start of "*" command.\nLine: "' + line + '"\n'
		acti_pieces = line.split()
		acti_tag = acti_pieces[1]
		if (not tag_regex.match(acti_tag)):
			raise ParsingException, 'Illegal character in tag of "*" command.\nLine: "' + line + '"\n'
		tag_repeat = 0
		for option in cur_state.act_options:
			if (option.tag == acti_tag):
				tag_repeat = 1
		if (tag_repeat):
			raise ParsingException, 'Tag of "*" command duplicates previously registered "*" command.\nLine: "' + line + '"\n'
		expression = " ".join(acti_pieces[2:])
		acti_text = eval_string(expression)
		cur_state.act_options.append(Option('*', parse_loc, pos, acti_tag, acti_text))
		if (stats_compute):
			com_count['*'] += 1
			parse_context = 1
			parseblock(f, parse_loc, next_depth+1)
			parse_context = 0
		return 1
	elif (ch == 'l' and not edit):
		if (len(strip_line)):
			raise ParsingException, 'Extra character(s) following "l" command.\nLine: "' + line + '"\n'
		if (stats_compute):
			com_count['l'] += 1
		if (cur_state.wl_state == 0):	# the "loss" state doesn't get set if a "win" state has been set this move
			cur_state.wl_state = 1
		return 0
	elif (ch == 's'):
		if (parse_context in [3, 4]):
			raise ParsingException, 'The repeat files can not be assigned state variables.\nLine: "' + line + '"\n'
		if (cur_state.location == 'start'):
			raise ParsingException, '"start.sma" can not be assigned state variables.\nLine: "' + line + '"\n'
		if (not len(strip_line)):
			raise ParsingException, 'No arguments supplied to "s" command.\nLine: "' + line + '"\n'
		if ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
			raise ParsingException, 'Space or tab expected after start of "s" command.\nLine: "' + line + '"\n'
		expression = line
		s_str = eval_string(expression)
		s_pieces = s_str.split()
		if (len(s_pieces) < 3):
			raise ParsingException, 'Not enough arguments supplied to "s" command.\nLine: "' + line + '"\n'
		if (opdict.has_key(s_pieces[1])):
			s_var = 's'
			s_op = s_pieces[1]
			s_exp = " ".join(s_pieces[2:])
		else:
			s_var = s_pieces[1]
			if not (('{' in strip_line) and ('{' not in s_var)):
				if (stats_compute):
					""" A negative sign can creep into a variable tag during stats_compute but
						still be a legal Smash tag if it never actually turns up negative
						during gameplay.  So we allow "-" with tag_regex2 if doing stats
						computation. """
	
					if (not tag_regex2.match(s_var)):
						raise ParsingException, 'Illegal character in variable tag of "s" command.\nLine: "' + line + '"\n'
				else:
					if (not tag_regex.match(s_var)):
						raise ParsingException, 'Illegal character in variable tag of "s" command.\nLine: "' + line + '"\n'
				if (s_var.isdigit()):
					raise ParsingException, 'Tag of state variable has only numeric characters in "s" command.\nLine: "' + line + '"\n'
			s_op = s_pieces[2]
			if (not opdict.has_key(s_op)):
				raise ParsingException, 'Illegal operator in "s" command.\nLine: "' + line + '"\n'
			s_exp = " ".join(s_pieces[3:])
		if (not stats_compute):
			if ((cur_state.gvars.has_key(s_var)) or (cur_state.up_timers.has_key(s_var)) or (cur_state.down_timers.has_key(s_var))):
				raise ParsingException, 'Illegal setting of state variable already in the namespace in another form.\nLine: "' + line + '"\n'
		if (s_op != '='):
			s_exp = " ".join([s_var, s_op, s_exp])
			s_op = "="
		expression = s_exp
		s_value = eval_exp(expression)
		if (stats_compute):
			com_count['s'] += 1
		else:
			mask = 0
			if (svarmasked.has_key(s_var)):
				if (svarmasked[s_var].has_key(cur_state.location)):
					mask = 1
			if (not mask):
				WriteText('\ts (' + s_var + ') = ' + str(s_value))
			if (s_value):
				if (not cur_state.svars.has_key(s_var)):
					cur_state.svars[s_var] = {}
				cur_state.svars[s_var][cur_state.location] = s_value
			else:
				if (cur_state.svars.has_key(s_var)):
					if (cur_state.svars[s_var].has_key(cur_state.location)):
						del cur_state.svars[s_var][cur_state.location]
					if (cur_state.svars[s_var] == {}):
						del cur_state.svars[s_var]
		return 0
	elif (ch == 'G' and not edit):
		if (parse_context in [1, 2]):
			error_text = 'Command "G" in the wrong context.' + context_error_text(0) + '\nLine: "' + line + '"\n'
			raise ParsingException, error_text
		if (len(strip_line)):
			raise ParsingException, 'Extra character(s) following "G" command.\nLine: "' + line + '"\n'
		if (stats_compute):
			com_count['G'] += 1
		if ((just_jumped == 2) or stats_compute):
			parseblock(f, parse_loc, next_depth+1)
			return 0
		else:
			return 1
	elif (ch == 'r'):
		if (not S2):
			error_text = 'Command "r" in the wrong context.' + context_error_text(2) + '\nLine: "' + line + '"\n'
			raise ParsingException, error_text
		if not len(strip_line):
			raise ParsingException, 'No tag-expression supplied to "r" command.\nLine: "' + line + '"\n'
		if ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
			raise ParsingException, 'Space or tab expected after start of "r" command.\nLine: "' + line + '"\n'
		if (stats_compute):
			com_count['r'] += 1
		expression = strip_line
		expression = eval_string(expression)
		S1 = eval_exp(expression)
		return 0
	elif (ch == 'L' and not edit):
		if (not len(strip_line)):
			raise ParsingException, 'No expression supplied to "L" command.\nLine: "' + line + '"\n'
		if ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
			raise ParsingException, 'Space or tab expected after start of "L" command.\nLine: "' + line + '"\n'
		expression = strip_line
		expression = eval_string(strip_line)
		condit = eval_exp(expression)
		if (stats_compute):
			com_count['L'] += 1
		if (condit or stats_compute):
			parseblock(f, parse_loc, next_depth+1)
			if (stats_compute):
				return 1
			else:
				return 5
		else:
			return 1
	elif (ch == 'e'):
		if (not len(strip_line)):
			raise ParsingException, 'No arguments supplied to "e" command.\nLine: "' + line + '"\n'
		if ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
			raise ParsingException, 'Space or tab expected after start of "e" command.\nLine: "' + line + '"\n'
		expression = line
		e_str = eval_string(expression)
		e_pieces = e_str.split()
		if (len(e_pieces) < 4):
			raise ParsingException, 'Not enough arguments supplied to "e" command.\nLine: "' + line + '"\n'
		e_tag = e_pieces[1]
		if (not tag_regex.match(e_tag)):
			raise ParsingException, 'Illegal character in tag of "e" command.\nLine: "' + line + '"\n'
		if (not enum_names.has_key(e_tag)):
			if not ((stats_compute) and ('{' in strip_line) and ('{' not in e_tag)):
				# no enumerated object of that tag
				raise ParsingException, 'Unrecognized tag in "e" command.\nLine: "' + line + '"\n'
		e_op = e_pieces[2]
		if (not opdict.has_key(e_op)):
			raise ParsingException, 'Illegal operator in "e" command.\nLine: "' + line + '"\n'
		e_exp = " ".join(e_pieces[3:])
		if (e_op != '='):
			e_exp = " ".join(['e:', e_tag, e_op, e_exp])
			e_op = "="
		expression = e_exp
		e_value = eval_exp(expression)
		if (stats_compute):
			com_count['e'] += 1
		else:
			if (not eobjmasked.has_key(e_tag)):
				WriteText('\te (' + e_tag + ') = ' + str(e_value) + '\t\t\t(' + enum_names[e_tag][1] + ')')
			if (e_value):
				cur_state.enum_inv[e_tag] = e_value
			elif (cur_state.enum_inv.has_key(e_tag)):
				del cur_state.enum_inv[e_tag]
			inv_change = 1
		return 0
	elif (ch == 'a'):
		if (not len(strip_line)):
			raise ParsingException, 'No tag-expression supplied to "a" command.\nLine: "' + line + '"\n'
		if ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
			raise ParsingException, 'Space or tab expected after start of "a" command.\nLine: "' + line + '"\n'
		expression = strip_line
		tag = eval_string(expression)
		if (len(tag.split()) != 1):
			raise ParsingException, 'More than one argument supplied to "a" command.\nLine: "' + line + '"\n'
		if (not tag_regex.match(tag)):
			raise ParsingException, 'Illegal character in tag of "a" command.\nLine: "' + line + '"\n'
		if (not reg_names.has_key(tag)):
			if not ((stats_compute) and ('{' in strip_line) and ('{' not in tag)):
				raise ParsingException, 'Unrecognized tag in "a" command.\nLine: "' + line + '"\n'
		if (stats_compute):
			com_count['a'] += 1
		else:
			cur_state.reg_inv[tag] = 1
			if (not robjmasked.has_key(tag)):
				WriteText('\ta (' + tag + ')\t\t\t(' + reg_names[tag] + ')')
			inv_change = 1
		return 0
	elif (ch == 'd'):
		if (not len(strip_line)):
			raise ParsingException, 'No tag-expression supplied to "d" command.\nLine: "' + line + '"\n'
		if ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
			raise ParsingException, 'Space or tab expected after start of "d" command.\nLine: "' + line + '"\n'
		expression = strip_line
		tag = eval_string(expression)
		if (len(tag.split()) != 1):
			raise ParsingException, 'More than one argument supplied to "d" command.\nLine: "' + line + '"\n'
		if (not tag_regex.match(tag)):
			raise ParsingException, 'Illegal character in tag of "d" command.\nLine: "' + line + '"\n'
		if (not reg_names.has_key(tag)):
			if not ((stats_compute) and ('{' in strip_line) and ('{' not in tag)):
				raise ParsingException, 'Unrecognized tag in "d" command.\nLine: "' + line + '"\n'
		if (stats_compute):
			com_count['d'] += 1
		else:
			if cur_state.reg_inv.has_key(tag):
				del cur_state.reg_inv[tag]
				inv_change = 1
			if (not robjmasked.has_key(tag)):
				WriteText('\td (' + tag + ')\t\t\t(' + reg_names[tag] + ')')
		return 0
	elif (ch == 't'):
		if (not len(strip_line)):
			raise ParsingException, 'No arguments supplied to "t" command.\nLine: "' + line + '"\n'
		if ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
			raise ParsingException, 'Space or tab expected after start of "t" command.\nLine: "' + line + '"\n'
		expression = line
		t_str = eval_string(expression)
		t_pieces = t_str.split()
		if ((t_pieces[1] == '<<') or (t_pieces[1] == '>>')):
			expression = " ".join(t_pieces[2:])
			t_value = eval_exp(expression)
			if (t_pieces[1] == '<<'):
				if (stats_compute):
					com_count['t<<'] += 1
				else:
					for t_key in cur_state.up_timers.keys():
						cur_state.up_timers[t_key] -= t_value
					for t_key in cur_state.down_timers.keys():
						cur_state.down_timers[t_key] += t_value
					if (cur_state.up_timers != {} or cur_state.down_timers != {}):
						WriteText('All timers rewound by ' + str(t_value))
			else:
				if (stats_compute):
					com_count['t>>'] += 1
				else:
					for t_key in cur_state.up_timers.keys():
						cur_state.up_timers[t_key] += t_value
					for t_key in cur_state.down_timers.keys():
						cur_state.down_timers[t_key] -= t_value
					if (cur_state.up_timers != {} or cur_state.down_timers != {}):
						WriteText('All timers advanced by ' + str(t_value))
		elif ((t_pieces[1] == '+') or (t_pieces[1] == '-')):
			if (len(t_pieces) < 5):
				raise ParsingException, 'Not enough arguments supplied to "t" command.\nLine: "' + line + '"\n'
			t_name = t_pieces[2]
			if (t_name.isdigit()):
				raise ParsingException, 'Tag of timer has only numeric characters in "t" command.\nLine: "' + line + '"\n'
			if (not tag_regex.match(t_name)):
				raise ParsingException, 'Illegal character in tag of "t" command.\nLine: "' + line + '"\n'
			if (not stats_compute):
				if ((cur_state.gvars.has_key(t_name)) or (cur_state.svars.has_key(t_name))):
					raise ParsingException, 'Illegal setting of timer already in the namespace in another form.\nLine: "' + line + '"\n'
			t_op = t_pieces[3]
			if (not opdict.has_key(t_op)):
				raise ParsingException, 'Illegal operator in "t" command.\nLine: "' + line + '"\n'
			t_exp = " ".join(t_pieces[4:])
			if (t_op != '='):
				t_exp = " ".join([t_name, t_op, t_exp])
				t_op = "="
			expression = t_exp
			t_value = eval_exp(expression)
			if (t_pieces[1] == '+'):
				if (stats_compute):
					com_count['t+'] += 1
				else:
					cur_state.up_timers[t_name] = t_value
					if cur_state.down_timers.has_key(t_name):
						del cur_state.down_timers[t_name]
					if (not timermasked.has_key(t_name)):
						WriteText('\tt+ (' + t_name + ') = ' + str(t_value))
			else:
				if (stats_compute):
					com_count['t-'] += 1
				else:
					cur_state.down_timers[t_name] = t_value
					if cur_state.up_timers.has_key(t_name):
						del cur_state.up_timers[t_name]
					if (not timermasked.has_key(t_name)):
						WriteText('\tt- (' + t_name + ') = ' + str(t_value))
		else:
			raise ParsingException, '"+", "-", "<<", or ">>" expected after start of "t" command.\nLine: "' + line + '"\n'
		return 0
	elif (ch.upper() == 'I'):
		if ((parse_context in [1, 2]) and (ch == 'i')):
			error_text = 'Command "i" in the wrong context.' + context_error_text(0) + '\nLine: "' + line + '"\n'
			raise ParsingException, error_text
		if (not len(strip_line)):
			raise ParsingException, 'No image file name supplied to "' + ch + '" command.\nLine: "' + line + '"\n'
		if ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
			raise ParsingException, 'Space or tab expected after start of "' + ch + '" command.\nLine: "' + line + '"\n'
		if (stats_compute):
			com_count[ch] += 1
		elif (ch == 'i'):
			cur_state.area_image = os.path.join(cur_state.directory, strip_line)
		else:
			cur_state.act_image = os.path.join(cur_state.directory, strip_line)
		return 0
	elif (ch == '#' and not edit):
		if (stats_compute):
			com_count['#'] += 1
		return 0
	elif ((ch.upper() == 'W') and (not edit)):
		if (cur_state.wl_state == 0):
			expression = strip_line
			expression = eval_string(expression)
			w_pieces = expression.split("|")
			try:
				UpdateHoFData()
			except:
				raise ParsingException, 'Unable to process the win command.\nLine: "' + line + '"\n'
			if (expression == ""):
				if len(hof_fields):
					param_num = 3
				else:
					param_num = 0
			else:
				param_num = len(w_pieces) + 3
			if (param_num != len(hof_fields)):
				WriteText('param_num = ' + str(param_num) + ' and len(hof_fields) = ' + str(len(hof_fields)))
				raise ParsingException, 'Length of win parameter list doesn\'t match the number of Hall of Fame fields.\n\nLine: "' + line + '"\n'
			par_eval = [playername, str(cur_state.moves)]
			for index in range (2, param_num-1):	# we skip the indices for the "Name", "Moves", and "Date"
				parameter = w_pieces[index-2]
				type = hof_fields[index][1]
				if (type < 2):
					expression = parameter
					par_eval.append(str(eval_exp(expression)))
				else:
					if (parameter == ''):
						raise ParsingException, 'Win parameter #' + str(index-1) + ' is blank.\n\nLine: "' + line + '"\n'
					par_eval.append(parameter)
					if (type == 4):
						if (parameter not in hof_fields[index][2]):
							raise ParsingException, 'Win parameter #' + str(index-1) + ' is not in the custom entry list for the corresponding Hall of Fame field.\n\nLine: "' + line + '"\n'
			if ((not stats_compute) and (param_num)):
				writenew = 0
				try:
					hallsize = os.path.getsize(hall_filename)
					hallfile = open(hall_filename, 'rU')

					# here is code for checking that the hof_fields agree with the text in the HoF file

					index = 0
					incompatible = 0
					while 1:
						hof_line = hallfile.readline().strip()
						if (hof_line == ''):
							test = hallfile.read(1)
							if (test == ''):
								incompatible = 1
								break
							hallfile.seek(-1,1)
						elif (hof_line[0] == '|'):
							break
						elif (index == param_num):
							incompatible = 1
						hof_line_pieces = hof_line.split("|")
						field_name = hof_line_pieces[0]
						try:
							field_type = int(hof_line_pieces[1])
						except IndexError:
							hallfile.close()
							raise ParsingException, 'Syntax error on line ' + str(index+1) + ' of "' + hall_filename + '"\n'
						field_custom = hof_line_pieces[2:]
						if (index > len(hof_fields)):
							incompatible = 1
							break
						if (hof_fields[index] != [field_name, field_type, field_custom]):
							incompatible = 1
							break
						index += 1
					if (index != param_num):
						incompatible = 1
					if (incompatible):
						# code for renaming old HoF before it gets erased
						hallfile.close()
						writenew = 2
				except OSError:
					writenew = 1
				if (writenew):
					# if writenew == 1, we're writing a new file entirely.  If 2, we have to rename the
					# old one to preserve it.

					if (writenew == 2):
						int_extension = 1
						try:
							hallfile_pieces = os.path.splitext(hall_filename)
							while 1:
								hallfile = open(hallfile_pieces[0] + str(int_extension) + hallfile_pieces[1], 'rU')
								hallfile.close()
								int_extension += 1
						except IOError:
							""" int_extension now has the lowest positive N for which
								'<HoF_name>N.<HoF_extension>' wasn't already taken. """
							pass
						new_hallfile = hallfile_pieces[0] + str(int_extension) + hallfile_pieces[1]
						# here we also change the HoF filename and save the config.agc file
						os.rename(hall_filename, new_hallfile)
						WriteText('Existing Hall of Fame\'s fields are incompatible with the current data of "game.txt" -- the old Hall of Fame will be renamed as "' + new_hallfile + '"\n')
						

					hallfile = open(hall_filename, 'w')
					for field in hof_fields:
						hallfile.write(field[0])
						hallfile.write('|' + str(field[1]))
						for custom_entry in field[2]:
							hallfile.write('|' + custom_entry)
						hallfile.write('\n')
					hallfile.write('|\n')
				else:
					hallfile.close()
					hallfile = open(hall_filename, 'a')
				# here goes code for when it's ready to print new HoF entries (positioned after last field line and separating "|")
				for index in range (0, param_num-1):
					hallfile.write(par_eval[index] + '\n')

				# replace with real date/time:
				hallfile.write(time.strftime("%a, %b %d, %Y, at %H:%M:%S", time.localtime(time.time())) + '\n')
				hallfile.write('|\n')
				hallfile.close()
				UpdateHoFData()
		if (stats_compute):
			com_count[ch] += 1
		if ((cur_state.wl_state == 0) or (cur_state.wl_state == 3)):
			if (ch == 'w'):
				cur_state.wl_state = 2
			else:
				cur_state.wl_state = 3
		return 0
	elif (ch == '@'):
		if (not len(strip_line)):
			raise ParsingException, 'No arguments supplied to "@" command.\nLine: "' + line + '"\n'
		if ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
			raise ParsingException, 'Space or tab expected after start of "@" command.\nLine: "' + line + '"\n'
		cons_pieces = line.split()
		if (len(cons_pieces) < 3):
			raise ParsingException, 'Not enough arguments supplied to "@" command.\nLine: "' + line + '"\n'
		if ((cons_pieces[1] == "+") or (cons_pieces[1] == "-")):
			if ((parse_loc == 'copts') and (next_depth == 0)):
				raise ParsingException, 'Only constant option definitions are allowed in "copts.sma".\nLine: "' + line + '"\n'
			if (len(cons_pieces) != 3):
				raise ParsingException, 'One tag needed after "' + cons_pieces[1] + '" in "@" command.\nLine: "' + line + '"\n'
			cons_tag = cons_pieces[2]
			found_opt = 0
			if (stats_compute):
				if (cons_pieces[1] == "+"):
					com_count['@+'] += 1
				else:
					com_count['@-'] += 1
				return 0
			for option in cur_state.const_options:
				if (option.tag == cons_tag):
					found_opt = 1
					if ((cons_pieces[1] == "+") and (option.visible == 0)):
						option.visible = 1
						cur_state.visible_opt_count += 1
					elif ((cons_pieces[1] == "-") and (option.visible == 1)):
						option.visible = 0
						cur_state.visible_opt_count -= 1
					break
			if (not found_opt):
				raise ParsingException, 'Tag of "@" command not recognized as the tag of a constant option.\nLine: "' + line + '"\n'
		else:
			cons_tag = cons_pieces[1]
			expression = " ".join(cons_pieces[2:])
			cons_text = eval_string(expression)
			if (stats_compute):
				if (parse_loc == 'copts'):
					com_count['@'] += 1
					parse_context = 1
					parseblock(f, 'copts', next_depth+1)
					parse_context = 0
				else:
					com_count['@r'] += 1
				return 0
			rename = 0
			for option in cur_state.const_options:
				if (option.tag == cons_tag):
					if ((parse_loc == 'copts') and (next_depth == 0)):
						raise ParsingException, 'Only constant option definitions are allowed in "copts.sma".\nLine: "' + line + '"\n'
					rename = 1
					option.text = cons_text
					break
			if (not rename):
				if (parse_loc == 'copts'):
					cur_state.visible_opt_count += 1
					cur_state.const_options.append(Option('@', 'copts', pos, cons_tag, cons_text))
					return 1
				else:
					raise ParsingException, 'Illegal definition of new constant option outside "copts.sma", or tag of "@" command not recognized as previously registered constant option.\nLine: "' + line + '"\n'
		return 0
	elif (ch == '~' and not edit):
		if (next_depth):
			error_text = 'Command "~" in the wrong context.' + context_error_text(3) + '\nLine: "' + line + '"\n'
			WriteText('parse_context = ' + str(parse_context) + ' and S2 = ' + str(S2))
			raise ParsingException, error_text
		if (not len(strip_line)):
			raise ParsingException, 'No tag supplied to "~" command.\nLine: "' + line + '"\n'
		if ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
			raise ParsingException, 'Space or tab expected after start of "~" command.\nLine: "' + line + '"\n'
		if (not tag_regex.match(strip_line)):
			raise ParsingException, 'Illegal character in tag of "~" command.\nLine: "' + line + '"\n'
		if (stats_compute):
			com_count['~'] += 1
			S2 += 1		# this works to not flag "r" commands as errors in ~ blocks, but yes, outside
			parseblock(f, parse_loc, next_depth+1)
			S2 -= 1
			return 0
		return 1
	elif (ch == 'S'):
		if (not len(strip_line)):
			raise ParsingException, 'No arguments supplied to "S" command.\nLine: "' + line + '"\n'
		if ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
			raise ParsingException, 'Space or tab expected after start of "S" command.\nLine: "' + line + '"\n'
		status_pieces = line.split()
		if (len(status_pieces) < 2):
			raise ParsingException, 'No arguments supplied to "S" command.\nLine: "' + line + '"\n'
		if (not status_pieces[1].isdigit()):
			raise ParsingException, 'Non-numeric character in status line number of "S" command.\nLine: "' + line + '"\n'
		if (stats_compute):
			com_count['S'] += 1
			return 0
		status_num = int(status_pieces[1])
		if (len(status_pieces) == 2):	# remove the status line
			if (cur_state.status_lines.has_key(status_num)):
				del cur_state.status_lines[status_num]
				cur_state.status_list.remove(status_num)
		else:
			status_text = " ".join(status_pieces[2:])
			expression = status_text
			status_text = eval_string(expression)
			if (not cur_state.status_lines.has_key(status_num)):
				insort(cur_state.status_list, status_num)
			cur_state.status_lines[status_num] = status_text
		return 0
	elif (ch == 'D'):
		if (len(line.strip()) > 1):
			raise ParsingException, 'Extra character(s) following "D" command.\nLine: "' + line + '"\n'
		cur_state.enum_inv = {}
		cur_state.reg_inv = {}
		if (stats_compute):
			com_count['D'] += 1
			return 0
		else:
			inv_change = 1
		return 0
	elif (ch == 'z'):
		if (len(line.strip()) > 1):
			raise ParsingException, 'Extra character(s) following "z" command.\nLine: "' + line + '"\n'
		cur_state.svars = {}
		if (stats_compute):
			com_count['z'] += 1
			return 0
		return 0
	elif (ch == 'Z'):
		if (len(line.strip()) > 1):
			raise ParsingException, 'Extra character(s) following "Z" command.\nLine: "' + line + '"\n'
		cur_state.up_timers = {}
		cur_state.down_timers = {}
		cur_state.gvars = {}
		if (stats_compute):
			com_count['Z'] += 1
			return 0
		return 0
	elif ((ch == 'E') and not edit):
		if not len(strip_line):
			raise ParsingException, 'No string-expression supplied to "E" command.\nLine: "' + line + '"\n'
		if ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
			raise ParsingException, 'Space, tab, or end of line expected after start of "E" command.\nLine: "' + line + '"\n'
		expression = strip_line
		e_evaluated = eval_string(expression)
		if (not stats_compute):
			raise ParsingException, '"E" command encountered.  Literal exception text: "' + strip_line + '"\nEvaluated exception text: "' + e_evaluated + '"\n'
	elif ((ch.upper() == 'M') and edit):
		if (not len(strip_line)):
			raise ParsingException, 'No arguments supplied to "M" command.\nLine: "' + line + '"\n'
		if ((raw_line[0] != ' ') and (raw_line[0] != '\t')):
			raise ParsingException, 'Space or tab expected after start of "M" command.\nLine: "' + line + '"\n'
		expression = line
		m_evaluated = eval_string(expression)
		m_pieces = m_evaluated.split()
		if (len(m_pieces) != 3):
			raise ParsingException, 'Exactly two arguments must be supplied following "' + ch + '" command.\nLine: "' + line + '"\n'
		if ((m_pieces[1] != 'v') and (m_pieces[1] != 's') and (m_pieces[1] != 'a') and (m_pieces[1] != 'e') and (m_pieces[1] != 't')):
			raise ParsingException, 'First argument must be "v", "s", "a", "e", or "t" in "' + ch + '" command.\nLine: "' + line + '"\n'
		if (m_pieces[1] == 's'):
			s_pieces = m_pieces[2].split(".")
			if (len(s_pieces) != 2):
				raise ParsingException, 'To mask/unmask a state variable, the syntax must be in the form "m s location.variable" or "M s location.variable".\nLine: "' + line + '"\n'
			if ((not len(s_pieces[0])) or (not len(s_pieces[0]))):
				raise ParsingException, 'To mask/unmask a state variable, the syntax must be in the form "m s location.variable" or "M s location.variable".\nLine: "' + line + '"\n'
			if (ch == 'm'):
				if (not svarmasked.has_key(s_pieces[0])):
					svarmasked[s_pieces[0]] = {}
				svarmasked[s_pieces[0]][s_pieces[1]] = 1
			else:
				if (svarmasked.has_key(s_pieces[0])):
					if (svarmasked[s_pieces[0]].has_key(s_pieces[1])):
						del svarmasked[s_pieces[0]][s_pieces[1]]
						if (svarmasked[s_pieces[0]] == {}):
							del svarmasked[s_pieces[0]]
		else:
			tag = m_pieces[2]
			if (not tag_regex.match(tag)):
				raise ParsingException, 'Illegal character in tag of "' + ch + '" command.\nLine: "' + line + '"\n'
			if (m_pieces[1] == 'v'):
				if (ch == 'm'):
					gvarmasked[tag] = 1
				elif (gvarmasked.has_key(tag)):
					del gvarmasked[tag]
			elif (m_pieces[1] == 'a'):
				if (ch == 'm'):
					robjmasked[tag] = 1
				elif (robjmasked.has_key(tag)):
					del robjmasked[tag]
			elif (m_pieces[1] == 'e'):
				if (ch == 'm'):
					eobjmasked[tag] = 1
				elif (eobjmasked.has_key(tag)):
					del eobjmasked[tag]
			else:
				if (ch == 'm'):
					timermasked[tag] = 1
				elif (timermasked.has_key(tag)):
					del timermasked[tag]
	elif (ch.upper() == 'M'):
		raise ParsingException, 'Interpreter masking commands are not allowed in Smash code.\nLine: "' + line + '"\n'
	elif (edit):
		if ((ch.upper() == 'C') or (ch.upper() == 'L') or (ch == '^') or (ch == '*') or (ch.upper() == 'W') or (ch == '#') or (ch == '~') or (ch == 'E')):
			raise ParsingException, 'Edit error: this Smash command is not permitted with editing.\nLine: "' + line + '"\n'
		return 0
	else:
		raise ParsingException, 'Unknown Smash command.\nLine: "' + line + '"\n'
		
		
#-------------------EXPRESSION PARSING FUNCTIONS----------------------

""" Here is how the expression parsing functions call on each other:
	(A -> B means that A might call B after doing its share of processing)

	Three functions return strings, and the first can break into the numerical expression chain:
		eval_string -> eval_string, eval_exp
		eval_string_brackets -> eval_exp
		eval_string_imagetags -> eval_string

	The remaining, returning integers, are in one long chain, determined by the operator precedence chain:
		eval_exp -> eval_or_exp -> eval_and_exp -> eval_comparison_exp
			-> eval_bitor_exp -> eval_bitxor_exp -> eval_bitand_exp
			-> eval_shift_exp -> eval_concat_exp -> eval_additive_exp
			-> eval_modulo_exp -> eval_divide_exp -> eval_multiply_exp
			-> eval_power_exp -> eval_unary_exp -> eval_term """


def eval_string(line, c_done=0):
	""" Evaluates the insides of {...}, {'...}, {"...},
		{$...}, and {#...} expressions, passing the
		entire remaining string down the chain to the
		next evaluation function. """

	global expression

	if (line.find("{") == -1):
		return line

	i = 0
	while (i < len(line) - 1):
		if (line[i] == "{"):
			if (i and (line[i-1] == "\\")):
				line = line[:i-1] + line[i:]
				i -= 1
			else:
				exp_end = 0
				nest_count = 0
				saw_inner_brace = 0
				for j in range (i+1, len(line)):
					if (line[j] == "{"):
						nest_count += 1
						saw_inner_brace = 1
					elif (line[j] == "}"):
						nest_count -= 1
						if (nest_count == -1):
							exp_end = j
							break
				if (exp_end):
					if (not c_done):
						subline = line[i+1:j]
						subline = eval_c(subline)
						line = line[:i+1] + subline + line[j:]
						j += len(subline) - j + i + 1		# put j back at the outer } index
					inside_char = line[i+1]
					skip_char = (inside_char in ["'", '"', ',', '$', '#'])
					if (saw_inner_brace):
						subline = eval_string(line[i+1+skip_char:j], 1)
						expression = subline
					else:
						expression = line[i+1+skip_char:j]
					if (not skip_char):
						subline = str(eval_exp(expression, 1))
					elif (inside_char == "'"):
						expression = eval_exp(expression, 1)
						subline = decode(expression)
					elif (inside_char == '"'):
						subline = str(encode(expression))
					else:
						expression = eval_exp(expression, 1)
						subline = string_format(str(expression), inside_char)
					line = line[:i] + subline + line[j+1:]
					i += len(str(subline)) - 1
		i += 1
	return line


def eval_string_brackets(line):
	""" Evaluates the insides of [...], ['...], ["...],
		[$...], and [#...] expressions, returning
		the remaining string. """

	global expression

	if (line.find("[") == -1):
		return line
	
	i = 0
	while (i < len(line) - 1):
		if (line[i] == "["):
			if (i and (line[i-1] == "\\")):
				line = line[:i-1] + line[i:]
				i -= 1
			else:
				exp_end = 0
				nest_count = 0
				saw_inner_brace = 0
				for j in range (i+1, len(line)):
					if (line[j] == "["):
						nest_count += 1
						saw_inner_brace = 1
					elif (line[j] == "]"):
						nest_count -= 1
						if (nest_count == -1):
							exp_end = j
							break
				if (exp_end):
					subline = line[i+1:j]
					subline = eval_c(subline)
					line = line[:i+1] + subline + line[j:]
					j += len(subline) - j + i + 1		# put j back at the outer ] index

					inside_char = line[i+1]
					skip_char = (inside_char in ["'", '"', ',', '$', '#'])
					if (saw_inner_brace):
						subline = eval_string(line[i+1+skip_char:j], 1)
						expression = subline
					else:
						expression = line[i+1+skip_char:j]
					if (not skip_char):
						subline = str(eval_exp(expression, 1))
					elif (inside_char == "'"):
						expression = eval_exp(expression, 1)
						subline = decode(expression)
					elif (inside_char == '"'):
						subline = str(encode(expression))
					else:
						expression = eval_exp(expression, 1)
						subline = string_format(str(expression), inside_char)
					line = line[:i] + subline + line[j+1:]
					i += len(str(subline)) - 1
		i += 1
	return line

	
def eval_string_imagetags(line):
	""" Evaluates the insides of [ ] -- if a valid image file
		name is found, it replaces that part, including brackets,
		with the appropriate HTML <img src> tag.  The remaining
		string is passed on to eval_string. """

	if (line.find("[") == -1):
		return eval_string(line)
	
	i = 0
	while (i < len(line) - 1):
		if (line[i] == "["):
			if (i and (line[i-1] == "\\")):
				line = line[:i-1] + line[i:]
				i -= 1
			else:
				exp_end = 0
				nest_count = 0
				saw_inner_brace = 0
				for j in range (i+1, len(line)):
					if (line[j] == "["):
						nest_count += 1
						saw_inner_brace = 1
					elif (line[j] == "]"):
						nest_count -= 1
						if (nest_count == -1):
							exp_end = j
							break
				if (exp_end):
					if (saw_inner_brace):
						i = exp_end
					else:
						try:
							graphic_name = eval_string(line[i+1:j].strip())
							ftest = open(os.path.join(cur_state.directory, graphic_name),'rU')
							graphic_part = "<img src='" + os.path.join(cur_state.directory, graphic_name) + "' border='0'>"
							ftest.close()
							line = line[:i] + graphic_part + line[j+1:]
							i += len(graphic_part) - 1
						except IOError:
							# following isn't always appropriate, because might not
							# be actually trying to reference image file in []
							# WriteText('No "' + graphic_name + '" image file in directory.')
							i = j
		i += 1
	return eval_string(line)

	
def eval_exp(line, c_done=0):

	if (not c_done):
		line = eval_c(line)

	if (line.find("(") == -1):
		return eval_or_exp(line)
		
	i = 0
	while (i < len(line) - 1):
		if (line[i] == "("):
			exp_end = 0
			nest_count = 0
			saw_open_paren = 0
			for j in range (i+1, len(line)):
				if (line[j] == "("):
					nest_count += 1
					saw_open_paren = 1
				elif (line[j] == ")"):
					nest_count -= 1
					if (nest_count == -1):
						exp_end = j
						break
			if (exp_end):
				if (saw_open_paren):
					subline = eval_exp(line[i+1:j], 1)
				else:
					subline = eval_or_exp(line[i+1:j])
				line = line[:i] + str(subline) + line[j+1:]
				i += len(str(subline)) - 1
		i += 1
	return eval_or_exp(line)
	
def eval_c(line):
	""" Evaluates all the C: operators in the string, going from right
		to left, and returns the processed string. """
	
	for i in range(len(line)-2, -1, -1):
		if ((line[i] == 'C') and (line[i+1] == ':')):
			try:
				char = line[i+2]
			except IndexError:
				char = ' '
			new_string = str(ord(char))
			line = line[:i] + new_string + line[i+3:]
	return line
	
def eval_or_exp(line):

	if (line.find("|") == -1):
		return eval_and_exp(line)
		
	or_exps = []
	begin_index = 0
	for i in range (1, len(line) - 1):
		if (line[i] == "|"):
			if (line[i-1] != "."):
				or_exps.append(line[begin_index:i])
				begin_index = i + 1
	if (not begin_index):
		return eval_and_exp(line)
	or_exps.append(line[begin_index:])
	for exp in or_exps:
		if (eval_and_exp(exp)):
			return 1
	return 0


def eval_and_exp(line):

	if (line.find(",") == -1):
		return eval_comparison_exp(line)

	and_exps = line.split(",")
	for exp in and_exps:
		if (not eval_comparison_exp(exp)):
			return 0
	return 1
	
	
def eval_comparison_exp(line):

	comparison_exps = []
	operators = []
	begin_index = 0
	i = 1
	while (i < len(line) - 1):
		one_is_op = 0
		two_is_op = 0
		two_chars = line[i:i+2]
		if ((two_chars == "!=") or (two_chars == "<=") or (two_chars == ">=")):
			two_is_op = 1
		elif ((line[i] == '=') or (line[i] == '!') or (line[i] == '<') or (line[i] == '>')):
			if ((two_chars == "<<") or (two_chars == ">>")):
				i += 1
			else:
				one_is_op = 1
		if (one_is_op or two_is_op):
			left = line[begin_index:i].strip()
			if ((len(left)) and TagChars.has_key(left[len(left)-1])):
				comparison_exps.append(left)
				if (two_is_op):
					operators.append(two_chars)
					begin_index = i + 2
					i += 1
				else:
					operators.append(line[i])
					begin_index = i + 1
		i += 1
	comparison_exps.append(line[begin_index:])
	result_so_far = eval_bitor_exp(comparison_exps[0])
	for i in range (0, len(operators)):
		if (operators[i] == "="):
			result_so_far = (result_so_far == eval_bitor_exp(comparison_exps[i+1]))
		elif ((operators[i] == "!") or (operators[i] == "!=")):
			result_so_far = (result_so_far != eval_bitor_exp(comparison_exps[i+1]))
		elif (operators[i] == "<"):
			result_so_far = (result_so_far < eval_bitor_exp(comparison_exps[i+1]))
		elif (operators[i] == ">"):
			result_so_far = (result_so_far > eval_bitor_exp(comparison_exps[i+1]))
		elif (operators[i] == "<="):
			result_so_far = (result_so_far <= eval_bitor_exp(comparison_exps[i+1]))
		else:
			result_so_far = (result_so_far >= eval_bitor_exp(comparison_exps[i+1]))
	return int(result_so_far)
	
	
def eval_bitor_exp(line):

	if (line.find(".|") == -1):
		return eval_bitxor_exp(line)
		
	bitor_exps = line.split(".|")	
	result_so_far = eval_bitxor_exp(bitor_exps[0])
	for i in range (1, len(bitor_exps)):
		result_so_far |= eval_bitxor_exp(bitor_exps[i])
	return result_so_far
	
	
def eval_bitxor_exp(line):
	
	if (line.find(".^") == -1):
		return eval_bitand_exp(line)
	
	bitxor_exps = line.split(".^")
	result_so_far = eval_bitand_exp(bitxor_exps[0])
	for i in range (1, len(bitxor_exps)):
		result_so_far ^= eval_bitxor_exp(bitxor_exps[i])
	return result_so_far
	
	
def eval_bitand_exp(line):	
		
	if (line.find(".&") == -1):
		return eval_shift_exp(line)

	bitand_exps = line.split(".&")
	result_so_far = eval_shift_exp(bitand_exps[0])
	for i in range (1, len(bitand_exps)):
		result_so_far &= eval_shift_exp(bitand_exps[i])
	return result_so_far
	

def eval_shift_exp(line):
	
	if ((line.find("<<") == -1) and (line.find(">>") == -1)):
		return eval_concat_exp(line)

	shift_exps = []
	operators = []
	begin_index = 0
	i = 1
	while (i < len(line) - 2):
		op = line[i:i+2]
		if ((op == "<<") or (op == ">>")):
			shift_exps.append(line[begin_index:i])
			operators.append(op)
			begin_index = i + 2
			i += 1
		i += 1
	shift_exps.append(line[begin_index:])
	result_so_far = eval_concat_exp(shift_exps[0])
	for i in range (0, len(operators)):
		shift_amount = eval_concat_exp(shift_exps[i+1]) % 64
		if (operators[i] == "<<"):
			result_so_far = (long(result_so_far) << shift_amount)
		else:
			if (result_so_far < 0):
				result_so_far += 18446744073709551616
			result_so_far = (result_so_far >> shift_amount)
	return rangefix(result_so_far)
	
def eval_concat_exp(line):
	
	if (line.find("++") == -1):
		return eval_additive_exp(line)

	concat_exps = []
	operators = []
	begin_index = 0
	i = 1
	while (i < len(line) - 2):
		op = line[i:i+2]
		if (op == "++"):
			concat_exps.append(line[begin_index:i])
			operators.append(op)
			begin_index = i + 2
			i += 1
		i += 1
	concat_exps.append(line[begin_index:])
	result_so_far = eval_additive_exp(concat_exps[0])
	for i in range (0, len(operators)):
		rhs = eval_additive_exp(concat_exps[i+1])
		result_so_far = concat(result_so_far, rhs)
	return result_so_far

def eval_additive_exp(line):

	additive_exps = []
	operators = []
	begin_index = 0
	for i in range (1, len(line) - 1):
		if (((line[i] == '+') or (line[i] == "-")) and (line[i+1] != "/") and (line[i+1] != "%")):
			left = line[begin_index:i].strip()
			if ((len(left)) and TagChars.has_key(left[len(left)-1])):
				additive_exps.append(left)
				operators.append(line[i])
				begin_index = i + 1
	additive_exps.append(line[begin_index:])
	result_so_far = eval_modulo_exp(additive_exps[0])
	for i in range (0, len(operators)):
		if (operators[i] == "+"):
			result_so_far += eval_modulo_exp(additive_exps[i+1])
		else:
			result_so_far -= eval_modulo_exp(additive_exps[i+1])
	return rangefix(result_so_far)


def eval_modulo_exp(line):
	global expression
	global stats_compute

	if (line.find("%") == -1):
		return eval_divide_exp(line)
	
	mod_exps = []
	operators = []
	begin_index = 0
	i = 1
	while (i < len(line) - 1):
		one_is_op = 0
		two_is_op = 0
		if (line[i] == "%"):
			one_is_op = 1
		elif ((line[i] == "+") or (line[i] == "-")):
			if (line[i+1] == "%"):
				two_is_op = 1
		if (one_is_op or two_is_op):
			mod_exps.append(line[begin_index:i])
			if (two_is_op):
				operators.append(line[i:i+2])
				begin_index = i + 2
				i += 1
			else:
				operators.append(line[i])
				begin_index = i + 1
		i += 1
	mod_exps.append(line[begin_index:])
	result_so_far = eval_divide_exp(mod_exps[0])
	for i in range (0, len(operators)):
		rhs = eval_divide_exp(mod_exps[i+1])
		if (not rhs):
			if (stats_compute):
				return 0
			else:
				raise ParsingException, 'Expression error: modulo by 0.\nExpression: "' + expression + '"\n'
		intdiv = result_so_far / rhs
		modulus = result_so_far % rhs
		if (modulus):
			if (result_so_far < 0):
				if (intdiv < 0):
					if ((operators[i] == "+%") or (operators[i] == "%")):
						modulus -= rhs
				elif (operators[i] == "+%"):
					modulus -= rhs
			else:
				if (intdiv < 0):
					if ((operators[i] == "%") or (operators[i] == "+%")):
						modulus -= rhs
				elif (operators[i] == "+%"):
					modulus -= rhs
		result_so_far = modulus
	return result_so_far


def eval_divide_exp(line):
	global expression
	global stats_compute

	if (line.find("/") == -1):
		return eval_multiply_exp(line)

	divide_exps = []
	operators = []
	begin_index = 0
	i = 1
	while (i < len(line) - 1):
		one_is_op = 0
		two_is_op = 0
		if (line[i] == "/"):
			one_is_op = 1
		elif ((line[i] == "+") or (line[i] == "-")):
			if (line[i+1] == "/"):
				two_is_op = 1
		if (one_is_op or two_is_op):
			divide_exps.append(line[begin_index:i])
			if (two_is_op):
				operators.append(line[i:i+2])
				begin_index = i + 2
				i += 1
			else:
				operators.append(line[i])
				begin_index = i + 1
		i += 1
	divide_exps.append(line[begin_index:])
	result_so_far = eval_multiply_exp(divide_exps[0])
	for i in range (0, len(operators)):
		rhs = eval_multiply_exp(divide_exps[i+1])
		if (not rhs):
			if (stats_compute):
				return 0
			else:
				raise ParsingException, 'Expression error: division by 0.\nExpression: "' + expression + '"\n'
		intdiv = result_so_far / rhs
		if (result_so_far % rhs):
			if (intdiv < 0):
				if (operators[i] == "/" or operators[i] == "+/"):
					intdiv += 1
			else:
				if (operators[i] == "+/"):
					intdiv += 1
		result_so_far = intdiv
	return result_so_far


def eval_multiply_exp(line):
	
	if (line.find("*") == -1):
		return eval_unary_exp(line)

	multiply_exps = []
	begin_index = 0
	i = 1
	while (i < len(line) - 1):
		if (line[i] == "*"):
			if (line[i+1] == "*"):
				i += 1
			else:
				multiply_exps.append(line[begin_index:i])
				begin_index = i + 1
		i += 1
	multiply_exps.append(line[begin_index:])
	result_so_far = eval_power_exp(multiply_exps[0])
	for i in range (1, len(multiply_exps)):
		result_so_far *= eval_power_exp(multiply_exps[i])
	return rangefix(result_so_far)


def eval_power_exp(line):
	
	if (line.find("**") == -1):
		return eval_unary_exp(line)

	power_exps = line.split("**")
	result_so_far = eval_unary_exp(power_exps[0])
	for i in range (1, len(power_exps)):
		result_so_far = result_so_far ** eval_unary_exp(power_exps[i])
	return rangefix(result_so_far)


def eval_unary_exp(line):
	global expression, stats_compute
	global S1, S2, S3, S4, S5

	i = len(line) - 1
	while (i >= 0):
		if ((not TagChars.has_key(line[i])) and (line[i] != ".") and (not line[i].isspace())):
			base_term = line[i+1:].strip()
			break
		i -= 1
	if (i < 0):
		return eval_term(line)
	last_was_string = 1
	

	""" If string-based unary operators (a:) and number-based operators (!) are
		mixed, we might be switching between strings and numbers as we
		evaluate.  (Well, the first one will operate on a string, and the rest
		on numbers.)  So remember which one we've got at any given time.
		Use base_term to keep track of the last string evaluation,
		and result_so_far to keep track of the last numerical evaluation.
		NOTE: base_term should always be in stripped form. """
	
	while (i >= 0):
		if (line[i] == "!"):
			if (last_was_string):
				result_so_far = (not eval_term(base_term))
				last_was_string = 0
			else:
				result_so_far = (not result_so_far)
		elif (line[i] == "-"):
			if (last_was_string):
				result_so_far = -eval_term(base_term)
				last_was_string = 0
			else:
				result_so_far = -result_so_far
		elif (line[i] == "~"):
			if (last_was_string):
				result_so_far = ~eval_term(base_term)
				last_was_string = 0
			else:
				result_so_far = ~result_so_far
		elif ((line[i] == ":") and i):
			if (line[i-1] == "a"):
				if (last_was_string):
					if (reg_names.has_key(base_term)):
						result_so_far = cur_state.reg_inv.has_key(base_term)
						last_was_string = 0
					else:
						raise ParsingException, 'Expression error: unknown inventory tag referenced with "a:".\nExpression: "' + expression + '"\n'
				else:
					term = str(result_so_far)
					if (reg_names.has_key(term)):
						result_so_far = cur_state.reg_inv.has_key(term)
					else:
						raise ParsingException, 'Expression error: unknown inventory tag referenced with "a:".\nExpression: "' + expression + '"\n'
			elif (line[i-1] == "d"):
				if (last_was_string):
					if (reg_names.has_key(base_term)):
						result_so_far = not cur_state.reg_inv.has_key(base_term)
					else:
						raise ParsingException, 'Expression error: unknown inventory tag referenced with "d:".\nExpression: "' + expression + '"\n'
				else:
					term = str(result_so_far)
					if (reg_names.has_key(term)):
						result_so_far = not cur_state.reg_inv.has_key(term)
					else:
						raise ParsingException, 'Expression error: unknown inventory tag referenced with "d:".\nExpression: "' + expression + '"\n'
			elif (line[i-1] == "e"):
				if (last_was_string):
					if (enum_names.has_key(base_term)):
						result_so_far = cur_state.enum_inv.get(base_term, 0)
						last_was_string = 0
					else:
						raise ParsingException, 'Expression error: unknown inventory tag referenced with "e:".\nExpression: "' + expression + '"\n'
				else:
					term = str(result_so_far)
					if (enum_names.has_key(term)):
						result_so_far = cur_state.enum_inv.get(term, 0)
					else:
						raise ParsingException, 'Expression error: unknown inventory tag referenced with "e:".\nExpression: "' + expression + '"\n'
			elif (line[i-1] == "r"):
				if (last_was_string):
					result_so_far = eval_term(base_term)
					last_was_string = 0
				if (result_so_far <= 0):
					if stats_compute:
						return 0
					else:
						raise ParsingException, 'Expression error: non-positive range specified for "r:".\nExpression: "' + expression + '"\n'
				result_so_far = genrand_int32() % result_so_far
			elif (line[i-1] == "c"):
				if (last_was_string):
					last_was_string = 0
					result_so_far = encode(base_term)
				else:
					result_so_far = encode(str(result_so_far))
			elif (line[i-1] == "f"):
				if (stats_compute):
					result_so_far = 0
				else:
					if (last_was_string):
						last_was_string = 0
						function_operator(base_term)
					else:
						function_operator(str(result_so_far))
					result_so_far = S1
			elif (line[i-1] == "A"):
				if (last_was_string):
					result_so_far = eval_term(base_term)
					last_was_string = 0
				if (result_so_far < 0):
					result_so_far = -result_so_far
			elif (line[i-1] == "p"):
				if (last_was_string):
					result_so_far = eval_term(base_term)
					last_was_string = 0
				if (result_so_far == 1):
					result_so_far = 0
				else:
					result_so_far = 1
			elif (line[i-1] == "P"):
				if (last_was_string):
					result_so_far = eval_term(base_term)
					last_was_string = 0
				if (result_so_far == 1):
					result_so_far = 0
				else:
					result_so_far = 115
			elif (line[i-1] == "S"):
				if (last_was_string):
					result_so_far = eval_term(base_term)
					last_was_string = 0
				if (result_so_far == 0):
					result_so_far = cur_state.moves
				elif (result_so_far == 1):
					result_so_far = S1
				elif (result_so_far == 2):
					result_so_far = S2
				elif (result_so_far == 3):
					result_so_far = S3
				elif (result_so_far == 4):
					result_so_far = S4
				elif (result_so_far == 5):
					result_so_far = S5
				else:
					raise ParsingException, 'Expression error: invalid argument to "S:".\nExpression: "' + expression + '"\n'
			else:
				raise ParsingException, 'Expression error: invalid character preceding ":".\nExpression: "' + expression + '"\n'
			i -= 1
		elif (not line[i].isspace()):
			if stats_compute:
				WriteText('Possible expression error: invalid character.\nExpression: "' + expression + '"\n')
				# replace the invalid character with a valid character and continue backtracking
				line = line[:i] + 'a' + line[i+1:]
				result_so_far = 0
				while (i >= 0):
					if ((not TagChars.has_key(line[i])) and (line[i] != ".") and (not line[i].isspace())):
						base_term = line[i+1:].strip()
						break
					i -= 1
			else:
				raise ParsingException, 'Expression error: invalid character.\nExpression: "' + expression + '"\n'
		i -= 1
	return rangefix(result_so_far)


def eval_term(line):
	global expression

	result = 0
	line = line.strip()
	if (line == ""):
		raise ParsingException, 'Expression error: empty term.\nExpression: "' + expression + '"\n'
	elif (line.isdigit()):
		result = int(line)
	else:
		if (line[0] == "."):
			if (len(line) > 3):
				if (line[1:3] == ".."):
					timer_term = line[3:]
					if (cur_state.up_timers.has_key(timer_term)):
						result = cur_state.up_timers[timer_term]
					else:
						result = cur_state.down_timers.get(timer_term, 0)
			if ((len(line) < 4) or (line[1:3] != "..")):
				result = cur_state.gvars.get(line[1:], 0)
		elif (line.find(".") > 0):
			state_pieces = line.split(".")
			if (len(state_pieces) != 2):
				raise ParsingException, 'Expression error: extra "." scoping operator.\nExpression: "' + expression + '"\n'
			result = cur_state.svars.get(state_pieces[1], {}).get(state_pieces[0], 0)
		elif (len(line.split()) > 1):
			# a space separating base terms, such as "var 47"
			raise ParsingException, 'Expression error: space separating basic terms.\nExpression: "' + expression + '"\n'
		else:
			try:
				result = cur_state.svars[line][cur_state.location]
			except KeyError:
				try:
					result = cur_state.up_timers[line]
				except KeyError:
					try:
						result = cur_state.down_timers[line]
					except KeyError:
						result = cur_state.gvars.get(line, 0)
	return rangefix(result)


def concat(lhs, rhs):
	
	if (lhs < 0):
		lhs += 18446744073709551616
	if (rhs < 0):
		rhs += 18446744073709551616
	temp = rhs
	shift = 0
	while (temp):
		temp >>= 8
		shift += 8
	if (shift == 0):
		shift = 8
	lhs <<= shift
	result = (lhs + rhs) % 18446744073709551616
	if (result > 9223372036854775807):
		result -= 18446744073709551616
	return result

def decode(int_value):
	if (int_value < 0):
		int_value += 18446744073709551616
	result_so_far = ''
	while (int_value > 0):
		low_int = int_value % 256
		last_char = ''
		if (low_int > 26):
			last_char = chr(low_int)
		elif (low_int > 0):
			last_char = chr(96 + low_int)
		result_so_far = last_char + result_so_far
		int_value >>= 8
	return result_so_far

def encode(str_value):
	str_value = str_value[-8:]
	result_so_far = 0
	str_len = len(str_value)
	for i in range(0, str_len):
		result_so_far += long(ord(str_value[i])) << (8 * (str_len - i - 1))
	if (result_so_far > 9223372036854775807):
		result_so_far - 18446744073709551616
	return result_so_far

def string_format(input_string, format_type):
	string_so_far = ''
	left_decimal = long(input_string)
	if (format_type == '$'):
		string_so_far = '$'
		right_decimal = left_decimal % 100
		left_string = str(left_decimal / 100)
	elif (format_type == '#'):
		string_so_far = chr(163)
		right_decimal = left_decimal % 100
		left_string = str(left_decimal / 100)
	else:
		left_string = str(left_decimal)
	left_len = len(left_string)
	for i in range (0, left_len):
		if i and (not ((left_len - i)%3)):
			string_so_far += ','
		string_so_far += left_string[i]
	if not (format_type == ','):
		string_so_far += '.'
		if (right_decimal < 10):
			string_so_far += '0'
		string_so_far += str(right_decimal)
	return string_so_far


def function_operator(funcline):
	global cur_parse_file
	global expression
	global S1, S2

	func_pieces = funcline.split(".")
	if (len(func_pieces) > 2):
		# too many . separators
		raise ParsingException, 'Expression error: too many "." separators in argument of "f" operator.\nExpression: "' + expression + '"\n'
	if (len(func_pieces) == 2):
		if (not len(func_pieces[0]) or not len(func_pieces[1])):
			raise ParsingException, 'Expression error: empty string to one side of "." separator in argument of "f" operator.\nExpression: "' + expression + '"\n'
		func_loc = func_pieces[0]
		func_name = func_pieces[1]
	else:
		func_loc = cur_parse_file
		func_name = func_pieces[0]
	if (not tag_regex3.match(func_loc)):
		raise ParsingException, 'Uppercase character in location tag "' + func_loc + '" of "f" operator.\n'
	try:
		f4 = open(os.path.join(cur_state.directory, func_loc + '.sma'), 'rU')
		f4.seek(0)
		find_line = f4.readline()
		if (find_line == ''):
			f4.close()
			raise ParsingException, 'Function "' + func_name + '" not found in file "' + func_loc + '".\nLocation found, but no function with this tag located.\n'
		find_line = find_line.strip()
		while (find_line != ("~ " + func_name)):
			find_line = f4.readline()
			if (find_line == ''):
				f4.close()
				raise ParsingException, 'Expression error: function "' + func_name + '" not found in file "' + func_loc + '".\nLocation found, but no function with this tag located.\n'
			find_line = find_line.strip()
		S2 += 1
		S1 = 0
		temp_loc = cur_parse_file
		cur_parse_file = func_loc
		parseblock(f4, func_loc, 1)
		cur_parse_file = temp_loc
		f4.close()
		S2 -= 1
	except IOError:
		raise ParsingException, 'Expression error: function "' + func_name + '" called, but source file "' + func_loc + '" not found.\nExpression: "' + expression + '"\n'

def rangefix(x):
	return ((x + 9223372036854775808) % 18446744073709551616) - 9223372036854775808

def init_genrand(tag):
	global mt, mti
	N = 624
	
	mti = N + 1
	xor_total = 0
	for char in tag:
		xor_total ^= ord(char)

	cur_state.random_seed = cur_state.random_seed ^ cur_state.moves ^ (xor_total << ((cur_state.moves * 11) % 24))
	mt[0] = cur_state.random_seed
	for i in range (1, N):
		mt[i] = (1812433253 * (mt[i-1] ^ (mt[i-1] >> 30)) + i)
		mt[i] &= 4294967295
	
def genrand_int32():
	# the seed is the "checksum" kept track of in the saved state
	global mt
	global mti

	mag01 = [0, 2567483615]
	UPPER_MASK  = 2147483648
	LOWER_MASK  = 2147483647
	N = 624
	M = 397
	
	if (mti >= N):
		for kk in range (0, N-M):
			y = (mt[kk] & UPPER_MASK) | (mt[kk+1] & LOWER_MASK)
			mt[kk] = mt[kk+M] ^ (y >> 1) ^ mag01[y & 1]
		for kk in range (N-M, N-1):
			y = (mt[kk] & UPPER_MASK) | (mt[kk+1] & LOWER_MASK)
			mt[kk] = mt[kk+(M-N)] ^ (y >> 1) ^ mag01[y & 1]
		y = (mt[N-1] & UPPER_MASK) | (mt[0] & LOWER_MASK)
		mt[N-1] = mt[M-1] ^ (y >> 1) ^ mag01[y & 1]
		
		mti = 0
		
	y = mt[mti]
	mti += 1

	y ^= (y >> 11)
	y ^= (y << 7) & 2636928640
	y ^= (y << 15) & 4022730752
	y ^= (y >> 18)

	return y


#------------------------LOG CLASS-----------------------------

class MyLog(wx.PyLog):
	""" Used for logging text to the bottom right panel.  Code for time stamps is there,
		but with logTime=0 it won't be used. """

	def __init__(self, textCtrl, logTime=0):
		wx.PyLog.__init__(self)
		self.tc = textCtrl
		self.logTime = logTime

	def DoLogString(self, message, timeStamp):
		if self.logTime:
			message = time.strftime("%X", time.localtime(timeStamp)) + \
                      ": " + message
		if self.tc:
			self.tc.AppendText(message + '\n')
			
def NarText(text):
	""" Used for narrating the action text on each move to the narration file of choice. """

	global narrate
	global narrate_filename
	
	narf = open(os.path.join(cur_state.directory, narrate_filename), 'a')
	narf.write(text)
	narf.write('\n')
	narf.close()
		
def WriteText(text):
	""" "text" is a string passed in to be logged to the bottom right panel
		or the external file of choice, or both, or neither, depending
		on the settings chosen by the user. """

	global panellog, textlog
	global log_filename
	
	if panellog:
		wx.LogMessage(text)
	if textlog:
		logf = open(os.path.join(cur_state.directory, log_filename), 'a')
		logf.write(text)
		logf.write('\n')
		logf.close()





#------------------------GUI CLASSES---------------------------
class HoFWindow(wx.Frame):

	# initialize class variables
	
	def __init__(self, parent):
		global game_title, main_directory

		wx.Frame.__init__(self, parent, -1, game_title + ' Hall of Fame', size=(800, 600))
		# the icon for Rex -- hopefully present
		self.SetIcon(wx.Icon(os.path.join(main_directory, 'rex.ico'), wxBITMAP_TYPE_ICO))
		self.hall_sort_column = -1
		self.hall_entries = []
		self.hall_fields = []
		EVT_CLOSE(self, self.OnCloseWindow)
		self.hofwindow = wx.html.HtmlWindow(self, -1, style=wx.CLIP_CHILDREN|wx.VSCROLL|wx.HSCROLL)
		self.hofwindow.SetBackgroundColour('WHITE')
		self.RefreshData(0)
		self.Show(True)

	def OnCloseWindow(self, event):
		self.Destroy()

	def RefreshData(self, event):
		""" This should actually re-fetch *everything* including field names from the hall_filename file """

		global game_title, hall_filename

		try:
			# First we open the Hall of Fame file and read in the field data:

			self.hall_fields = []
			self.hall_entries = []
			self.hall_sort_column = -1
			hallfile = open(hall_filename, 'rU')
			line = ''
			while 1:
				hall_line = hallfile.readline().strip()
				if (hall_line == ""):
					test = hallfile.read(1)
					if (test == ''):
						raise "No |"
					else:
						hallfile.seek(-1,1)
						continue
				elif (hall_line[0] == '|'):
					break
				hall_line_pieces = hall_line.split("|")
				field_name = hall_line_pieces[0]
				field_type = int(hall_line_pieces[1])
				field_custom = hall_line_pieces[2:]
				self.hall_fields.append([field_name, field_type, field_custom])
			self.hall_fields.append(["Order", 1, []])	# A "hidden" field used to sort by "Date" easier

			# Then we read in the Hall of Fame entries:

			try:
				while 1:
					hall_entry = []
					while 1:
						hall_line = hallfile.readline().strip()
						if (hall_line == "|"):
							break
						if (hall_line == ""):
							test = hallfile.read(1)
							if (test == ''):
								raise "EOF"
							else:
								hallfile.seek(-1,1)
						hall_entry.append(hall_line)
					if (len(hall_entry) == len(self.hall_fields)-1):
						hall_entry.append(str(len(self.hall_entries)))
						self.hall_entries.append(hall_entry)
			except "EOF":
				pass
			hallfile.close()

			# Finally, a call to the method to display the table.

			self.DrawTable()
		except IOError:
			halltext = '<wxp class="wxButton" width="90" height="25"><param name="label" value="Try Again"><param name = "id" value = 3050></wxp><p>The Hall of Fame file could not be located -- check that the file still exists given in the main window\'s Options tab, then refresh.'
			wx.EVT_BUTTON(self, 3050, self.RefreshData)
			self.hofwindow.SetPage(halltext)
		except:
			WriteText('Error in Hall of Fame data -- the Hall of Fame table could not be displayed properly.')
			halltext = '<wxp class="wxButton" width="90" height="25"><param name="label" value="Try Again"><param name = "id" value = 3050></wxp><p>There was an error in the Hall of Fame text file, so a table could not be generated properly -- try checking the file\'s structure for errors, then refresh.'
			wx.EVT_BUTTON(self, 3050, self.RefreshData)
			self.hofwindow.SetPage(halltext)

	def DrawTable(self):
		""" Handles the generation of the HTML Table showing the Hall of Fame entries. """

		max_column = len(self.hall_fields) - 1
		halltext = '<center><table border=1 width=' + str(100 * (len(self.hall_fields)+2)) + '><tr><td><wxp class="wxButton" width="90" height="25"><param name="label" value="Refresh"><param name = "id" value = 3050></wxp></td><td colspan = ' + str(len(self.hall_fields)+1) + '><font size=+1><center><b>' + game_title + ' Hall of Fame:</b></center></font></td></tr><tr><td><center><b>Rank No.</b></center></td>'
		wx.EVT_BUTTON(self, 3050, self.RefreshData)
		if (self.hall_sort_column == -1):
			self.hall_sort_column = max_column - 1
		field_index = 0
		for hall_field in self.hall_fields:
			if (field_index == max_column):
				break
			if (field_index == self.hall_sort_column):
				halltext += '<td'
				if (field_index == max_column-1):
					halltext += ' colspan=3'
				halltext += ' bgcolor="#ffffdd"><center><font size=+1><b>' + hall_field[0] + '</b></font></center></td>'
			else:
				halltext += '<td'
				if (field_index == max_column-1):
					halltext += ' colspan=3'
				halltext += '><wxp class="wxButton" width="'
				if (field_index == max_column-1):
					halltext += '290'
				else:
					halltext += '90'
				halltext += '" height="25"><param name="label" value="' + hall_field[0] + '"><param name = "id" value = ' + str(2800+field_index) + '></wxp></td>'
				wx.EVT_BUTTON(self, (2800+field_index), self.Resort)
			field_index += 1
		halltext += "</tr>"
		row = 1
		for hall_entry in self.hall_entries:
			halltext += "<tr><td><center>" + str(row) + "</center></td>"
			for index in range (0, max_column):
				halltext += '<td'
				if (index == max_column-1):
					halltext += ' colspan=3'
				if (index == self.hall_sort_column):
					halltext += ' bgcolor="#ffffdd"'
				halltext += '><center>' + str(hall_entry[index]) + '</center></td>'
			halltext += "</tr>"
			row += 1
		halltext += "</table></center>"
		self.hofwindow.SetPage(halltext)

	def Resort(self, event):
		""" Sorts the Hall of Fame entries according to the sorting scheme belonging
			to the column selected.  In the case of ties, the entry with the earlier date
			appears first (really, what it does is sort by the "hidden" column which is the
			"Entry Number" field -- entry #1 has value "1" there, entry #2 has value "2",
			etc.)  After the re-sort, the table is redrawn. """

		column = event.GetId() - 2800
		max_col = len(self.hall_fields)-1
		try:
			sort_method = self.hall_fields[column][1]
			if (column == (max_col-1)):	# If we're sorting by date, really sort by the hidden "Entry Number" field
				self.hall_entries.sort(lambda a, b: int(a[max_col]) - int(b[max_col]))
			else:
				blanks = filter(lambda a: not len(a[column]), self.hall_entries)
				others = filter(lambda a: len(a[column]), self.hall_entries)
				if (sort_method == 0):	# Numeric, Descending
					others.sort(lambda a, b: ((int(a[column]) < int(b[column])) or ((int(a[column]) == int(b[column])) and (int(a[max_col]) > int(b[max_col])))) * 2 - 1)
				elif (sort_method == 1):	# Numeric, Ascending
					others.sort(lambda a, b: ((int(a[column]) > int(b[column])) or ((int(a[column]) == int(b[column])) and (int(a[max_col]) > int(b[max_col])))) * 2 - 1)
				elif (sort_method == 2):	# Alphabetic, Descending
					others.sort(lambda a, b: ((a[column].upper() < b[column].upper()) or ((a[column].upper() == b[column].upper()) and (int(a[max_col]) > int(b[max_col])))) * 2 - 1)
				elif (sort_method == 3):	# Alphabetic, Ascending
					others.sort(lambda a, b: ((a[column].upper() > b[column].upper()) or ((a[column].upper() == b[column].upper()) and (int(a[max_col]) > int(b[max_col])))) * 2 - 1)
				elif (sort_method == 4):	# Custom Sorting Scheme
					others.sort(lambda a, b: ((self.hall_fields[column][2].index(a[column]) > self.hall_fields[column][2].index(b[column])) or ((self.hall_fields[column][2].index(a[column]) == self.hall_fields[column][2].index(b[column])) and (int(a[max_col]) > int(b[max_col])))) * 2 - 1)
				self.hall_entries = others + blanks
			self.hall_sort_column = column
			self.DrawTable()
		except:
			WriteText('Error in Hall of Fame data -- table could not be resorted.')

class SmashFrame(wx.Frame):

	InfoWindowTitle = "Program Info"
	MainWindowTitle = "Game Text"
	StateWindowTitle = "Game State"
	DescWindowTitle = "Game Description"
	OptWindowTitle = "Options"
	FileWindowTitle = "Location File"

	def __init__(self, parent, id, title):
		global main_directory, textlog, panellog, narrate, nar_area, wordstats, hall_filename, hof_track, log_lines
	
		wx.Frame.__init__(self, parent, -1, title, size = (800, 600), style=wx.DEFAULT_FRAME_STYLE)

		self.window = None
		self.buttonset = []
		self.messageset = []
		self.Maximize()

		# the icon for Rex -- hopefully present
		self.SetIcon(wx.Icon(os.path.join(main_directory, 'rex.ico'), wxBITMAP_TYPE_ICO))

		wx.EVT_CLOSE(self, self.OnCloseWindow)

		self.Centre(wx.BOTH)
		self.CreateStatusBar(1, wx.ST_SIZEGRIP)

		# Define some splitter window hierarchy for dividing the main window into panels
		self.splitter1 = wx.SplitterWindow(self, -1, style=wx.NO_3D|wx.SP_3D)
		self.splitter2 = wx.SplitterWindow(self.splitter1, -1, style=wx.NO_3D|wx.SP_3D)
		self.splitter3 = wx.SplitterWindow(self.splitter1, -1, style=wx.NO_3D|wx.SP_3D)
		self.splitter4 = wx.SplitterWindow(self.splitter3, -1, style=wx.NO_3D|wx.SP_3D)

		def EmptyHandler(evt): pass
		wx.EVT_ERASE_BACKGROUND(self.splitter1, EmptyHandler)
		wx.EVT_ERASE_BACKGROUND(self.splitter2, EmptyHandler)
		wx.EVT_ERASE_BACKGROUND(self.splitter3, EmptyHandler)
		wx.EVT_ERASE_BACKGROUND(self.splitter4, EmptyHandler)
		
        	# Set up the "File" menu
		fmenu = wx.Menu()
		ID_SET_NEW = 300
		ID_RESET = 403
		ID_EXIT = 302
		ID_IMPORT_INV = 304
		ID_IMPORT_GAME = 305
		fmenu.Append(ID_SET_NEW, "Set game", "Choose a new Smash game and start it.")
		fmenu.Append(ID_RESET, "Reset game", "Reset the game and start over.")
		fmenu.AppendSeparator()
		fmenu.Append(ID_IMPORT_INV, "Re-import inventory\tCtrl-I", "Update Rex's item lists by rereading (e)objects.txt.")
		fmenu.Append(ID_IMPORT_GAME, "Re-import game.txt\tCtrl-G", "Update Rex's data from game.txt.")
		fmenu.AppendSeparator()
		fmenu.Append(ID_EXIT, "Quit\tCtrl-Q", "Exit the program.")
		
		# Set up the "State" menu
		smenu = wx.Menu()
		ID_SAVE = 400
		ID_LOAD = 401
		ID_BKU = 402
		ID_REENTER = 404
		ID_QSAVE = 405
		ID_QLOAD = 406
		smenu.Append(ID_SAVE, "Save state\tAlt-S", "Save the current game state.")
		smenu.Append(ID_LOAD, "Load state\tAlt-L", "Load a previously saved game state.")
		smenu.AppendSeparator()
		smenu.Append(ID_QSAVE, "Quick save\tCtrl-S", "Make a quick save.")
		smenu.Append(ID_QLOAD, "Quick load\tCtrl-L", "Load from the last quick save state.")
		smenu.AppendSeparator()
		smenu.Append(ID_BKU, "Back up one move\tCtrl-B", "Take back the last game move, restoring the previous state.")
		smenu.Append(ID_REENTER, "Re-enter location\tCtrl-R", "Re-parse this location's code.")
		
		# Set up the "Display" menu
		dmenu = wx.Menu()
		ID_ACTAREA = 500
		ID_ACT = 501
		ID_AREA = 502
		ID_AGLHTML = 503
		ID_DUMPLOG = 504
		dmenu.Append(ID_ACT, "Show action text only\tF1", "Show the action text and hide the area text.")
		dmenu.Append(ID_AREA, "Show area text only\tF2", "Show the area text and hide the action text.")
		dmenu.Append(ID_ACTAREA, "Show both (default)\tF3", "Show both the action and area description texts.")
		dmenu.AppendSeparator()
		dmenu.Append(ID_AGLHTML, "Generate AGL-style HTML\tF4", "Generate an AGL-style HTML file showing the current game text, options, and inventory.")
		dmenu.AppendSeparator()
		dmenu.Append(ID_DUMPLOG, "Purge log\tCtrl-P", "Erase the contents of the log in the bottom-left panel.")
		
		# Set up the "Location" menu
		lmenu = wx.Menu()
		ID_AREANAR = 631
		ID_GAMESTATS = 632
		ID_FILESIZES = 633
		lmenu.Append(ID_AREANAR, "Manually narrate current area text\tF6", "Append the current state's area description text to the narration file.")
		lmenu.AppendSeparator()
		lmenu.Append(ID_GAMESTATS, "Compute current game's statistics\tF7", "View statistics for this game's location and other files.")
		lmenu.Append(ID_FILESIZES, "Compute current game's size\tF8", "Compute the total size of this game's SMA and TXT files.")
		
		# Set up the "Window" menu
		wmenu = wx.Menu()
		ID_FLIP_LEFT_PANELS = 690
		ID_FLIP_BOT_PANELS = 691
		ID_LOG_INV_PANELS = 692
		wmenu.Append(ID_FLIP_LEFT_PANELS, "Toggle left panels\tF9", "Hide/unhide the left panels.")
		wmenu.Append(ID_FLIP_BOT_PANELS, "Toggle bottom panels\tF10", "Hide/unhide the bottom panels.")
		wmenu.Append(ID_LOG_INV_PANELS, "Toggle log/inv/both panels\tF11", "Switch between the log, inventory, or both at the bottom.")


		# Add menu command for toggling inventory/log showing (3 way toggling)


        	# Set up the menu bar with the menus we've created
		self.menuBar = wx.MenuBar()
		self.menuBar.Append(fmenu, "Game");
		self.menuBar.Append(smenu, "State");
		self.menuBar.Append(dmenu, "Display");
		self.menuBar.Append(lmenu, "Location");
		self.menuBar.Append(wmenu, "Window");
		self.SetMenuBar(self.menuBar)
		
		# Link the menu events to handlers
		wx.EVT_MENU(self, ID_SET_NEW, self.TryInitGame)
		wx.EVT_MENU(self, ID_EXIT,  self.TimeToQuit)
		wx.EVT_MENU(self, ID_SAVE, self.TrySave)
		wx.EVT_MENU(self, ID_LOAD, self.TryLoad)
		wx.EVT_MENU(self, ID_QSAVE, self.TryQuickSave)
		wx.EVT_MENU(self, ID_QLOAD, self.TryQuickLoad)
		wx.EVT_MENU(self, ID_BKU, self.BackState)
		wx.EVT_MENU(self, ID_RESET, self.ResetState)
		wx.EVT_MENU(self, ID_IMPORT_INV, self.ImportInv)
		wx.EVT_MENU(self, ID_IMPORT_GAME, self.ImportGame)
		wx.EVT_MENU(self, ID_REENTER, self.Reenter)
		wx.EVT_MENU(self, ID_DUMPLOG, self.DumpLog)
		wx.EVT_MENU(self, ID_ACTAREA, self.BothText)
		wx.EVT_MENU(self, ID_ACT, self.ActText)
		wx.EVT_MENU(self, ID_AREA, self.AreaText)
		wx.EVT_MENU(self, ID_AGLHTML, self.AGLHTML)
		wx.EVT_MENU(self, ID_AREANAR, self.AreaNarrate)
		wx.EVT_MENU(self, ID_GAMESTATS, self.GameStats)
		wx.EVT_MENU(self, ID_FILESIZES, self.FileSizes)
		wx.EVT_MENU(self, ID_FLIP_LEFT_PANELS, self.ToggleLeftPanels)
		wx.EVT_MENU(self, ID_FLIP_BOT_PANELS, self.ToggleBotPanels)
		wx.EVT_MENU(self, ID_LOG_INV_PANELS, self.ToggleLogInvPanels)
		wx.EVT_MOVE(self, self.Refresh())
                
		# Make the inventory/edit/mask panel and status panel, children of splitter4 (horizontal)
		self.lp = wx.html.HtmlWindow(self.splitter4, -1, style=wx.CLIP_CHILDREN|wx.VSCROLL)
		self.stat = wx.html.HtmlWindow(self.splitter4, -1, style=wx.CLIP_CHILDREN|wx.VSCROLL)

		# Set up the log panel in the bottom left, child of splitter2 (vertical)
		self.log = wx.TextCtrl(self.splitter2, -1, style = wx.TE_MULTILINE|wx.TE_READONLY)
		font = wx.Font(10, wx.MODERN, wx.NORMAL, wx.NORMAL)
        	self.log.SetFont(font)
		wx.Log_SetActiveTarget(MyLog(self.log))
		
		# Set up the inventory panel in the bottom right, child of splitter2 (vertical)
		self.inv = wx.html.HtmlWindow(self.splitter2, -1, style=wx.CLIP_CHILDREN|wx.VSCROLL)
		wx.EVT_SIZE(self.inv, self.UpdateInvWindow)


		# Create a notebook for the upper right panel, child of splitter3 (vertical)
		self.nb = wx.Notebook(self.splitter3, 21000, style=wx.CLIP_CHILDREN)

		# Make a panel in which to set the "Program Info" window and add it to the notebook
		panel0 = wx.Panel(self.nb, -1, style=wx.CLIP_CHILDREN)
		self.infowindow = wx.html.HtmlWindow(panel0, -1, size=(784, 500))
		self.nb.AddPage(panel0, self.InfoWindowTitle)

		# Make a panel in which to set the "Game Text" window and add it to the notebook
		panel = wx.Panel(self.nb, -1, style=wx.CLIP_CHILDREN)
		self.mainwindow = wx.html.HtmlWindow(panel, -1, size=(784, 500))
		self.nb.AddPage(panel, self.MainWindowTitle)
		
		# Make a panel in which to set the "Game State" window and add it to the notebook
		panel4 = wx.Panel(self.nb, -1, style=wx.CLIP_CHILDREN)
		self.statewindow = wx.html.HtmlWindow(panel4, -1, size=(784,500))
		self.nb.AddPage(panel4, self.StateWindowTitle)
		
		# Make a panel in which to set the "Game Description" window and add it to the notebook
		panel7 = wx.Panel(self.nb, -1, style=wx.CLIP_CHILDREN)
		self.descwindow = wx.html.HtmlWindow(panel7, -1, size=(784,500))
		self.nb.AddPage(panel7, self.DescWindowTitle)

		# Make a panel in which to set the "Options" window and add it to the notebook
		self.panel5 = wx.Panel(self.nb, -1, style=wx.CLIP_CHILDREN)
		self.panel5.SetBackgroundColour('WHITE')	# this one's not an HTML window, so it needs it
		self.nb.AddPage(self.panel5, self.OptWindowTitle)

		# Make a panel in which to set the "Location File" window and add it to the notebook
		panel6 = wx.Panel(self.nb, -1, style=wx.CLIP_CHILDREN)
		self.filewindow = wx.TextCtrl(panel6, -1,
                              style = wx.TE_MULTILINE|wx.TE_READONLY|
                              wx.HSCROLL|wx.TE_RICH2|wx.TE_NOHIDESEL)
		self.nb.AddPage(panel6, self.FileWindowTitle)

		""" Detect when a notebook tab is selected, refreshing the Game State
			and File Window text -- more efficient than having these refresh on every move. """
		wx.EVT_NOTEBOOK_PAGE_CHANGED(self, 21000, self.NotebookTab)

		# Have all those panels fix their sizes when the main frame is resized
		def FixSize1(evt, window=self.mainwindow):
			window.SetSize(evt.GetSize())
		def FixSize2(evt, window=self.infowindow):
			window.SetSize(evt.GetSize())
		def FixSize3(evt, window=self.statewindow):
			window.SetSize(evt.GetSize())	
		def FixSize4(evt, window=self.panel5):
			window.SetSize(evt.GetSize())
		def FixSize5(evt, window=self.filewindow):
			window.SetSize(evt.GetSize())
		def FixSize6(evt, window=self.descwindow):
			window.SetSize(evt.GetSize())
		wx.EVT_SIZE(panel, FixSize1)
		wx.EVT_SIZE(panel0, FixSize2)
		wx.EVT_SIZE(panel4, FixSize3)
		wx.EVT_SIZE(self.panel5, FixSize4)
		wx.EVT_SIZE(panel6, FixSize5)
		wx.EVT_SIZE(panel7, FixSize6)
	
		wx.EVT_ERASE_BACKGROUND(panel, EmptyHandler)

		# Place the Rex information text in the Program Info window
		self.infowindow.SetPage(Start_Text)
		self.Show(True)

		# Link the button event to its handler
		wx.EVT_BUTTON(self, 20000, self.LicenseShow)

		# Add the main window's panels to the splitters and split them
		self.splitter1.SplitHorizontally(self.splitter3, self.splitter2, 500)
		self.splitter2.SplitVertically(self.log, self.inv, 500)
		self.splitter3.SplitVertically(self.splitter4, self.nb, 125)
		self.splitter4.SplitHorizontally(self.lp, self.stat, 200)

		# Don't allow the user to drag the sashes too close to the edges
		self.splitter1.SetMinimumPaneSize(20)
		self.splitter2.SetMinimumPaneSize(20)
		self.splitter3.SetMinimumPaneSize(20)
		self.splitter4.SetMinimumPaneSize(20)


		# Create the Options window checkboxes and listbox, and initialize the settings to
		# whatever was read from "config.agc" (or use the defaults if that wasn't possible)
		wx.StaticText(self.panel5, -1, "Player Name:", wx.Point(10,10))
		self.name_text = wx.StaticText(self.panel5, -1, playername, wx.Point(130,10))


		self.namechange_button = wx.Button(self.panel5, 1810, "Change name", wx.Point(130, 30), wx.Size(90, 25))
		wx.EVT_BUTTON(self, 1810, self.ChooseName)

		wx.StaticText(self.panel5, -1, "Hall of Fame file:", wx.Point(10,62))
		self.cb7 = wx.CheckBox(self.panel5, 1811, "  Record wins in a Hall of Fame", wx.Point(130, 60), wx.Size(170, 20))
		self.cb7.SetValue(hof_track)
		self.cb7.SetBackgroundColour('WHITE')
		self.hall_text = wx.StaticText(self.panel5, -1, "", wx.Point(130,80))
		if (hof_track):
			self.hall_text.SetLabel("Hall of Fame: " + hall_filename)

		wx.StaticText(self.panel5, -1, "Logging:", wx.Point(10,102))
		self.cb1 = wx.CheckBox(self.panel5, 1801, "  Log to panel", wx.Point(130, 100), wx.Size(150, 20))
		self.cb1.SetValue(panellog)
		self.cb1.SetBackgroundColour('WHITE')

		self.cb2 = wx.CheckBox(self.panel5, 1802, "  Log to external file:", wx.Point(130, 120), wx.Size(150, 20))
		self.log_text = wx.StaticText(self.panel5, -1, "", wx.Point(130,140))
		if textlog:
			self.log_text.SetLabel("Log file: " + log_filename)
		self.cb2.SetValue(textlog)
		self.cb2.SetBackgroundColour('WHITE')

		self.cb9 = wx.CheckBox(self.panel5, 1809, " Log all parsed lines", wx.Point(130, 160), wx.Size(150, 20))
		self.cb9.SetValue(log_lines)
		self.cb9.SetBackgroundColour('WHITE')

		wx.StaticText(self.panel5, -1, "Narration:", wx.Point(10,190))
		self.lb1 = wx.ListBox(self.panel5, 1803, wxPoint(130, 190), wxSize(120, 45),
                       ["Don't narrate", "Narrate as plain text", "Narrate as HTML"], wxLB_SINGLE)
		self.lb1.SetSelection(narrate)
		self.narrate_text = wx.StaticText(self.panel5, -1, "", wx.Point(130,240))
		if narrate:
			self.narrate_text.SetLabel("Narration file: " + narrate_filename)

		self.cb8 = wx.CheckBox(self.panel5, 1808, " Narrate area text", wx.Point(130, 260), wx.Size(150, 20))
		self.cb8.SetValue(nar_area)
		self.cb8.SetBackgroundColour('WHITE')

		wx.StaticText(self.panel5, -1, "Development Mode:", wx.Point(10,292))
		self.cb5 = wx.CheckBox(self.panel5, 1805, "", wx.Point(130, 290), wx.Size(150, 20))
		self.cb5.SetValue(develop)
		self.cb5.SetBackgroundColour('WHITE')
		
		wx.StaticText(self.panel5, -1, "Word Statistics:", wx.Point(10,312))
		self.cb6 = wx.CheckBox(self.panel5, 1806, "Compile results", wx.Point(130, 310), wx.Size(150, 20))
		self.cb6.SetValue(wordstats)
		self.cb6.SetBackgroundColour('WHITE')
		self.ws_text = wx.StaticText(self.panel5, -1, "", wx.Point(130,330))
		if wordstats:
			self.ws_text.SetLabel("Word statistics file: " + ws_filename)

		wx.StaticText(self.panel5, -1, "Minimum inventory", wx.Point(10, 350))
		wx.StaticText(self.panel5, -1, "column width:", wx.Point(10, 364))
		self.width_text = wx.StaticText(self.panel5, -1, str(min_col_width), wx.Point(130, 350))

		self.widthchange_button = wx.Button(self.panel5, 1812, "Change width", wx.Point(130, 370), wx.Size(90, 25))
		wx.EVT_BUTTON(self, 1812, self.ChooseWidth)
		
		# Link the events for using those controls to appropriate handlers
		wx.EVT_CHECKBOX(self, 1801, self.PLToggle)
		wx.EVT_CHECKBOX(self, 1802, self.TLToggle)
		wx.EVT_LISTBOX(self, 1803, self.NarrateChoice)
		wx.EVT_CHECKBOX(self, 1805, self.DevToggle)
		wx.EVT_CHECKBOX(self, 1806, self.WSToggle)
		wx.EVT_CHECKBOX(self, 1808, self.NarAreaToggle)
		wx.EVT_CHECKBOX(self, 1809, self.LogLinesToggle)
		wx.EVT_CHECKBOX(self, 1811, self.HoFToggle)

		# load the "global" config.agc file, if it's present -- had to set up log panel first
		self.LoadConfig(os.path.join(main_directory, 'config.agc'))

		# Try to load from the quick-saved state, if it's present
		is_quicksave = self.LoadState(os.path.join(main_directory, 'quick_save.ags'), 1)

		# If there *was* a quicksave (is_quicksave = 1), set the
		# notebook tab to selection #1 (Game Text) immediately, for convenience
		self.nb.SetSelection(is_quicksave)
		
		# Set the edit button and masking box
		self.RefreshLP()
	#----------------------------------------------

	def RefreshOptTab(self):
	
		global playername, hof_track, hall_filename, panellog, textlog, log_filename, log_lines
		global narrate, narrate_filename, nar_area, develop, wordstats, ws_filename, min_col_width
	
		self.name_text.SetLabel(playername)
		self.cb7.SetValue(hof_track)
		if (hof_track):
			self.hall_text.SetLabel("Hall of Fame: " + hall_filename)
		else:
			self.hall_text.SetLabel("")
		self.cb1.SetValue(panellog)
		self.cb2.SetValue(textlog)
		if (textlog):
			self.log_text.SetLabel("Log file: " + log_filename)
		else:
			self.log_text.SetLabel("")
		self.cb9.SetValue(log_lines)
		self.lb1.SetSelection(narrate)
		if narrate:
			self.narrate_text.SetLabel("Narration file: " + narrate_filename)
		else:
			self.narrate_text.SetLabel("")
		self.cb8.SetValue(nar_area)
		self.cb5.SetValue(develop)
		self.cb6.SetValue(wordstats)
		if wordstats:
			self.ws_text.SetLabel("Word statistics file: " + ws_filename)
		else:
			self.ws_text.SetLabel("")
		self.width_text.SetLabel(str(min_col_width))

	def ChooseWidth(self, event):
		""" Brings up a dialog for the user to choose a value for min_col_width, the
			minimum width an inventory column can be, in pixels, before the inventory
			is shown with fewer columns (down to a minimum of one column).  If the
			user OKs a change, the change is saved to the configuration file."""
		global min_col_width

		dlg = wx.TextEntryDialog(self, 'Enter a value for the minimum column width (minimum 5, default 120):', 'Choosing inventory column width', '')
		if dlg.ShowModal() == wx.ID_OK:
			newwidth = dlg.GetValue().strip()
			try:
				newwidth = int(newwidth)
			except ValueError:
				dlg = wx.MessageDialog(self, "Only integral values 5 or greater are valid.",
					"Illegal value", wx.OK | wx.ICON_INFORMATION)
				dlg.ShowModal()
				dlg.Destroy()
			else:
				if (newwidth < 5):
					dlg = wx.MessageDialog(self, "Only integral values 5 or greater are valid.",
						"Illegal value", wx.OK | wx.ICON_INFORMATION)
					dlg.ShowModal()
					dlg.Destroy()
				else:
					min_col_width = int(newwidth)
					self.width_text.SetLabel(str(min_col_width))
					self.TrySaveConfig()
					self.UpdateInvWindow(0)
		dlg.Destroy()

	def ChooseName(self, event):
		""" Brings up a dialog for the user to choose a playing name (for Hall of
			Fame registrations only).  If the user OKs a name change, the
			change is saved to the configuration file. """
		global playername

		dlg = wx.TextEntryDialog(self, 'Enter a name to play under:', 'Choosing player name', '')
		if dlg.ShowModal() == wx.ID_OK:
			newname = dlg.GetValue()
			if (newname.strip() == ''):
				newname = 'Player'
			elif ('|' in newname):
				dlg = wx.MessageDialog(self, "The | character can't be used in your name.",
					"Illegal character", wx.OK | wx.ICON_INFORMATION)
				dlg.ShowModal()
				dlg.Destroy()
			else:
				playername = newname
				self.name_text.SetLabel(playername)
				self.TrySaveConfig()
		dlg.Destroy()

	def NarrateChoice(self, event):
		""" Change the global "narrate" variable depending on the user's selection,
			navigates for a file/directory if narration was not set to off,
			and saves the change to the configuration file. """

		global narrate
		global narrate_filename

		choice = self.lb1.GetSelection()
		narrate = choice

		filename_pieces = os.path.split(narrate_filename)

		if narrate:
			result = ''
			dlg = wx.FileDialog(self, "Select a file to narrate to:", filename_pieces[0], filename_pieces[1], "", wx.SAVE)
			if dlg.ShowModal() == wx.ID_OK:
				result = dlg.GetPath()
			else:
				result = narrate_filename
			dlg.Destroy()
			self.narrate_text.SetLabel("Narration file: " + result)
			narrate_filename = result
		else:
			self.narrate_text.SetLabel("")
		self.TrySaveConfig()
		
	def NarAreaToggle(self, event):
		""" Change the global "nar_area" variable and save the change
			to the area text narration setting to the configuration file. """

		global nar_area
		
		if event.Checked():
			nar_area = 1
		else:
			nar_area = 0
		self.TrySaveConfig()

	def LogLinesToggle(self, event):
		""" Change the global "log_lines" variable and save the change
			to the configuration file. """

		global log_lines
		
		if event.Checked():
			log_lines = 1
		else:
			log_lines = 0
		self.TrySaveConfig()
		
	def PLToggle(self, event):
		""" Change the global "panellog" variable and save the change
			to the panel logging setting to the configuration file. """

		global panellog
		
		if event.Checked():
			panellog = 1
		else:
			panellog = 0
		self.TrySaveConfig()
		
	def TLToggle(self, event):
		""" Change the global "textlog" variable, navigates for a file/directory
			if it was toggled on, and saves the change to the configuration file. """

		global textlog
		global log_filename
		
		if event.Checked():
			textlog = 1


			filename_pieces = os.path.split(log_filename)

			result = ''
			dlg = wx.FileDialog(self, "Select a file to log to:", filename_pieces[0], filename_pieces[1], "", wx.SAVE)
			if dlg.ShowModal() == wx.ID_OK:
				result = dlg.GetPath()
			else:
				result = log_filename
			dlg.Destroy()
			self.log_text.SetLabel("Log file: " + result)
			log_filename = result
		else:
			textlog = 0
			self.log_text.SetLabel("")
		self.TrySaveConfig()
		
	def DevToggle(self, event):
		""" Changes the global "develop" variable, refreshes the two status panels
			affected by this change, and saves the change
			to the development setting to the configuration file. """

		global develop
		
		if event.Checked():
			develop = 1
		else:
			develop = 0
		self.RefreshStatus()
		self.RefreshLP()
		self.TrySaveConfig()

	def WSToggle(self, event):
		""" Changes the global "wordstats" variable, navigates for a file/directory
			if it was toggled on, and saves the change to the configuration file. """

		global wordstats
		global ws_filename
		
		if event.Checked():
			wordstats = 1

			filename_pieces = os.path.split(ws_filename)

			result = ''
			dlg = wx.FileDialog(self, "Select a file to compile to:", filename_pieces[0], filename_pieces[1], "", wx.SAVE)
			if dlg.ShowModal() == wx.ID_OK:
				result = dlg.GetPath()
			else:
				result = ws_filename
			dlg.Destroy()
			self.ws_text.SetLabel("Word statistics file: " + result)
			ws_filename = result
		else:
			wordstats = 0
			self.ws_text.SetLabel("")
		self.TrySaveConfig()

	def HoFToggle(self, event):
		""" Changes the global "hof_track" variable, navigates for a file/directory
			if it was toggled on, and saves the change to the configuration file. """

		global hof_track
		global hall_filename

		if event.Checked():
			hof_track = 1

			filename_pieces = os.path.split(hall_filename)

			result = ''
			dlg = wx.FileDialog(self, "Select a place for the Hall of Fame:", filename_pieces[0], filename_pieces[1], "", wx.SAVE)
			if dlg.ShowModal() == wx.ID_OK:
				result = dlg.GetPath()
			else:
				result = hall_filename
			dlg.Destroy()
			self.hall_text.SetLabel("Hall of Fame: " + result)
			hall_filename = result
		else:
			hof_track = 0
			self.hall_text.SetLabel("")
		self.TrySaveConfig()

	def OptClick(self, event):
		""" Take the event ID from clicking the button corresponding to an option and
			run process_option to set the Smash parser into action.  Then refresh the
			Game Text window, possibly refresh the inventory window, and possibly
			narrate the action text. """

		global narrate, inv_change
		
		if ((cur_state.wl_state == 0) or (cur_state.wl_state == 3)):
			choice_number = event.GetId() - 10000
			if (choice_number <= (cur_state.visible_opt_count + len(cur_state.move_options) + len(cur_state.act_options))):
				process_option(choice_number)
			all_text = self.TextComposite()
			if narrate:
				self.NarStuff()
			self.SetMainWindow(all_text)
			if ((inv_change) or (cur_state.wl_state == 1) or (cur_state.wl_state == -1)):
				self.UpdateInvWindow(0)
			
	def DumpLog(self, event):
		self.log.Clear()
		
	def ToggleLeftPanels(self, event):
		""" Hides or unhides the two left (status) panels.  If it hides them, it remembers
			the sash position, so that if the user unhides the panels later, the sash
			will be back where it was originally. """

		global left_shown, bot_shown, left_sash, bot_sash

		if (left_shown == 0):
			self.splitter3.SetMinimumPaneSize(20)
			self.splitter3.SetSashPosition(left_sash)
			left_shown = 1
		else:
			left_sash = self.splitter3.GetSashPosition()
			self.splitter3.SetMinimumPaneSize(0)
			self.splitter3.SetSashPosition(0)
			left_shown = 0

		# Refresh the left panels, because the text can get compressed into one spot when unhiding.
		self.RefreshStatus()
		self.RefreshLP()
		
	def ToggleBotPanels(self, event):
		""" Hides or unhides the bottom (logging/inventory) panels.  If it hides
			them, it remembers the sash position, so that if the user unhides the panels
			later, the sash will be back where it was originally. """

		global bot_shown, bot_sash, loginv_shown
		
		if (bot_shown == 0):
			self.splitter1.SetMinimumPaneSize(20)
			self.splitter1.SetSashPosition(bot_sash)
			bot_shown = 1
			if (loginv_shown == 1):
				self.splitter2.SetSashPosition(10000)
		else:
			bot_sash = self.splitter1.GetSashPosition()
			self.splitter1.SetMinimumPaneSize(0)
			self.splitter1.SetSashPosition(10000)
			bot_shown = 0

	def ToggleLogInvPanels(self, event):
		""" Toggles between showing just the log panel, just the inventory panel, or both in the
			bottom panel area of the interface. """	

		global loginv_shown, loginv_sash

		if (loginv_shown == 0):
			loginv_sash = self.splitter2.GetSashPosition()
			self.splitter2.SetMinimumPaneSize(0)
			self.splitter2.SetSashPosition(10000)
			loginv_shown = 1
		elif (loginv_shown == 1):
			self.splitter2.SetSashPosition(0)
			loginv_shown = 2
		else:
			self.splitter2.SetMinimumPaneSize(20)
			self.splitter2.SetSashPosition(loginv_sash)
			loginv_shown = 0
			
	
	def NotebookTab(self, event):
		""" Executes when a notebook tab is selected.  Refreshes the Game
			State window and the Location File window. """

		self.RefreshStateWindow(1750)
		self.RefreshFileWindow(1800)

	def NarStuff(self):
		""" Executes after the action/area text for a turn have been finalized, appending this
			text to the narration file. """

		global narrate, nar_area
		
		if (narrate == 1):
			fixed_text = cur_state.act_text.replace("<br>","\n").replace("<p>","\n\n")
			if ((nar_area) and (cur_state.wl_state in [0, 3])):
				fixed_text += '\n\n{\t' + cur_state.area_text.replace("<br>","\n").replace("<p>","\n\n") + '\t}'
		else:
			if ((nar_area) and (cur_state.wl_state in [0, 3])):
				fixed_text = cur_state.act_text + '<blockquote><font size=-1>' + cur_state.area_text + '</font></blockquote>'
			else:
				fixed_text = cur_state.act_text + '<p>'
		NarText(fixed_text)

	def AreaNarrate(self, event):
		""" Executes when the menu command for manually narrating the current state's area description
			text is chosen.  Normally used when, by default, area text is not narrated, but the user
			wants to narrate it for that move.  Also it can be used to force narration of area text which
			would not normally be displayed (upon a lose screen or a parsing error screen). """

		global narrate_filename, narrate

		if not narrate:
			WriteText('------AREA TEXT NARRATION FAILED: narration in either text or HTML form must be enabled first------')
			return
		if (cur_state.directory == ''):
			dlg = wx.MessageDialog(self, "There is currently\nno game state.",
					"Error", wx.OK | wx.ICON_INFORMATION)
			dlg.ShowModal()
			dlg.Destroy()
			return
		if (narrate == 1):
			fixed_text = '\n{\t' + cur_state.area_text.replace("<br>","\n").replace("<p>","\n\n") + '\t}'
		else:
			fixed_text = '<blockquote><font size=-1>' + cur_state.area_text + '</font></blockquote>'
		NarText(fixed_text)
		WriteText('------AREA TEXT NARRATED------')


	def RefreshStateWindow(self, event):
		""" Refreshes the Game State window, providing a button to execute this function again. """

		statetext = ''
		if (cur_state.location != ''):
			statetext += '<font size=+1><b>Location: </b></font>' + cur_state.location + '<br>'
			statetext += '<font size=+1><b>Moves: </b></font>' + str(cur_state.moves) + '<br>'
			statetext += '<font size=+1><b>Random Number Seed: </b></font>' + str(cur_state.random_seed) + '<hr>'
			statetext += '<font size=+1><b>State Variables:</b></font><br>'
			if (cur_state.svars != {}):
				svar_list = []
				var_keys = cur_state.svars.keys()
				for var_key in var_keys:
					loc_keys = cur_state.svars[var_key].keys()
					for loc_key in loc_keys:
						svar_list.append([loc_key, var_key])
				svar_list.sort(lambda a, b: ((a[0]>b[0]) or (b[0]==a[0] and a[1]>b[1])) * 2 - 1)
				for svar_pair in svar_list:
					statetext += '<b>' + svar_pair[0] + '.' + svar_pair[1] + '</b> = ' + str(cur_state.svars[svar_pair[1]][svar_pair[0]]) + '<br>'
			statetext += '<hr><font size=+1><b>Global Variables:</b></font><br>'
			if (cur_state.gvars != {}):
				var_keys = cur_state.gvars.keys()
				var_keys.sort()
				for var_key in var_keys:
					statetext += '<b>' + var_key + '</b> = ' + str(cur_state.gvars[var_key]) + '<br>'
			statetext += '<hr><font size=+1><b>Upward Timers:</b></font><br>'
			if (cur_state.up_timers != {}):
				ut_keys = cur_state.up_timers.keys()
				ut_keys.sort()
				for ut_key in ut_keys:
					statetext += '<b>' + ut_key + '</b> = ' + str(cur_state.up_timers[ut_key]) + '<br>'
			statetext += '<hr><font size=+1><b>Downward Timers:</b></font><br>'
			if (cur_state.down_timers != {}):
				dt_keys = cur_state.down_timers.keys()
				dt_keys.sort()
				for dt_key in dt_keys:
					statetext += '<b>' + dt_key + '</b> = ' + str(cur_state.down_timers[dt_key]) + '<br>'
		else:
			statetext += '<font size=+1><b>No game has been selected.</b></font>'
		self.statewindow.SetPage(statetext)
		
	def RefreshFileWindow(self, event):
		""" This executes automatically when the window gets
			focus, so the user never needs to do it manually. """

		if (cur_state.location != ''):
			try:
				ffile = open(os.path.join(cur_state.directory, cur_state.location + '.sma'),'rU')
				text = "Location: " + cur_state.location + "\n-------------------------\n\n"
				text += ffile.read()
				ffile.close()
			except IOError:
				text = 'I/O Error: unable to open "' + cur_state.location + '.sma" for viewing.'
        	else:
        		text = 'No game has been selected.'
		self.filewindow.SetValue(text)
		font = wx.Font(10, wx.MODERN, wx.NORMAL, wx.NORMAL)
        	self.filewindow.SetFont(font)
			
	def EditDialog(self, event):
		""" This executes when the user clicks the Edit State button in the top left panel. """

		global next_jump
		global just_jumped
		global cur_parse_file
		global expression
		global act_list, area_list, block_text_lines
		global com_count
			
		if (cur_state.directory == ''):
			dlg = wx.MessageDialog(self, "There is currently\nno game state.",
					"Error", wx.OK | wx.ICON_INFORMATION)
			dlg.ShowModal()
			dlg.Destroy()
		else:
			dlg = wx.TextEntryDialog(self, 'Enter a line of Smash code:', 'Editing Current State', '')
			if dlg.ShowModal() == wx.ID_OK:
				line = dlg.GetValue()
				WriteText('------EDIT LINE: "' + line + '"------')
				success = 1
				if (line[0] == 'x'):
					line = line.strip()
					try:
						if (line[1] != ' '):
							WriteText('Syntax error: " " expected following start of "x" command, or "x" must be alone on the input line.')
							WriteText('------EDIT UNSUCCESSFUL------')
							dlg.Destroy()
							return
					except IndexError:
						line = "x " + cur_state.location
					x_pieces = line.split()
					if (len(x_pieces) != 2):
						WriteText('Syntax Error: exactly one location tag must follow "x" command.')
						WriteText('------EDIT UNSUCCESSFUL------')
						dlg.Destroy()
						return
					try:
						temp_state.load(cur_state)
						com_count = {'p':0, 'P':0, 'A':0, 'I':0, 'i':0, 'g':0, '.':0, ',':0, ';':0, 'c':0, 'C':0, 'L':0, 'G':0, 'D':0, 'z':0, 'Z':0, 'f':0, 'r':0, 'v':0, 's':0, 'a':0, 'd':0, '^':0, '@':0, '@r':0, '@+':0, '@-':0, 'E':0, '*':0, 'l':0, 'w':0, 'W':0, 'S':0, 'e':0, 't+':0, 't-':0, 't<<':0, 't>>':0, '#':0, '~':0, 'words':0, 'wordbytes':0}
						locname = x_pieces[1]
						error_found = self.LocStats(locname)
						WriteText('\n---FILE STATISTICS---')
						WriteText('"p" text commands: ' + str(com_count['p']))
						WriteText('"P" text commands: ' + str(com_count['P']))
						WriteText('"A" text commands: ' + str(com_count['A']))
						WriteText('"." text commands: ' + str(com_count['.']))
						WriteText('"," text commands: ' + str(com_count[',']))
						WriteText('";" text commands: ' + str(com_count[';']))
						WriteText('"c" conditional commands: ' + str(com_count['c']))
						WriteText('"C" conditional commands: ' + str(com_count['C']))
						WriteText('"L" loop commands: ' + str(com_count['L']))
						WriteText('"G" commands: ' + str(com_count['G']))
						WriteText('"D" commands: ' + str(com_count['D']))
						WriteText('"z" commands: ' + str(com_count['z']))
						WriteText('"Z" commands: ' + str(com_count['Z']))
						WriteText('"@" constant options: ' + str(com_count['@']))
						WriteText('"@" redefinitions: ' + str(com_count['@r']))
						WriteText('"@ +" commands: ' + str(com_count['@+']))
						WriteText('"@ -" commands: ' + str(com_count['@-']))
						WriteText('"^" movement options: ' + str(com_count['^']))
						WriteText('"*" action options: ' + str(com_count['*']))
						WriteText('"v" setting commands: ' + str(com_count['v']))
						WriteText('"s" setting commands: ' + str(com_count['s']))
						WriteText('"a" setting commands: ' + str(com_count['a']))
						WriteText('"d" setting commands: ' + str(com_count['d']))
						WriteText('"e" setting commands: ' + str(com_count['e']))
						WriteText('"f" function commands: ' + str(com_count['f']))
						WriteText('"r" return value commands: ' + str(com_count['r']))
						WriteText('"g" jump commands: ' + str(com_count['g']))
						WriteText('"t" up-timer set commands: ' + str(com_count['t+']))
						WriteText('"t" down-timer set commands: ' + str(com_count['t-']))
						WriteText('"t" timer rewind commands: ' + str(com_count['t<<']))
						WriteText('"t" timer ffwd commands: ' + str(com_count['t>>']))
						WriteText('"w" win commands: ' + str(com_count['w']))
						WriteText('"W" win commands: ' + str(com_count['W']))
						WriteText('"l" lose commands: ' + str(com_count['l']))
						WriteText('"i" commands: ' + str(com_count['i']))
						WriteText('"I" commands: ' + str(com_count['I']))
						WriteText('"S" status set commands: ' + str(com_count['S']))
						WriteText('"#" comments: ' + str(com_count['#']))
						WriteText('"~" function definitions: ' + str(com_count['~']))
						WriteText('"E" error commands: ' + str(com_count['E']))
						WriteText('Total words used in .sma file description text: ' + str(com_count['words']))
						WriteText('Total bytes of these words, including spaces: ' + str(com_count['wordbytes']) + '\n---------------------')
					except IOError:	
						WriteText('IOError: location not found for "x" command.')
						WriteText('------EDIT UNSUCCESSFUL------')
					cur_state.load(temp_state)
				else:
					try:
						fe = open(os.path.join(cur_state.directory, cur_state.location + '.sma'), 'rU')

						""" Back up the current state, in case an error happens after
							some changes, but not all, have been made when ParsingException's
							raised (e.g. if an "f" command does a bunch of things). """

						temp_state.load(cur_state)
						cur_parse_file = cur_state.location
						parsestatus = parseline(fe, 0, 0, line, cur_state.location, 1)
						if (parsestatus == 2):
							act_entry = "".join(block_text_lines)
							expression = act_entry
							act_entry = eval_string_imagetags(expression)
							act_list.append(act_entry)
							block_text_lines = []
						if (parsestatus == 3):
							area_entry = "".join(block_text_lines)
							expression = area_entry
							area_entry = eval_string_imagetags(expression)
							area_list.append(area_entry)
							block_text_lines = []
						all_text = self.TextComposite()
					except IOError:
						WriteText('Unable to open current location file, "' + os.path.join(cur_state.directory, cur_state.location + ".sma") + '"')
						WriteText('------EDIT UNSUCCESSFUL------')
					except ParsingException, message:
						WriteText('PARSING ERROR:\n' + str(message))
						WriteText('------EDIT UNSUCCESSFUL------')
						cur_state.load(temp_state)
						just_jumped = 0
					else:
						self.SetMainWindow(all_text)
						if (line[0].upper() == 'M'):
							self.TrySaveConfig()
							self.RefreshLP()
					fe.close()
				if ((inv_change) or (cur_state.wl_state == 1) or (cur_state.wl_state == -1)):
					self.UpdateInvWindow(0)
			dlg.Destroy()
					
	def ExamineInventory(self, event):
		""" This executes when the user clicks the "Use/Examine Item" button in the inventory panel. """

		global develop, just_jumped, next_jump, cur_parse_file, parse_context, narrate
		global S1, S2, S3, S4, S5
	
		if (cur_state.directory == ''):
			dlg = wx.MessageDialog(self, "There is currently\nno game state.",
					"Error", wx.OK | wx.ICON_INFORMATION)
			dlg.ShowModal()
			dlg.Destroy()
		else:
			choice_list = []
			number_nicks = {}
			j = 0
			for tag in enum_namekeys:
				if (cur_state.enum_inv.has_key(tag)):
					if (cur_state.enum_inv[tag] == 1):
						name_part = enum_names[tag][0]
					else:
						name_part = enum_names[tag][1]
					if (develop):
						choice_list.append(name_part + ' (' + tag + ')')
					else:
						choice_list.append(name_part + ' (#' + str(j+1) + ')')
						number_nicks[j] = tag
					j += 1
			k = 0
			for tag in reg_namekeys:
				if (cur_state.reg_inv.has_key(tag)):
					if (develop):
						choice_list.append(reg_names[tag] + ' (' + tag + ')')
					else:
						choice_list.append(reg_names[tag] + ' (#' + str(j+k+1) + ')')
						number_nicks[j + k] = tag
					k += 1
			dlg = wx.SingleChoiceDialog(self, 'Select an item:', 'Selection', choice_list, wx.OK|wx.CANCEL)
			if dlg.ShowModal() == wx.ID_OK:
				line = dlg.GetStringSelection()
				line_pieces = line.split()
				line_tag = line_pieces[-1][1:-1]
				if (not develop):
					line_tag = number_nicks[int(line_tag[1:])-1]
				try:
					if (choice_list.index(line) < j):	# it was an enumerated object selection
						itemcall = 'f eobjectactions.' + line_tag
						fi = open(os.path.join(cur_state.directory, 'eobjectactions.sma'), 'rU')
						if (cur_state.enum_inv[line_tag] == 1):
							enum_fullname = enum_names[line_tag][0]
						else:
							enum_fullname = enum_names[line_tag][1]
						WriteText('~ ' + str(cur_state.moves + 1) + ' Enumerated Object Action : ' + line_tag + ' (' + enum_fullname + ')')
						if (narrate == 1):
							NarText('---(' + str(cur_state.moves + 1) + ' / ' + cur_state.location + ')  ~ ' + ' Enumerated Object Action : ' + line_tag + ' (' + enum_fullname + ')')
						if (narrate == 2):
							NarText('<b>(' + str(cur_state.moves + 1) + ' / ' + cur_state.location + ')  ~ ' + ' Enumerated Object Action : ' + line_tag + ' (' + enum_fullname + ')' + '</b><br>')
					else:
						itemcall = 'f objectactions.' + line_tag
						fi = open(os.path.join(cur_state.directory, 'objectactions.sma'), 'rU')
						WriteText('~ ' + str(cur_state.moves + 1) + ' Regular Object Action : ' + line_tag + ' (' + reg_names[line_tag] + ')')
						if (narrate == 1):
							NarText('---(' + str(cur_state.moves + 1) + ' / ' + cur_state.location + ')  ~ ' + ' Regular Object Action : ' + line_tag + ' (' + reg_names[line_tag] + ')')
						if (narrate == 2):
							NarText('<b>(' + str(cur_state.moves + 1) + ' / ' + cur_state.location + ')  ~ ' + ' Regular Object Action : ' + line_tag + ' (' + reg_names[line_tag] + ')' + '</b><br>')
				except IOError:
					if (choice_list.index(line) < j):
						warn('Unable to open file "' + os.path.join(cur_state.directory, 'eobjectactions.sma') + '"\n')
					else:
						warn('Unable to open file "' + os.path.join(cur_state.directory, 'objectactions.sma') + '"\n')
					cur_state.wl_state = -1
				else:
					last10_state.load(last9_state)
					last9_state.load(last8_state)
					last8_state.load(last7_state)
					last7_state.load(last6_state)
					last6_state.load(last5_state)
					last5_state.load(last4_state)
					last4_state.load(last3_state)
					last3_state.load(last2_state)
					last2_state.load(last_state)
					last_state.load(cur_state)
					cur_state.moves += 1
					cur_state.act_text = ''
					act_list = []
					cur_state.move_options = []
					cur_state.act_options = []
					cur_state.act_image = ''
					cur_state.area_image = ''
	
					cur_state.wl_state = 0
					
					init_genrand(line_tag)
					adjust_timers()
					
					try:
						""" S2 set to -1 because examining an item will involve a function
							call, which increments S2 in parseline, but in Smash, S2
							is supposed to be 0 examining an item. """
						S1 = 0
						S2 = -1
						S3 = 126
						S4 = encode(line_tag)
						S5 = encode(cur_state.location)

						cur_parse_file = cur_state.location
						parse_context = 2
						parsestatus = parseline(fi, 0, 0, itemcall, cur_state.location)
	
						# fix S2
						S2 = 0
						
						""" Determine if a jump is required, then process the new
							location's code, generating new options.  If a win/loss
							command is encountered, don't parse the new location. """
						
						if ((cur_state.wl_state == 0) or (cur_state.wl_state == 3)):
							try:
								f_pre = open(os.path.join(cur_state.directory, 'repeatpre.sma'), 'rU')
								parse_context = 3
								cur_parse_file = 'repeatpre'
								parseblock(f_pre, 'repeatpre')
								f_pre.close()
							except IOError:
								pass
						parse_context = 0
						if ((cur_state.wl_state == 0) or (cur_state.wl_state == 3)):
							S5 = encode(next_jump)
							cur_parse_file = next_jump
							try:
								newloc_parse()
							except IOError:
								warn('Unable to open file "' + os.path.join(cur_state.directory, next_jump) + '.sma"\n')
								cur_state.wl_state = -1
								just_jumped = 0
						if ((cur_state.wl_state == 0) or (cur_state.wl_state == 3)):
							try:
								f_post = open(os.path.join(cur_state.directory, 'repeatpost.sma'), 'rU')
								parse_context = 4
								cur_parse_file = 'repeatpost'
								parseblock(f_post, 'repeatpost')
								f_post.close()
							except IOError:
								pass
						parse_context = 0
					except ParsingException, message:
						WriteText('PARSING ERROR:\n' + str(message))
						cur_state.wl_state = -1
						just_jumped = 0
						parse_context = 0
					fi.close()
				all_text = self.TextComposite()
				if narrate:
					self.NarStuff()
				self.SetMainWindow(all_text)
				self.nb.SetSelection(1)
				if ((inv_change) or (cur_state.wl_state == 1) or (cur_state.wl_state == -1)):
					self.UpdateInvWindow(0)
			dlg.Destroy()

	def RefreshLP(self):
		""" Refreshes the upper left panel. """

		global develop, gvarmasked, svarmasked, robjmasked, eobjmasked	
		
		lp_text = ''
				
		# Editing button:
		if (develop and ((cur_state.wl_state == 0) or (cur_state.wl_state == 3))):
			lp_text += '<wxp class="wxButton" width="90" height="25"><param name="label" value="Edit State"><param name="id" value=1600></wxp><br>'
			wx.EVT_BUTTON(self, 1600, self.EditDialog)

		# Masked variables/inventory/timers box:
		if develop:
			if ((gvarmasked != {}) or (svarmasked != {}) or (robjmasked != {}) or (eobjmasked != {})or (timermasked != {})):
				lp_text += '<br><table cellpadding=4 cellspacing=0 bgcolor="#ffddff"><tr><td><font size=+1><b>Masked:</b></font><br>'
				if (gvarmasked != {}):
					lp_text += '<b>Global Vars:</b><br>'
					g_keys = gvarmasked.keys()
					g_keys.sort()
					for g_key in g_keys:
						lp_text += g_key + '<br>'
				if (svarmasked != {}):
					lp_text += '<b>State Vars:</b><br>'
					s_locs = svarmasked.keys()
					s_locs.sort()
					for s_loc in s_locs:
						s_vars = svarmasked[s_loc].keys()
						s_vars.sort()
						for s_var in s_vars:
							lp_text += s_loc + '.' + s_var + '<br>'
				if (eobjmasked != {}):
					lp_text += '<b>Enum. Inv.:</b><br>'
					e_keys = eobjmasked.keys()
					e_keys.sort()
					for e_key in e_keys:
						lp_text += e_key + '<br>'
				if (robjmasked != {}):
					lp_text += '<b>Reg. Inv.:</b><br>'
					r_keys = robjmasked.keys()
					r_keys.sort()
					for r_key in r_keys:
						lp_text += r_key + '<br>'
				if (timermasked != {}):
					lp_text += '<b>Timers:</b><br>'
					t_keys = timermasked.keys()
					t_keys.sort()
					for t_key in t_keys:
						lp_text += t_key + '<br>'
			lp_text += '</td></tr></table>'
		self.lp.SetPage(lp_text)
		

	def RefreshStatus(self):
		""" Refreshes the bottom left status panel. """

		global develop
		global expression
		
		if (cur_state.wl_state != -1):
			status_text = ''
			if (develop):
				status_text += '<table cellpadding=4 cellspacing=0 bgcolor="#eeeeff"><tr><td><b>'
				status_text += cur_state.location + '</b></td></tr></table>'
			if (len(cur_state.status_list)):
				status_error = 0
				status_text += '<table cellpadding=4 cellspacing=0 bgcolor="#ffffdd"><tr><td>'
				try:
					for status_num in cur_state.status_list:
						if (develop):
							status_text += '<b>' + str(status_num) + '</b> '
						expression = cur_state.status_lines[status_num]
						status_text += eval_string_brackets(cur_state.status_lines[status_num]) + '<br>'
				except ParsingException, message:
					WriteText('PARSING ERROR (in status line):\n' + str(message))
					cur_state.wl_state = -1
					all_text = self.TextComposite()
					self.SetMainWindow(all_text)
				status_text += '</td></tr></table>'
			
			self.stat.SetPage(status_text)
				
	def SetMainWindow(self, text):
		""" Refreshes the Game Text window, normally after a move is made.  Since changes to the
			status lines often occur alongside, we'll run the refresh function for that panel
			too. """

		self.mainwindow.SetPage(text)
		self.RefreshStatus()

	def UpdateInvWindow(self, event):
		""" Recalculates the HTML for the inventory table, depending on the items in the inventory
			and the window width, which can force the number of columns in the table below 5. """

		global min_col_width, inv_change

		inv_change = 0

		# num_columns = determined by window width
		window_width = self.inv.GetSizeTuple()[0]
		if (window_width < min_col_width + 50):
			window_width = min_col_width + 50	# force at *least* one column in the next computation
		num_columns = (window_width - 50) / min_col_width	# 50 is "overhead" for borders, scroll bar, etc.
		j = 0
		ordered_inv = {}
		for tag in enum_namekeys:
			if (cur_state.enum_inv.has_key(tag)):
				ordered_inv[j] = str(cur_state.enum_inv[tag]) + ' '
				if (cur_state.enum_inv[tag] == 1):
					ordered_inv[j] += enum_names[tag][0]
				else:
					ordered_inv[j] += enum_names[tag][1]
				j += 1
		for tag in reg_namekeys:
			if (cur_state.reg_inv.has_key(tag)):
				ordered_inv[j] = reg_names[tag]
				j += 1
		inv_text = ''
		if (j==0):
			if (cur_state.directory != ''):
				inv_text = '<center><table cellpadding=0 cellspacing=0 border=0><tr><td align=center><b>You\'re not carrying anything.</b></td></tr></table></center>'
		else:
			column_length = 1 + (j - 1) / num_columns
			# effectively: column_span = j +/ column_length
			column_span = j / column_length + (j % column_length != 0)
			inv_text = '<center><table cellpadding=4 cellspacing=0 bgcolor="#e8ffe8"><tr><td><table cellpadding=0 cellspacing=0 border=0><tr><td colspan=' + str(column_span) + ' align=center>'
			if ((cur_state.wl_state == 0) or (cur_state.wl_state == 2) or (cur_state.wl_state == 3)):
				inv_text += '<wxp class="wxButton" width="110" height="25"><param name="label" value="Use/Examine Item"><param name="id" value=30000></wxp>'
				wx.EVT_BUTTON(self, 30000, self.ExamineInventory)
			inv_text += '<b>&nbsp;&nbsp;You\'re Carrying:&nbsp;<font color="red">(' + str(j) + ')</font></b></td></tr><tr><td><font size=-5>&nbsp;</font></td></tr>'
			for row in range(0, column_length):
				inv_text += '<tr>'
				for col in range(0, column_span):
					inv_text += '<td>'
					spotnum = row + (col * column_length)
					if (spotnum < j):
						#inv_text += '<wxp class="wxButton" width="10" height="10"><param name="label" value=" "><param name="id" value=' + str(30000+spotnum) + '></wxp>&nbsp;<font size=-1>'
						inv_text += '<font size=-1>'
						inv_text += ordered_inv[spotnum]
						inv_text += '</font>'
					inv_text += '</td>'
				inv_text += '</tr>'
			inv_text += '</table></td></tr></table></center>'
		self.inv.SetPage(inv_text)
		
	def TextComposite(self):
		""" Generates the HTML which will be set as the Game Text window's text.  Buttons for
			each option are embedded as well, and the button events are linked to the
			handler OptClick. """

		global act_list, area_list, moved_back, expression, hof_fields, hide
	
		showing_act = 1
		showing_area = 1
		
		
		if (moved_back == 0):
			cur_state.act_text = "".join(act_list)
			cur_state.area_text = "".join(area_list)
		moved_back = 0
		if ((cur_state.wl_state != 0) or (hide == 1) or (cur_state.area_text == '')):
			showing_area = 0
		if ((hide == 2) or (cur_state.act_text == '')):
			showing_act = 0
		if (showing_act and showing_area):
			main_text = (''.join(['<html><body bgcolor="white" text="black"><center><br><table cellpadding=7 cellspacing=3><tr><td bgcolor="#ffe4dd">', cur_state.act_text, '</td><td bgcolor="#ccccff">', cur_state.area_text, '</td></tr></table>']))
		elif (showing_act and not showing_area):
			main_text = (''.join(['<html><body bgcolor="white" text="black"><center><br><table cellpadding=7 cellspacing=3><tr><td bgcolor="#ffe4dd">', cur_state.act_text,'</td></tr></table>']))
		elif (showing_area and not showing_act):
			main_text = (''.join(['<html><body bgcolor="white" text="black"><center><br><table cellpadding=7 cellspacing=3><tr><td bgcolor="#ccccff">', cur_state.area_text,'</td></tr></table>']))
		else:
			main_text = (''.join(['<html><body bgcolor="white" text="black"><center>']))
		if (cur_state.wl_state == 1):
			try:
				losefile = open(os.path.join(cur_state.directory, 'lose.txt'),'rU')
				temp_state.load(cur_state)
				main_text += '</center>'
				lose_text = losefile.read()
				losefile.close()
				expression = lose_text
				lose_text = eval_string_imagetags(expression)
				cur_state.load(temp_state)
				main_text += lose_text
			except IOError:
				pass
			except ParsingException, message:
				WriteText('PARSING ERROR:\n' + str(message))
				main_text += '<b>[PARSING ERROR in lose.txt { } expression -- see log]</b>'
				cur_state.load(temp_state)
			ID_BKU2 = 600
			main_text += '<center><hr><font size=+3>You Lost!</font><p>Oops.<br>Press the button below to undo that last fatal move.<p><wxp class="wxButton" width="150" height="30"><param name="label" value="Undo Last Move"><param name="id" value=600></wxp>'
			wx.EVT_BUTTON(self, 600, self.BackState)
		if ((cur_state.wl_state == 2) or (cur_state.wl_state == 3)):
			try:
				winfile = open(os.path.join(cur_state.directory, 'win.txt'),'rU')
				temp_state.load(cur_state)
				main_text += '</center>'
				win_text = winfile.read()
				winfile.close()
				expression = win_text
				win_text = eval_string_imagetags(expression)
				cur_state.load(temp_state)
				main_text += win_text
			except IOError:
				pass
			except ParsingException, message:
				WriteText('PARSING ERROR:\n' + str(message))
				main_text += '<b>[PARSING ERROR in win.txt { } expression -- see log]</b>'
				cur_state.load(temp_state)
			ID_BKU2 = 600
			main_text += '<center><hr><font size=+3>You Won!</font><p>Congratulations!'
			if (hof_fields != []):
				main_text += '  You have been added to the Hall of Fame for this game.  Good job!'
			if (cur_state.wl_state == 2):
				main_text += '<p><wxp class="wxButton" width="150" height="30"><param name="label" value="Undo last move"><param name="id" value=600></wxp>'
				wx.EVT_BUTTON(self, 600, self.BackState)
			else:
				main_text += '<p><font size=+3>But Wait!  It\'s Not Over!</font><p>Although you have successfully won this game, there is still more to play!<hr><table cellpadding=7 cellspacing=3><tr><td bgcolor="#ccccff">' + cur_state.area_text + '</td></tr></table>'
		if (cur_state.wl_state == -1):
			ID_BKU2 = 600
			main_text += '<center><p><font size=+1>Smash Parsing Error!  View the log file to see the type of error encountered and other information.</font><p><wxp class="wxButton" width="150" height="30"><param name="label" value="Undo last move"><param name="id" value=600></wxp>'
			wx.EVT_BUTTON(self, 600, self.BackState)
		if ((cur_state.wl_state == 0) or (cur_state.wl_state == 3)):
			main_text += '<p><table cellpadding=0 cellspacing=0><tr>'
			icount = 0
			if (cur_state.visible_opt_count > 0):
				main_text += '<td valign=top>'
				for opt in cur_state.const_options:
					if (opt.visible):
						icount += 1
						main_text += '<table cellpadding=0 cellspacing=3><tr><td ALIGN=RIGHT><wxp class="wxButton" width="25" height="17"><param name="label" value=&nbsp;><param name="id" value=' + str(10000 + icount) + '></wxp></td><td WIDTH="80%" NOWRAP>' + opt.text + '</td></tr></table>'
						wx.EVT_BUTTON(self, (10000 + icount), self.OptClick)
				main_text += '</td>'
			if (cur_state.move_options != []):
				main_text += '<td valign=top>'
				for opt in cur_state.move_options:
					icount += 1
					main_text += '<table cellpadding=0 cellspacing=3><tr><td ALIGN=RIGHT><wxp class="wxButton" width="25" height="17"><param name="label" value=&nbsp;><param name="id" value=' + str(10000 + icount) + '></wxp></td><td WIDTH="80%">' + opt.text + '</td></tr></table>'
					wx.EVT_BUTTON(self, (10000 + icount), self.OptClick)
				main_text += '</td>'
			if (cur_state.act_options != []):
				main_text += '<td valign=top>'
				for opt in cur_state.act_options:
					icount += 1
					main_text += '<table cellpadding=0 cellspacing=3><tr><td ALIGN=RIGHT><wxp class="wxButton" width="25" height="17"><param name="label" value=&nbsp;><param name="id" value=' + str(10000 + icount) + '></wxp></td><td WIDTH="80%">' + opt.text + '</td></tr></table>'
					wx.EVT_BUTTON(self, (10000 + icount), self.OptClick)
				main_text += '</td>'
			main_text += '</tr></table>'
			if (cur_state.act_image != ''):
				main_text += '<br><img src="' + cur_state.act_image + '">'
			elif (cur_state.area_image != ''):
				main_text += '<br><img src="' + cur_state.area_image + '">'
		main_text += '</center>'
		if ((cur_state.wl_state != 0) and (cur_state.wl_state != 3)):
			self.RefreshLP()
			
		return main_text

	def AGLHTML(self, event):
		global act_list, area_list, moved_back, expression, hof_fields

		
		if (cur_state.directory == ''):
			dlg = wx.MessageDialog(self, "No game has been started.",
					"Error", wx.OK | wx.ICON_INFORMATION)
			dlg.ShowModal()
			dlg.Destroy()
		else:
			showing_area = 1
			showing_act = 1
			if ((cur_state.wl_state != 0) or (cur_state.area_text == '')):
				showing_area = 0
			if (cur_state.act_text == ''):
				showing_act = 0
			if (showing_act and showing_area):
				main_text = (''.join(['<html><body bgcolor="white" text="black"><center><br><table cellpadding=7 cellspacing=3><tr><td bgcolor="#ffe4dd">', cur_state.act_text, '</td><td bgcolor="#ccccff">', cur_state.area_text, '</td></tr></table></center>']))
			elif (showing_act and not showing_area):
				main_text = (''.join(['<html><body bgcolor="white" text="black"><center><br><table cellpadding=7 cellspacing=3><tr><td bgcolor="#ffe4dd">', cur_state.act_text,'</td></tr></table></center>']))
			elif (showing_area and not showing_act):
				main_text = (''.join(['<html><body bgcolor="white" text="black"><center><br><table cellpadding=7 cellspacing=3><tr><td bgcolor="#ccccff">', cur_state.area_text,'</td></tr></table></center>']))
			else:
				main_text = (''.join(['<html><body bgcolor="white" text="black">']))
			if (cur_state.wl_state == 1):
				try:
					losefile = open(os.path.join(cur_state.directory, 'lose.txt'),'rU')
					temp_state.load(cur_state)
					lose_text = losefile.read()
					losefile.close()
					expression = lose_text
					lose_text = eval_string_imagetags(expression)
					cur_state.load(temp_state)
					main_text += lose_text
				except IOError:
					pass
				except ParsingException, message:
					WriteText('PARSING ERROR:\n' + str(message))
					main_text += '<b>[PARSING ERROR in lose.txt { } expression -- see log]</b>'
					cur_state.load(temp_state)
				main_text += '<center><hr><font size=+3>You Lost!</font><p>Oops.</center>'
			if ((cur_state.wl_state == 2) or (cur_state.wl_state == 3)):
				try:
					winfile = open(os.path.join(cur_state.directory, 'win.txt'),'rU')
					temp_state.load(cur_state)
					win_text = winfile.read()
					winfile.close()
					expression = win_text
					win_text = eval_string_imagetags(expression)
					cur_state.load(temp_state)
					main_text += win_text
				except IOError:
					pass
				except ParsingException, message:
					WriteText('PARSING ERROR:\n' + str(message))
					main_text += '<b>[PARSING ERROR in win.txt { } expression -- see log]</b>'
					cur_state.load(temp_state)
				main_text += '<center><hr><font size=+3>You Won!</font><p>Congratulations!'
				if (hof_fields != []):
					main_text += '  You have been added to the Hall of Fame for this game.  Good job!'
				if (cur_state.wl_state == 2):
					main_text += '<p>'
				else:
					main_text += '<p><font size=+3>But Wait!  It\'s Not Over!</font><p>Although you have successfully won this game, there is still more to play!<hr><table cellpadding=7 cellspacing=3><tr><td bgcolor="#ccccff">' + cur_state.area_text + '</td></tr></table>'
				main_text += '</center>'
			if (cur_state.wl_state == -1):
				ID_BKU2 = 600
				main_text += '<center><p><font size=+1>Smash Parsing Error!  View the log file to see the type of error encountered and other information.</font><p></center>'
			if ((cur_state.wl_state == 0) or (cur_state.wl_state == 3)):
				main_text += '<p><center><table cellpadding=0 cellspacing=0><tr>'
				icount = 0
				if (cur_state.visible_opt_count > 0):
					main_text += '<td valign=top><table cellpadding=0 cellspacing=0>'
					for opt in cur_state.const_options:
						if (opt.visible):
							icount += 1
							main_text += '<tr><td><input type=radio name=choice value="0"></td><td>&nbsp;</td><td>' + opt.text + '</td></tr>'
					main_text += '</table></td>'
				if (cur_state.move_options != []):
					main_text += '<td valign=top><table cellpadding=0 cellspacing=0>'
					for opt in cur_state.move_options:
						icount += 1
						main_text += '<tr><td><input type=radio name=choice value="0"></td><td>&nbsp;</td><td>' + opt.text + '</td></tr>'
					main_text += '</table></td>'
				if (cur_state.act_options != []):
					main_text += '<td valign=top><table cellpadding=0 cellspacing=0>'
					for opt in cur_state.act_options:
						icount += 1
						main_text += '<tr><td><input type=radio name=choice value="0"></td><td>&nbsp;</td><td>' + opt.text + '</td></tr>'
					main_text += '</table></td>'
				main_text += '<td><table cellspacing=5 cellpadding=5><tr><td><input type="submit" value=" OK "></td></tr></table></td>'
				main_text += '</tr></table>'
				if (cur_state.act_image != ''):
					main_text += '<br><img src="' + cur_state.act_image + '">'
				elif (cur_state.area_image != ''):
					main_text += '<br><img src="' + cur_state.area_image + '">'
				main_text += '</center>'
			num_columns = 5
			j = 0
			ordered_inv = {}
			for tag in enum_namekeys:
				if (cur_state.enum_inv.has_key(tag)):
					ordered_inv[j] = str(cur_state.enum_inv[tag]) + ' '
					if (cur_state.enum_inv[tag] == 1):
						ordered_inv[j] += enum_names[tag][0]
					else:
						ordered_inv[j] += enum_names[tag][1]
					j += 1
			for tag in reg_namekeys:
				if (cur_state.reg_inv.has_key(tag)):
					ordered_inv[j] = reg_names[tag]
					j += 1
			if j:
				column_length = 1 + (j - 1) / num_columns
				# effectively: column_span = j +/ column_length
				column_span = j / column_length + (j % column_length != 0)
				main_text += '<p><center><table cellpadding=4 cellspacing=0 bgcolor="e8ffe8"><tr><td><table cellpadding=0 cellspacing=0 border=0><tr><td colspan=' + str(column_span) + ' align=center><b>&nbsp;&nbsp;You\'re Carrying:</b></td></tr><tr><td><font size=-5>&nbsp;</font></td></tr>'
				for row in range(0, column_length):
					main_text += '<tr>'
					for col in range(0, column_span):
						main_text += '<td>'
						spotnum = row + (col * column_length)
						if (spotnum < j):
							main_text += '<table cellpadding=0 cellspacing=0><tr><td><input type=radio name=choice value="0"></td><td><font size=-1>&nbsp;</font></td><td><font size=-1>'
							main_text += ordered_inv[spotnum]
							main_text += '</font></td></tr></table>'
						main_text += '</td>'
					main_text += '</tr>'
				main_text += '</table></td></tr></table></center>'
			result = ''
			dlg = wx.FileDialog(self, "Select a file to output the HTML to:", cur_directory, "", "*.html", wx.SAVE)
			if dlg.ShowModal() == wx.ID_OK:
				result = dlg.GetPath()
			dlg.Destroy()
			if (result != ''):
				if (result[-5:] != '.html'):
					result += '.html'
				aglfile = open(result, 'w')
				aglfile.write(main_text)
				aglfile.close()
				WriteText('AGL-style HTML file "' + result + '" written.')

		
	def TryInitGame(self, event):
		""" Executed when the user opts to start a new game, creating a file dialog
			for locating the correct directory.  If one is chosen, the game is
			initialized with InitGame. """

		global cur_directory
	
		result = ''
		dlg = wx.FileDialog(self, "Select a start file:", cur_directory, wildcard="Smash Start Files (start.sma)|start.sma")
		if dlg.ShowModal() == wx.ID_OK:
			result = dlg.GetPath()
		dlg.Destroy()
		if (result != '' and (result[-9:] == 'start.sma')):
			self.InitGame(result)
		
		
	def TryLoad(self, event):
		""" Executed when the user opts to load a game state, creating a file dialog
			for locating the desired save file.  If one is chosen, the game is
			loaded with LoadState. """
		
		global cur_directory
		
		result = ''
		dlg = wx.FileDialog(self, "Select a saved state:", cur_directory, wildcard="Rex Interpreter States (*.ags)|*.*")
		if dlg.ShowModal() == wx.ID_OK:
			result = dlg.GetPath()
		dlg.Destroy()
		if (result != '' and (result[-4:] == '.ags')):
			self.LoadState(result)
			
	def TrySave(self,event):
		""" Executed when the user opts to save a gane state, creating a file dialog
			for locating the desired save file.  If one is chosen, the game is
			saved with SaveState. """

		global cur_directory
		
		if (cur_state.directory == ''):
			dlg = wx.MessageDialog(self, "There is currently\nno game state.",
					"Error", wx.OK | wx.ICON_INFORMATION)
			dlg.ShowModal()
			dlg.Destroy()
		else:
			result = ''
			dlg = wx.FileDialog(self, "Select a saved state:", cur_directory, "", "*.ags", wx.SAVE)
			if dlg.ShowModal() == wx.ID_OK:
				result = dlg.GetPath()
			dlg.Destroy()
			if (result != ''):
				if (result[-4:] != '.ags'):
					result += '.ags'
				self.SaveState(result)
				
	def TryQuickSave(self, event):
		""" Provides one step of the quick-saving process, checking whether a game
			has actually been started before passing on to SaveState. """

		global main_directory
	
		if (cur_state.directory == ''):
			dlg = wx.MessageDialog(self, "There is currently\nno game state.",
					"Error", wx.OK | wx.ICON_INFORMATION)
			dlg.ShowModal()
			dlg.Destroy()
		else:
			self.SaveState(os.path.join(main_directory, 'quick_save.ags'))
			
			
	def LocStats(self, locname):
		""" Parses the location "locname" and compiles statistics for it.  As
			changes to cur_state shouldn't be kept, anything which calls this
			function should back up cur_state into temp_state first, then
			copy temp_state back to cur_state afterward. """

		global cur_parse_file, stats_compute, parse_context, S2
				
		cur_state.const_options = []
		cur_state.move_options = []
		cur_state.act_options = []
		cur_state.location = locname
		WriteText('Parsing "' + locname + '.sma":')
		try:
			f4 = open(os.path.join(cur_state.directory, locname + '.sma'), 'rU')
			cur_parse_file = locname
			stats_compute = 1
			parse_context = 0
			S2 = 0
			parseblock(f4, locname)
			f4.close()
			stats_compute = 0
		except ParsingException, message:
			warn('PARSING ERROR:\n' + str(message))
			stats_compute = 0
			return 1
		return 0
			
			
	def GameStats(self, event):
		""" This is the statistics computation routine acting on the entire game's files.  It
			creates a list of all the appropriate .sma files (discounting admin.sma) and
			runs the single-location LocStats function on each.  "win.txt" and "lose.txt"
			are counted as well for their description text word counts.  cur_state is
			backed up to temp_state in advance so that any changes occurring from parsing
			aren't permanent. """

		global com_count
		global wordcount_dictionary
		global wordstats
		global ws_filename
	
		temp_state.load(cur_state)
		if (cur_state.directory == ''):
			dlg = wx.MessageDialog(self, "No game has been started.",
					"Error", wx.OK | wx.ICON_INFORMATION)
			dlg.ShowModal()
			dlg.Destroy()
		else:
			start_time = time.clock()
			filelist = os.listdir(cur_state.directory)
			filelist.sort()
			WriteText('\n------PARSING ALL LOCATIONS, PLEASE WAIT------')
			count = 0
			error_count = 0
			com_count = {'p':0, 'P':0, 'A':0, 'I':0, 'i':0, 'g':0, '.':0, ',':0, ';':0, 'c':0, 'C':0, 'L':0, 'G':0, 'D':0, 'z':0, 'Z':0, 'f':0, 'r':0, 'v':0, 's':0, 'a':0, 'd':0, '^':0, '@':0, '@r':0, '@+':0, '@-':0, 'E':0, '*':0, 'l':0, 'w':0, 'W':0, 'S':0, 'e':0, 't+':0, 't-':0, 't<<':0, 't>>':0, '#':0, '~':0, 'words':0, 'wordbytes':0}
			temp_state.load(cur_state)
			for locfile in filelist:
				loc_pieces = locfile.split(".")
				if (len(loc_pieces) == 2):
					if ((loc_pieces[1] == "sma") and (loc_pieces[0] != "admin")):
						locname = loc_pieces[0]
						count += 1
						error_found = self.LocStats(locname)
						if (error_found):
							error_count += 1
			WriteText('----------------------------------------------\nSMA file statistics:')
			
			if (error_count > 0):
				WriteText(str(count) + ' files parsed, ' + str(error_count) + ' with syntax error(s).\n')
			else:
				WriteText(str(count) + ' files parsed, no syntax errors found.\n')
			WriteText('"p" text commands: ' + str(com_count['p']))
			WriteText('"P" text commands: ' + str(com_count['P']))
			WriteText('"A" text commands: ' + str(com_count['A']))
			WriteText('"." text commands: ' + str(com_count['.']))
			WriteText('"," text commands: ' + str(com_count[',']))
			WriteText('";" text commands: ' + str(com_count[';']))
			WriteText('"c" conditional commands: ' + str(com_count['c']))
			WriteText('"C" conditional commands: ' + str(com_count['C']))
			WriteText('"L" loop commands: ' + str(com_count['L']))
			WriteText('"G" commands: ' + str(com_count['G']))
			WriteText('"D" commands: ' + str(com_count['D']))
			WriteText('"z" commands: ' + str(com_count['z']))
			WriteText('"Z" commands: ' + str(com_count['Z']))
			WriteText('"@" constant options: ' + str(com_count['@']))
			WriteText('"@" redefinitions: ' + str(com_count['@r']))
			WriteText('"@ +" commands: ' + str(com_count['@+']))
			WriteText('"@ -" commands: ' + str(com_count['@-']))
			WriteText('"^" movement options: ' + str(com_count['^']))
			WriteText('"*" action options: ' + str(com_count['*']))
			WriteText('"v" setting commands: ' + str(com_count['v']))
			WriteText('"s" setting commands: ' + str(com_count['s']))
			WriteText('"a" setting commands: ' + str(com_count['a']))
			WriteText('"d" setting commands: ' + str(com_count['d']))
			WriteText('"e" setting commands: ' + str(com_count['e']))
			WriteText('"f" function commands: ' + str(com_count['f']))
			WriteText('"r" return value commands: ' + str(com_count['r']))
			WriteText('"g" jump commands: ' + str(com_count['g']))
			WriteText('"t" up-timer set commands: ' + str(com_count['t+']))
			WriteText('"t" down-timer set commands: ' + str(com_count['t-']))
			WriteText('"t" timer rewind commands: ' + str(com_count['t<<']))
			WriteText('"t" timer ffwd commands: ' + str(com_count['t>>']))
			WriteText('"w" win commands: ' + str(com_count['w']))
			WriteText('"W" win commands: ' + str(com_count['W']))
			WriteText('"l" lose commands: ' + str(com_count['l']))
			WriteText('"i" commands: ' + str(com_count['i']))
			WriteText('"I" commands: ' + str(com_count['I']))
			WriteText('"S" status set commands: ' + str(com_count['S']))
			WriteText('"#" comments: ' + str(com_count['#']))
			WriteText('"~" function definitions: ' + str(com_count['~']))
			WriteText('"E" error commands: ' + str(com_count['E']))
			WriteText('\nTotal words used in .sma file description text: ' + str(com_count['words']))
			WriteText('Total bytes of these words, including spaces: ' + str(com_count['wordbytes']))
			WriteText('\n---READING WIN/LOSE FILES---')
			try:
				f4 = open(os.path.join(cur_state.directory, 'win.txt'), 'rU')
				WriteText('Reading "win.txt"...')
				while (get_depth(f4) != -1):
					line = f4.readline()
					if (line[-1] == '\n'):
						line = line[:-1]
					wordcount = CountWords(line)
					com_count['words'] += wordcount
					com_count['wordbytes'] += (len(line))
				f4.close()
			except IOError:
				WriteText('No "win.txt" file found')
			try:
				f4 = open(os.path.join(cur_state.directory, 'lose.txt'), 'rU')
				WriteText('Reading "lose.txt"...')
				while (get_depth(f4) != -1):
					line = f4.readline()
					if (line[-1] == '\n'):
						line = line[:-1]
					wordcount = CountWords(line)
					com_count['words'] += wordcount
					com_count['wordbytes'] += (len(line))
				f4.close()
			except IOError:
				WriteText('No "lose.txt" file found')
			WriteText('\nTotal words used in all files\' description text: ' + str(com_count['words']))
			WriteText('Total bytes of these words, including spaces: ' + str(com_count['wordbytes']))
			
			if wordstats:
				wordlist = wordcount_dictionary.items()
				wordlist.sort(lambda a,b: ((b[1]-a[1]>0) or (b[1]==a[1] and a[0]>b[0])) * 2 - 1)
				wordf = open(os.path.join(cur_state.directory, ws_filename), 'w')
				for word in wordlist:
					number = str(word[1])
					wordf.write(number +  ' ' * (8-len(number)) + word[0] + '\n')
				wordf.close()
				wordcount_dictionary = {}
				WriteText('\nWord totals written to "' + ws_filename + '"')
			stop_time = time.clock()
			WriteText('\nElapsed time: ' + str(stop_time - start_time) + ' seconds.')
			WriteText('----------------------------------------------')
			
			cur_state.load(temp_state)

	def FileSizes(self,event):
		""" This function simply counts how many .sma files and Smash .txt files the game uses, not
			counting admin.sma (as it's not relevant to the player), then finds the total
			size of all these files added together. """

		if (cur_state.directory == ''):
			dlg = wx.MessageDialog(self, "No game has been started.",
					"Error", wx.OK | wx.ICON_INFORMATION)
			dlg.ShowModal()
			dlg.Destroy()
		else:
			sma_size = 0
			sma_count = 0
			other_size = 0
			other_count = 0
			filelist = os.listdir(cur_state.directory)
			for sizefile in filelist:
				size_pieces = sizefile.split(".")
				if (len(size_pieces) == 2):
					if ((size_pieces[1] == "sma") and (size_pieces[0] != "admin")):
						sma_size += os.path.getsize(os.path.join(cur_state.directory, sizefile))
						sma_count += 1
					if (size_pieces[1] == "txt"):
						if (size_pieces[0] == "game") or (size_pieces[0] == "objects") or (size_pieces[0] == "eobjects") or (size_pieces[0] == "win") or (size_pieces[0] == "lose"):
							other_size += os.path.getsize(os.path.join(cur_state.directory, sizefile))
							other_count += 1
			WriteText('\n--------------------------------------------------------\nFile count and total size:')
			WriteText('(Not counting admin.sma and non-Smash .txt files)')
			WriteText('\n' + str(sma_count) + ' SMA files: ' + str(sma_size) + ' bytes')
			WriteText(str(other_count) + ' TXT files: ' + str(other_size) + ' bytes')
			WriteText('\nTOTAL:\n' + str(sma_count + other_count) + ' files: ' + str(sma_size + other_size) + ' bytes')


	def SetObjects(self, directory):
		""" Executed when a new game is set (whether directly or because
			of a load operation) or when the user manually re-imports
			the inventory. """
	
		global reg_names
		global enum_names
		global reg_namekeys
		global enum_namekeys
		global tag_regex
	
		reg_names = {}
		enum_names = {}
		reg_namekeys = []
		enum_namekeys = []
	
		inv_drop = 0
		
		WriteText('------IMPORTING (E)OBJECT.TXT NAMES------')
		try:
			objfile = open(os.path.join(directory, 'objects.txt'),'rU')
			WriteText('Loading regular object tags/names from "' + os.path.join(directory, 'objects.txt') + '"')
			objline = objfile.readline().strip()
			line = 1
			while (objline != ''):
				objline_pieces = objline.split("|")
				if (len(objline_pieces) == 1):
					WriteText('No "|" character on line ' + str(line) + 'of "objects.txt", skipping over this line.')
					pass
				elif (len(objline_pieces) > 2):
					WriteText('More than one "|" character on line ' + str(line) + 'of "objects.txt", skipping over this line.')
					pass
				elif reg_names.has_key(objline_pieces[0]):
					WriteText('Duplicated tag on line ' + str(line) + 'of "objects.txt", skipping over this line.')
				elif (not len(objline_pieces[0])):
					WriteText('Zero-length tag on line ' + str(line) + 'of "objects.txt", skipping over this line.')
				elif (not len(objline_pieces[1])):
					WriteText('Zero-length object full name on line ' + str(line) + ' of "objects.txt", skipping over this line.')
				else:
					if tag_regex.match(objline_pieces[0]):
						reg_names[objline_pieces[0]] = objline_pieces[1]
						reg_namekeys.append(objline_pieces[0])
					else:
						WriteText('Illegal tag character on line ' + str(line) + ' of "objects.txt", skipping over this line.')
				objline = objfile.readline().strip()
				line += 1
			objfile.close()
			for robj in cur_state.reg_inv.keys():
				if (not reg_names.has_key(robj)):
					WriteText('*** WARNING *** Regular object tagged "' + robj + '" no longer found, so it will be removed from the inventory.')
					del cur_state.reg_inv[robj]
					inv_drop = 1
		except IOError:
			WriteText('No "objects.txt" file openable in directory.')
		try:
			eobjfile = open(os.path.join(directory, 'eobjects.txt'),'rU')
			WriteText('Loading enumerated object tags/names from "' + os.path.join(directory, 'eobjects.txt') + '"')
			eobjline = eobjfile.readline().strip()
			line = 1
			while (eobjline != ''):
				eobjline_pieces = eobjline.split("|")
				if (len(eobjline_pieces) < 3):
					WriteText('Fewer than two "|" characters on line ' + str(line) + ' of "eobjects.txt", skipping over this line.')
					pass
				elif (len(eobjline_pieces) > 3):
					WriteText('More than two "|" characters on line ' + str(line) + ' of "eobjects.txt", skipping over this line.')
					pass
				elif enum_names.has_key(eobjline_pieces[0]):
					WriteText('Duplicated tag on line ' + str(line) + 'of "eobjects.txt", skipping over this line.')
				elif (not len(eobjline_pieces[0])):
					WriteText('Zero-length tag on line ' + str(line) + 'of "eobjects.txt", skipping over this line.')
				elif ((not len(eobjline_pieces[1])) or (not len(eobjline_pieces[2]))):
					WriteText('Zero-length object full name on line ' + str(line) + 'of "eobjects.txt", skipping over this line.')
				else:
					if tag_regex.match(eobjline_pieces[0]):
						enum_names[eobjline_pieces[0]] = [eobjline_pieces[1], eobjline_pieces[2]]
						enum_namekeys.append(eobjline_pieces[0])
					else:
						WriteText('Illegal tag character on line ' + str(line) + 'of "eobjects.txt", skipping over this line.')
				eobjline = eobjfile.readline().strip()
				line += 1
			eobjfile.close()
			for eobj in cur_state.enum_inv.keys():
				if (not enum_names.has_key(eobj)):
					WriteText('*** WARNING *** Enumerated object tagged "' + eobj + '" no longer found, so it will be removed from the inventory.')
					del cur_state.enum_inv[eobj]
					inv_drop = 1
		except IOError:
			WriteText('No "eobjects.txt" file openable in directory.')
		if ((inv_change) or (cur_state.wl_state == -1)):
			self.UpdateInvWindow(0)


	def ImportInv(self, event):
		""" Executes when the user manually re-imports the game's inventory.  The Game Text
			window's text is also refreshed in case any inventory items are dropped due
			to their tags being outdated and no longer found in the game's "objects.txt"
			or "eobjects.txt". """

		self.SetObjects(cur_state.directory)
		all_text = self.TextComposite()
		self.SetMainWindow(all_text)

	def ImportGame(self, event):
		""" Executes when the user tries to manually re-imports the
			game.txt file data -- logs the attempt, then passes on to
			the next step. """

		if (cur_state.directory == ''):
			dlg = wx.MessageDialog(self, "No game has been started.",
					"Error", wx.OK | wx.ICON_INFORMATION)
			dlg.ShowModal()
			dlg.Destroy()
		else:
			WriteText('------IMPORTING GAME.TXT DATA------')
			self.NewGameData()

	def NewGameData(self):
		""" When the current game is switched for any reason (or when data related to
			the Hall of Fame is updated), this function
			executes.  It passes most of the work to the global function UpdateHoFData,
			which needs to be called by the Smash parser as well.  The title bar is
			then updated with the new game's title, and the Game Description window
			is refreshed. """

		global game_title

		try:
			UpdateHoFData()
			self.SetTitle(game_title + " - Rex")
			self.UpdateDescWindow()
		except ParsingException, line:
			if (str(line) != "No title"):
				WriteText('*** WARNING ***\nSyntax error in "game.txt" -- data not imported.\nLine: "' + str(line) + '"')
			raise "fail"
		except IOError:
			WriteText('*** WARNING ***\n"game.txt" not openable in the game\'s directory.')
			raise "fail"

	def UpdateDescWindow(self):
		
		global game_title, difficulty, credits, synopsis_long, hof_fields, hall_filename

		desctext = '<center>'
		try:
			gfile = open(os.path.join(cur_state.directory, 'graphic.gif'), 'rU')
			gfile.close()
			desctext += "<img src= '" + os.path.join(cur_state.directory, 'graphic.gif') + "'>"
		except IOError:
			desctext += "<font size=+3>" + game_title + "</font>"
		if (difficulty >= 0):
			desctext += "<br>Difficulty: " + str(difficulty)
		desctext += "</center>"
		desctext += "<p>"
		if (credits != ""):
			desctext += "<font size=+1><b>Credits:</b></font><br>" + credits + "<p>"
		if (synopsis_long != ""):
			desctext += "<font size=+1><b>Synopsis:</b></font><br>" + synopsis_long + "<p>"
		desctext += '<font size=+1><b>Hall of Fame fields:</b></font>'
		if (hof_fields == []):
			desctext += "<br>This game will not have a Hall of Fame."
		else:
			desctext += "<p><table border=1><tr><td><b>Field Name</b></td><td><b>Sorting Method</b></td><td><b>Custom Entry Values</b></td></tr>"
			for hof_field in hof_fields:
				desctext += '<tr><td>' + hof_field[0] + '</td><td>'
				if (hof_field[1] == 0):
					desctext += 'Numeric, Descending'
				elif (hof_field[1] == 1):
					desctext += 'Numeric, Ascending'
				elif (hof_field[1] == 2):
					desctext += 'Alphabetic, Descending'
				elif (hof_field[1] == 3):
					desctext += 'Alphabetic, Ascending'
				else:
					desctext += 'Customized'
				desctext += '</td><td>'
				custom_count = len(hof_field[2])
				for index in range (0, custom_count-1):
					desctext += hof_field[2][index] + ', '
				if (custom_count):
					desctext += hof_field[2][custom_count-1]
				desctext += '</td></tr>'
			try:
				hallsize = os.path.getsize(hall_filename)
				desctext += '</table><p><center><wxp class="wxButton" width="150" height="25"><param name="label" value="View Hall of Fame"><param name = "id" value = 2700></wxp></center>'
				wx.EVT_BUTTON(self, 2700, self.CreateHoFWindow)
			except OSError:
				desctext += '</table>'
		self.descwindow.SetPage(desctext)
		
	def CreateHoFWindow(self, event):
		""" Executed when the user clicks on the button (in the Game Description
			tab) to view the Hall of Fame.  Creates a separate HTML window
			for showing the HoF entries (if any). """
		
		HallOfFame = HoFWindow(self)
		HallOfFame.Show(True)

	def TryQuickLoad(self, event):
		""" Executes upon startup.  If a quick-save file is present, it is automatically loaded
			for the user's convenience. """

		global main_directory
	
		self.LoadState(os.path.join(main_directory, 'quick_save.ags'))


	def InitGame(self, result):
		""" Executes when a new game is set or when the current game is reset. """

		global next_jump, just_jumped, cur_directory, cur_parse_file
		global act_list, area_list, moved_back, parse_context
		global game_title, reg_names, enum_names, reg_namekeys, enum_namekeys
		global S1, S2, S3, S4, S5
		global narrate
	
		temp_state.load(cur_state)
		temp_game_title = game_title
		temp_reg_names = reg_names
		temp_enum_names = enum_names
		temp_reg_namekeys = reg_namekeys
		temp_enum_namekeys = enum_namekeys
		try:
			cur_state.directory = result[:-9]
			self.NewGameData()
			cur_state.wl_state = 0
			cur_state.moves = 0
			cur_state.location = 'start'
			cur_state.act_text = ''
			cur_state.area_text = ''
			cur_state.gvars = {}
			cur_state.svars = {}
			cur_state.status_lines = {}
			cur_state.status_list = []
			cur_state.up_timers = {}
			cur_state.down_timers = {}
			cur_state.const_options = []
			cur_state.move_options = []
			cur_state.act_options = []
			cur_state.visible_opt_count = 0
			cur_state.enum_inv = {}
			cur_state.reg_inv = {}
			cur_state.act_image = ''
			cur_state.area_image = ''
			cur_state.random_seed = 1788195708
			self.SetObjects(cur_state.directory)
			act_list = []
			area_list = []
			S1 = 0
			S2 = 0
			S3 = 0
			S4 = 0
			S5 = 0
			WriteText('------STARTING A NEW GAME IN "' +  cur_state.directory + '"-------')
			parse_context = 0
			try:
				next_jump = 'copts'
				newloc_parse()
			except IOError:
				WriteText('"copts.sma" not found.')
			next_jump = 'start'
			newloc_parse()
			cur_state.visible_opt_count -= (len(cur_state.move_options) + len(cur_state.act_options))
			cur_state.move_options = []
			cur_state.act_options = []
			if ((cur_state.wl_state == 0) or (cur_state.wl_state == 3)):
				try:
					f_pre = open(os.path.join(cur_state.directory, 'repeatpre.sma'), 'rU')
					parse_context = 3
					cur_parse_file = 'repeatpre'
					parseblock(f_pre, 'repeatpre')
					f_pre.close()
				except IOError:
					pass
			parse_context = 0
			if ((cur_state.wl_state == 0) or (cur_state.wl_state == 3)):
				S5 = encode(next_jump)
				cur_parse_file = next_jump
				newloc_parse()
			if ((cur_state.wl_state == 0) or (cur_state.wl_state == 3)):
				try:
					f_post = open(os.path.join(cur_state.directory, 'repeatpost.sma'), 'rU')
					parse_context = 4
					cur_parse_file = 'repeatpost'
					parseblock(f_post, 'repeatpost')
					f_post.close()
				except IOError:
					pass
			moved_back = 0
			all_text = self.TextComposite()
			if narrate:
				self.NarStuff()
			self.SetMainWindow(all_text)
			self.nb.SetSelection(1)
			last10_state.load(last9_state)
			last9_state.load(last8_state)
			last8_state.load(last7_state)
			last7_state.load(last6_state)
			last6_state.load(last5_state)
			last5_state.load(last4_state)
			last4_state.load(last3_state)
			last3_state.load(last2_state)
			last2_state.load(last_state)
			last_state.load(temp_state)
			if (cur_state.directory != temp_state.directory):
				self.TryLoadConfig()
			self.RefreshStatus()
			self.UpdateInvWindow(0)
			cur_directory = cur_state.directory
		except ParsingException, message:
			WriteText('PARSING ERROR:\n' + str(message))
			cur_state.load(temp_state)
			just_jumped = 0
			game_title = temp_game_title
			reg_names = copy.deepcopy(temp_reg_names)
			enum_names = copy.deepcopy(temp_enum_names)
			reg_namekeys = copy.deepcopy(temp_reg_namekeys)
			enum_namekeys = copy.deepcopy(temp_enum_namekeys)
			try:
				self.NewGameData()
			except:
				pass
		except "fail":
			WriteText('Game initialization failed.')
			cur_state.load(temp_state)
			just_jumped = 0
			game_title = temp_game_title
			reg_names = copy.deepcopy(temp_reg_names)
			enum_names = copy.deepcopy(temp_enum_names)
			reg_namekeys = copy.deepcopy(temp_reg_namekeys)
			enum_namekeys = copy.deepcopy(temp_enum_namekeys)
			if (cur_state.directory != ''):
				try:
					self.NewGameData()
				except:
					pass
		self.RefreshLP()

	def TryLoadConfig(self):
		""" When a switch to the current game is made (or a new one is started, if no game is
			in progress), it will try to load in a "config.agc" file in that game's directory.
			If no such file is found, a new one will be copied there from the global "config.agc" """

		try:
			confile = open(os.path.join(cur_state.directory, 'config.agc'), 'rU')
			confile.close()
			self.LoadConfig(os.path.join(cur_state.directory, 'config.agc'))
		except IOError:
			try:
				confile = open(os.path.join(main_directory, 'config.agc'), 'rU')
				confile.close()
				self.LoadConfig(os.path.join(main_directory, 'config.agc'))
			except IOError:
				pass
		
	def LoadConfig(self, result):
		""" Executes upon startup to recover the user's global options which are set in
			the Options tab and saved in "config.agc" in Rex's directory.  Also executes
			when a new game is set/loaded and the options need to be changed.  """

		global gvarmasked, svarmasked, robjmasked, eobjmasked, timermasked, develop
		global narrate, narrate_filename, nar_area
		global playername, panellog, textlog, log_filename, log_lines, min_col_width
		global wordstats, ws_filename
		global hof_track, hall_filename
		
		try:
			confile = open(result, 'rU')
			temp_playername = confile.readline().strip()
			if ('|' in temp_playername):
				temp_playername = 'Player'
			confile.readline()
			line = confile.readline()
			temp_gvarmasked = {}
			temp_svarmasked = {}
			temp_robjmasked = {}
			temp_eobjmasked = {}
			temp_timermasked = {}
			while (line[0] != "-"):
				g_var = line.strip()
				temp_gvarmasked[g_var] = 1
				line = confile.readline()
			line = confile.readline()
			while (line[0] != "-"):
				s_pieces = line.strip().split(".")
				if (not svarmasked.has_key(s_pieces[0])):
					temp_svarmasked[s_pieces[0]] = {}
				temp_svarmasked[s_pieces[0]][s_pieces[1]] = 1
				line = confile.readline()
			line = confile.readline()
			while (line[0] != "-"):
				the_reg = line.strip()
				temp_robjmasked[the_reg] = 1
				line = confile.readline()
			line = confile.readline()
			while (line[0] != "-"):
				the_enum = line.strip()
				temp_eobjmasked[the_enum] = 1
				line = confile.readline()
			line = confile.readline()
			while (line[0] != "-"):
				the_timer = line.strip()
				temp_timermasked[the_timer] = 1
				line = confile.readline()
			line = confile.readline()
			temp_hof_track = int(line[0])
			if (temp_hof_track not in [0,1]):
				temp_hof_track = 0
			line = confile.readline()
			temp_hall_filename = line.strip()
			line = confile.readline()
			temp_panellog = int(line[0])
			if (temp_panellog not in [0,1]):
				temp_panellog = 0
			line = confile.readline()
			temp_textlog = int(line[0])
			if (temp_textlog not in [0,1]):
				temp_textlog = 0
			line = confile.readline()
			temp_log_filename = line.strip()
			line = confile.readline()
			temp_log_lines = int(line[0])
			if (temp_log_lines not in [0,1]):
				temp_log_lines = 0
			line = confile.readline()
			temp_narrate = int(line[0])
			if (temp_narrate not in [0,1,2]):
				temp_narrate = 0
			line = confile.readline()
			temp_narrate_filename = line.strip()
			line = confile.readline()
			temp_nar_area = int(line[0])
			if (temp_nar_area not in [0,1]):
				temp_nar_area = 0
			line = confile.readline()
			temp_develop = int(line[0])
			if (temp_develop not in [0,1]):
				temp_develop = 0
			line = confile.readline()
			temp_wordstats = int(line[0])
			if (temp_wordstats not in [0,1]):
				temp_wordstats = 0
			line = confile.readline()
			temp_ws_filename = line.strip()
			line = confile.readline()
			temp_min_col_width = int(line.strip())
			if (temp_min_col_width < 5):
				temp_min_col_width = 5
			confile.close()
			WriteText('------CONFIGURATION FILE "' + result + '" LOADED------')

			# As the config file was successfully loaded, we can
			# officially make the changes to the real options:

			playername = temp_playername
			gvarmasked = copy.deepcopy(temp_gvarmasked)
			svarmasked = copy.deepcopy(temp_svarmasked)
			robjmasked = copy.deepcopy(temp_robjmasked)
			eobjmasked = copy.deepcopy(temp_eobjmasked)
			timermasked = copy.deepcopy(temp_timermasked)
			hof_track = temp_hof_track
			hall_filename = temp_hall_filename
			panellog = temp_panellog
			textlog = temp_textlog
			log_filename = temp_log_filename
			log_lines = temp_log_lines
			narrate = temp_narrate
			narrate_filename = temp_narrate_filename
			nar_area = temp_nar_area
			develop = temp_develop
			wordstats = temp_wordstats
			ws_filename = temp_ws_filename
			min_col_width = temp_min_col_width
			self.RefreshOptTab()
		except IOError:
			WriteText('------CONFIGURATION FILE "' + result + '" NOT FOUND, OPTIONS NOT CHANGED------')
			pass
		except:
			WriteText('------ERROR IN CONFIGURATION FILE "' + result + '", OPTIONS NOT CHANGED------')		
		

	def LoadState(self, result, flag=0):
		""" Executes whenever a saved game state is loaded, whether manually by the user or
			automatically upon startup.  In the latter case, flag is passed the value 1
			by the calling function. """
	
		global next_jump
		global cur_directory
		global act_list, area_list
		global reg_names, enum_names, reg_namekeys, enum_namekeys
		

		# Back up cur_state, in case an error occurs in loading and cur_state needs to be recovered
		temp_state.load(cur_state)
		temp_reg_names = copy.deepcopy(reg_names)
		temp_enum_names = copy.deepcopy(enum_names)
		temp_reg_namekeys = copy.deepcopy(reg_namekeys)
		temp_enum_namekeys = copy.deepcopy(enum_namekeys)
		found_file = 0
		try:
			loadfile = open(result,'rU')
			if flag:
				WriteText('------AUTO-LOADING FROM "' + os.path.join(main_directory, 'quick_save.ags') + '"------')
			found_file = 1
			cur_state.gvars = {}
			cur_state.svars = {}
			cur_state.enum_inv = {}
			cur_state.reg_inv = {}
			cur_state.up_timers = {}
			cur_state.down_timers = {}
			cur_state.const_options = []
			cur_state.move_options = []
			cur_state.act_options = []
			cur_state.status_lines = {}
			cur_state.status_list = []
			line = loadfile.readline()
			cur_state.directory = line.split("=")[1].strip()
			reg_names = {}
			enum_names = {}
			reg_namekeys = []
			enum_namekeys = []
			self.SetObjects(cur_state.directory)
			line = loadfile.readline()
			cur_state.moves = int(line.split("=")[1].strip())
			line = loadfile.readline()
			cur_state.wl_state = int(line.split("=")[1].strip())
			line = loadfile.readline()
			cur_state.random_seed = int(line.split("=")[1].strip())
			line = loadfile.readline()
			cur_state.location = line.split("=")[1].strip()
			line = loadfile.readline()
			cur_state.visible_opt_count = int(line.split("=")[1].strip())
			loadfile.readline()
			cur_state.act_text = loadfile.readline().strip()
			act_list = [cur_state.act_text]
			loadfile.readline()
			cur_state.area_text = loadfile.readline().strip()
			area_list = [cur_state.area_text]
			loadfile.readline()
			opt_type = loadfile.read(1)
			i = 0
			while (opt_type != "-"):
				opt_text = loadfile.readline().strip()
				line = loadfile.readline()
				opt_loc = line.split("=")[1].strip()
				line = loadfile.readline()
				opt_tag = line.split("=")[1].strip()
				line = loadfile.readline()
				opt_pos = int(line.split("=")[1].strip())
				line = loadfile.readline()
				opt_vis = int(line.split("=")[1].strip())
				cur_state.const_options.append(Option('@', opt_loc, opt_pos, opt_tag, opt_text))
				cur_state.const_options[i].visible = opt_vis
				i += 1
				opt_type = loadfile.read(1)
			loadfile.readline()
			opt_type = loadfile.read(1)
			while (opt_type != "-"):
				opt_text = loadfile.readline().strip()
				line = loadfile.readline()
				opt_loc = line.split("=")[1].strip()
				line = loadfile.readline()
				opt_tag = line.split("=")[1].strip()
				line = loadfile.readline()
				opt_pos = int(line.split("=")[1].strip())
				cur_state.move_options.append(Option('^', opt_loc, opt_pos, opt_tag, opt_text))
				opt_type = loadfile.read(1)
			loadfile.readline()
			opt_type = loadfile.read(1)
			while (opt_type != "-"):
				opt_text = loadfile.readline().strip()
				line = loadfile.readline()
				opt_loc = line.split("=")[1].strip()
				line = loadfile.readline()
				opt_tag = line.split("=")[1].strip()
				line = loadfile.readline()
				opt_pos = int(line.split("=")[1].strip())
				cur_state.act_options.append(Option('*', opt_loc, opt_pos, opt_tag, opt_text))
				opt_type = loadfile.read(1)
			loadfile.readline()
			statusline = loadfile.readline()
			while (statusline[0] != "-"):
				status_pieces = statusline.split()
				status_num = int(status_pieces[0])
				status_text = " ".join(status_pieces[1:])
				cur_state.status_lines[status_num] = status_text
				cur_state.status_list.append(status_num)
				statusline = loadfile.readline()
			globline = loadfile.readline()
			while (globline[0] != "-"):
				glob_pieces = globline.split("=")
				gname = glob_pieces[0].strip()
				gval = glob_pieces[1].strip()
				cur_state.gvars[gname] = int(gval)
				globline = loadfile.readline()
			stateline = loadfile.readline()
			while (stateline[0] != "-"):
				state_pieces = stateline.split("=")
				loc_pieces = state_pieces[0].split(".")
				sloc = loc_pieces[0].strip()
				svar = loc_pieces[1].strip()
				sval = int(state_pieces[1].strip())
				
				# if there are already state variables for that variable tag (hence an existing dictionary)
				# then just update the new location-value, otherwise create a new dictionary for that
				# variable
			
				if (cur_state.svars.has_key(svar)):
					cur_state.svars[svar][sloc] = int(sval)
				else:
					cur_state.svars[svar] = {sloc:int(sval)}
				stateline = loadfile.readline()
			utimerline = loadfile.readline()
			while (utimerline[0] != "-"):
				utimer_pieces = utimerline.split("=")
				uname = utimer_pieces[0].strip()
				uval = utimer_pieces[1].strip()
				cur_state.up_timers[uname] = int(uval)
				utimerline = loadfile.readline()
			dtimerline = loadfile.readline()
			while (dtimerline[0] != "-"):
				dtimer_pieces = dtimerline.split("=")
				dname = dtimer_pieces[0].strip()
				dval = dtimer_pieces[1].strip()
				cur_state.down_timers[dname] = int(dval)
				dtimerline = loadfile.readline()
			enumline = loadfile.readline()
			while (enumline[0] != "-"):
				enum_pieces = enumline.split()
				enum_num = int(enum_pieces[0])
				enum_tag = enum_pieces[-1][1:-1]
				if (enum_names.has_key(enum_tag)):
					cur_state.enum_inv[enum_tag] = enum_num
				else:
					WriteText('Enumerated object tag "' + enum_tag + '" not recognized, skipping over this line.')
				enumline = loadfile.readline()
			regline = loadfile.readline()
			while (regline[0] != "-"):
				reg_pieces = regline.split()
				reg_tag = reg_pieces[-1][1:-1]
				if (reg_names.has_key(reg_tag)):
					cur_state.reg_inv[reg_tag] = 1
				else:
					WriteText('Regular object tag "' + reg_tag + '" not recognized, skipping over this line.')
				regline = loadfile.readline()
			cur_state.act_image = loadfile.readline().strip()
			cur_state.area_image = cur_state.act_image
			loadfile.close()
			self.NewGameData()
			WriteText('------GAME "' + result + '" LOADED------')
			next_jump = cur_state.location
			if (flag == 1):
				# (flag == 1): just loaded a quick-save on starting program, so copy current state
				# back all the way (on normal loads, might want to back up, cancelling the load)
				last_state.load(cur_state)
				last2_state.load(cur_state)
				last3_state.load(cur_state)
				last4_state.load(cur_state)
				last5_state.load(cur_state)
				last6_state.load(cur_state)
				last7_state.load(cur_state)
				last8_state.load(cur_state)
				last9_state.load(cur_state)
				last10_state.load(cur_state)
			else:
				last10_state.load(last9_state)
				last9_state.load(last8_state)
				last8_state.load(last7_state)
				last7_state.load(last6_state)
				last6_state.load(last5_state)
				last5_state.load(last4_state)
				last4_state.load(last3_state)
				last3_state.load(last2_state)
				last2_state.load(last_state)
				last_state.load(temp_state)
			all_text = self.TextComposite()
			self.SetMainWindow(all_text)
			self.TryLoadConfig()
			self.RefreshStatus()
			self.RefreshLP()
			self.nb.SetSelection(1)
			self.UpdateInvWindow(0)
			cur_directory = cur_state.directory
			return 1
		except "fail":
			cur_state.load(temp_state)
			if flag:
				WriteText('Quick load cancelled.')
			reg_names = copy.deepcopy(temp_reg_names)
			enum_names = copy.deepcopy(temp_enum_names)
			reg_namekeys = copy.deepcopy(temp_reg_namekeys)
			enum_namekeys = copy.deepcopy(temp_enum_namekeys)
			return 0
		except:
			cur_state.load(temp_state)
			if flag:
				if (found_file == 1):
					WriteText('"' + os.path.join(main_directory, 'quick_save.ags') + '" found, but loading was unsuccessful -- syntax error(s) in file.')
			else:
				if (found_file == 0):
					WriteText('IO Error: unable to find file to load that state.')
				else:
					WriteText('Syntax error in saved state file, unable to load.')
			reg_names = copy.deepcopy(temp_reg_names)
			enum_names = copy.deepcopy(temp_enum_names)
			reg_namekeys = copy.deepcopy(temp_reg_namekeys)
			enum_namekeys = copy.deepcopy(temp_enum_namekeys)
			return 0
	
	def TrySaveConfig(self):
		""" Executed when any changes are made to the options whose settings are stored
			in a configuration file, or when changes are made to masked
			variable/objects/timers.  If a game is currently being played, the changes are
			written to "config.agc" in that game's directory.  Otherwise, they're written
			to "config.agc" in Rex's directory. """

		if (cur_state.directory == ''):
			self.SaveConfig(os.path.join(main_directory, 'config.agc'))
		else:
			self.SaveConfig(os.path.join(cur_state.directory, 'config.agc'))
			
	def SaveConfig(self, result):
		""" Saves the user's preferences for masking, narrating, logging, and word statistics
			compilation options.  It executes automatically when any changes to these options
			are made or when anything is masked or unmasked through the editor. """

		global gvarmasked, svarmasked, robjmasked, eobjmasked, timermasked, develop
		global narrate, narrate_filename, nar_area
		global playername, panellog, textlog, log_filename, log_lines, min_col_width
		global wordstats, ws_filename
		global hof_track, hall_filename
		
		confile = open(result, 'w')
		confile.write(playername + "\n")
		confile.write("---Masked Global Variables---\n")
		g_keys = gvarmasked.keys()
		g_keys.sort()
		for g_var in g_keys:
			confile.write(g_var)
			confile.write("\n")
		confile.write("---Masked State Variables---\n")
		s_loc_keys = svarmasked.keys()
		s_loc_keys.sort()
		for s_loc in s_loc_keys:
			s_var_keys = svarmasked[s_loc].keys()
			s_var_keys.sort()
			for s_var in s_var_keys:
				confile.write(s_loc + '.' + s_var)
				confile.write("\n")
		confile.write("---Masked Regular Objects---\n")
		reg_keys = robjmasked.keys()
		reg_keys.sort()
		for the_reg in reg_keys:
			confile.write(the_reg)
			confile.write("\n")
		confile.write("---Masked Enumerated Objects---\n")
		enum_keys = eobjmasked.keys()
		enum_keys.sort()
		for the_enum in enum_keys:
			confile.write(the_enum)
			confile.write("\n")
		confile.write("---Masked Timers---\n")
		timer_keys = timermasked.keys()
		timer_keys.sort()
		for the_timer in timer_keys:
			confile.write(the_timer)
			confile.write("\n")
		confile.write("---Options: (1=log to Hall of Fame) / (name of Hall of Fame file) / (1=log to panel) / (1=log to external file) / (name of external log file) / (1=log all parsed lines) / (0=no narration, 1=narrate as text, 2=narrate as HTML) / (name of narration file) / (1=narrate area text, too) / (1=development mode on) / (1=word statistics compilation on) / (name of word statistics file) / (minimum width of inventory column) ---\n")
		confile.write(str(hof_track))
		confile.write("\n")
		confile.write(hall_filename)
		confile.write("\n")
		confile.write(str(panellog))
		confile.write("\n")
		confile.write(str(textlog))
		confile.write("\n")
		confile.write(log_filename)
		confile.write("\n")
		confile.write(str(log_lines))
		confile.write("\n")
		confile.write(str(narrate))
		confile.write("\n")
		confile.write(narrate_filename)
		confile.write("\n")
		confile.write(str(nar_area))
		confile.write("\n")
		confile.write(str(develop))
		confile.write("\n")
		confile.write(str(wordstats))
		confile.write("\n")
		confile.write(ws_filename)
		confile.write("\n")
		confile.write(str(min_col_width))
		confile.write("\n")
		confile.write("------")
		confile.close()
			

	def SaveState(self, result):
		""" Executes when the game state is saved (usually by the player choosing
			to save or quick-save, but also upon an automatic quick-save when
			the program is exited through a menu command ).

			In the case the user tries to save from a "losing" screen,
			it actually saves the settings in "last_state" to avoid
			troubling scenarios such as saving, then restarting Rex
			and having no "good" state to back up to.  For that reason,
			we back up cur_state first. """
	
		temp_state.load(cur_state)
		if ((temp_state.wl_state != 0) and (temp_state.wl_state != 3)):
			cur_state.load(last_state)
		try:
			savefile = open(result, 'w')
			savefile.write("Game Directory = ")
			savefile.write(cur_state.directory)
			savefile.write("\nMove Count = ")
			savefile.write(str(cur_state.moves))
			savefile.write("\nWin / Loss State = ")
			savefile.write(str(cur_state.wl_state))
			savefile.write("\nRandom Number Seed = ")
			savefile.write(str(cur_state.random_seed))
			savefile.write("\nLocation = ")
			savefile.write(cur_state.location)
			savefile.write("\nVisible Constant Option Count = ")
			savefile.write(str(cur_state.visible_opt_count))
			savefile.write("\n---Action Text:\n")
			savefile.write(cur_state.act_text)
			savefile.write("\n---Area Text:\n")
			savefile.write(cur_state.area_text)
			savefile.write("\n---Constant Options:\n")
			for opt in cur_state.const_options:
				savefile.write("@ " + opt.text)
				savefile.write("\n\tLocation = ")
				savefile.write(opt.location)
				savefile.write("\n\tTag = ")
				savefile.write(str(opt.tag))
				savefile.write("\n\tFile Position = ")
				savefile.write(str(opt.position))
				savefile.write("\n\tVisible = ")
				savefile.write(str(opt.visible))
				savefile.write("\n")
			savefile.write("---Movement Options:\n")
			for opt in cur_state.move_options:
				savefile.write("^ " + opt.text)
				savefile.write("\n\tLocation = ")
				savefile.write(opt.location)
				savefile.write("\n\tTag = ")
				savefile.write(str(opt.tag))
				savefile.write("\n\tFile Position = ")
				savefile.write(str(opt.position))
				savefile.write("\n")
			savefile.write("---Action Options:\n")
			for opt in cur_state.act_options:
				savefile.write("* " + opt.text)
				savefile.write("\n\tLocation = ")
				savefile.write(opt.location)
				savefile.write("\n\tTag = ")
				savefile.write(str(opt.tag))
				savefile.write("\n\tFile Position = ")
				savefile.write(str(opt.position))
				savefile.write("\n")
			savefile.write("---Status Lines:\n")
			for stat in cur_state.status_list:
				savefile.write(str(stat) + " " + cur_state.status_lines[stat] + "\n")
			savefile.write("---Global Variables:\n")
			g_list = cur_state.gvars.keys()
			g_list.sort()
			for gname in g_list:
				savefile.write(gname + " = ")
				savefile.write(str(cur_state.gvars[gname]))
				savefile.write("\n")
			savefile.write("---State Variables:\n")
			svar_list = []
			var_keys = cur_state.svars.keys()
			for var_key in var_keys:
				loc_keys = cur_state.svars[var_key].keys()
				for loc_key in loc_keys:
					svar_list.append([loc_key, var_key])
			svar_list.sort(lambda a, b: ((a[0]>b[0]) or (b[0]==a[0] and a[1]>b[1])) * 2 - 1)
			for svar_pair in svar_list:
				savefile.write(svar_pair[0] + "." + svar_pair[1] + " = " + str(cur_state.svars[svar_pair[1]][svar_pair[0]]) + "\n")
			savefile.write("---Upward-Counting Timers:\n")
			timer_list = cur_state.up_timers.keys()
			timer_list.sort()
			for timer in timer_list:
				savefile.write(timer + " = " + str(cur_state.up_timers[timer]) + "\n")
			savefile.write("---Downward-Counting Timers:\n")
			timer_list = cur_state.down_timers.keys()
			timer_list.sort()
			for timer in timer_list:
				savefile.write(timer + " = " + str(cur_state.down_timers[timer]) + "\n")
			savefile.write("---Enumerated Inventory:\n")
			for tag in enum_namekeys:
				if (cur_state.enum_inv.has_key(tag)):
					enum_val = cur_state.enum_inv[tag]
					savefile.write(str(enum_val) + " ")
					if (enum_val == 1):
						savefile.write(enum_names[tag][1])
					else:
						savefile.write(enum_names[tag][0])
					savefile.write(" (" + tag + ")\n")
			savefile.write("---Regular Inventory:\n")
			for tag in reg_namekeys:
				if (cur_state.reg_inv.has_key(tag)):
					savefile.write(reg_names[tag] + " (" + tag + ")\n")
			savefile.write("---Image:\n")
			if (cur_state.act_image != ''):
				savefile.write(cur_state.act_image)
			else:
				savefile.write(cur_state.area_image)
			savefile.close()
			WriteText('-----GAME "' + result + '" SAVED-----')
			self.nb.SetSelection(1)
		except ValueError:
			pass
		if ((temp_state.wl_state != 0) and (temp_state.wl_state != 3)):
			cur_state.load(temp_state)

	def BackState(self, event):
		""" Executed when the user backs up, either from the menu command or
			using the button provided on a "loss" screen. """

		global next_jump
		global moved_back
		global area_list, act_list
		global cur_directory
		
		if (cur_state.directory == ''):
			dlg = wx.MessageDialog(self, "No game has been started.",
					"Error", wx.OK | wx.ICON_INFORMATION)
			dlg.ShowModal()
			dlg.Destroy()
		elif (last_state.directory != ''):
			try:
				death_back = (cur_state.wl_state == 1)
				temp_state.load(cur_state)
				reimport = (cur_state.directory != last_state.directory)
				cur_state.load(last_state)
				if (reimport):
					self.NewGameData()
					cur_directory = cur_state.directory
				last_state.load(last2_state)
				last2_state.load(last3_state)
				last3_state.load(last4_state)
				last4_state.load(last5_state)
				last5_state.load(last6_state)
				last6_state.load(last7_state)
				last7_state.load(last8_state)
				last8_state.load(last9_state)
				last9_state.load(last10_state)
				if reimport:
					self.SetObjects(cur_state.directory)
					self.TryLoadConfig()
					self.RefreshStatus()
				act_list = [cur_state.act_text]
				area_list = [cur_state.area_text]
				next_jump = cur_state.location
				moved_back = 1
				all_text = self.TextComposite()
				self.SetMainWindow(all_text)
				self.RefreshLP()
				self.UpdateInvWindow(0)
				self.nb.SetSelection(1)
				if (death_back):
					WriteText('------DIED, UNDID MOVE------')
					if (narrate == 1):
						NarText('------DIED, UNDID MOVE------')
					if (narrate == 2):
						NarText('<b>------DIED, UNDID MOVE------</b><br>')
				else:
					WriteText('------BACKED UP 1 MOVE------')
					if (narrate == 1):
						NarText('------BACKED UP 1 MOVE------')
					if (narrate == 2):
						NarText('<b>------BACKED UP 1 MOVE------</b><br>')
			except:
				cur_state.load(temp_state)
				WriteText('Unable to back up from the last move.')

	def ResetState(self, event):
		""" Executed when the user opts to reset the current game, calling
			the same InitGame used when a new game is set. """

		if (cur_state.directory == ''):
			dlg = wx.MessageDialog(self, "No game has been started.",
					"Error", wx.OK | wx.ICON_INFORMATION)
			dlg.ShowModal()
			dlg.Destroy()
		else:
			self.InitGame(os.path.join(cur_state.directory, 'start.sma'))
		
	def Reenter(self, event):
		""" Executed when the user selects the "Re-enter location" menu command.
			The location code is parsed again to regenerate the area text
			and options that may have changed since the file was last parsed,
			but the other steps normally involved in making a move in Smash
			(updating timers and the move count, parsing "repeatpre.sma" and
			"repeatpost.sma", etc.) are not executed. """

		global just_jumped
		
		if (cur_state.directory == ''):
			dlg = wx.MessageDialog(self, "No game has been started.",
					"Error", wx.OK | wx.ICON_INFORMATION)
			dlg.ShowModal()
			dlg.Destroy()
		else:
			if ((cur_state.wl_state == 0) or (cur_state.wl_state == 3)):
				cur_state.wl_state = 0
				cur_state.act_text = ''
				act_list = []
				cur_state.move_options = []
				cur_state.act_options = []
				cur_state.act_image = ''
				cur_state.area_image = ''
				WriteText('------RE-ENTERED LOCATION------')
				try:
					newloc_parse()
				except ParsingException, message:
					WriteText('PARSING ERROR:\n' + str(message))
					cur_state.wl_state = -1
					just_jumped = 0
				all_text = self.TextComposite()
				self.SetMainWindow(all_text)
		
	def ActText(self, event):
		""" Executed when the user opts to hide the area text
			temporarily, for whatever reason -- only lasts
			until the next move. """
		global hide

		hide = 1
		all_text = self.TextComposite()
		self.SetMainWindow(all_text)
		
	def AreaText(self, event):
		""" Executed when the user opts to hide the action text
			temporarily, for whatever reason -- only lasts
			until the next move. """
		global hide

		hide = 2
		all_text = self.TextComposite()
		self.SetMainWindow(all_text)
		
	def BothText(self, event):
		""" Executed when the user chooses to switch back
			to showing both text boxes. """
		
		global hide

		hide = 0
		all_text = self.TextComposite()
		self.SetMainWindow(all_text)

	def TimeToQuit(self, event):
		""" Executed when Rex is exited through the menu command.
			Before closing the main window, it performs a
			backup save of the current state into the quick-save
			file. """

		global main_directory
	
		if (cur_state.directory != ''):
			self.SaveState(os.path.join(main_directory, 'quick_save.ags'))
		self.Close(1)

	def OnCloseWindow(self, event):
		""" Executed on shutdown. """

		self.dying = True
		self.window = None
		self.mainmenu = None
		self.Destroy()

	def LicenseShow(self, event):
		""" Handles the event of the user clicking the License button, creating a
			scrolled window with the GNU GPL text in it. """

		global LicenseText

		dlg = wxScrolledMessageDialog(self, LicenseText, "GNU General Public License")
		dlg.ShowModal()
		dlg.Destroy()

#---------------------------------------------------------------------------

class MyApp(wx.App):
	def OnInit(self):
		wx.InitAllImageHandlers()
		frame = SmashFrame(None, -1, "Rex")
		frame.Show()
		return True

#---------------------------------------------------------------------------

# Environment
main_directory = os.path.normpath(os.path.join(os.getcwd(), os.path.dirname(sys.argv[0])))	# constant
cur_directory = main_directory

Start_Text = """<html><body>Rex Copyright 2004 Andrew Walters.  Rex comes
with ABSOLUTELY NO WARRANTY.  This is free software released under version
2 of the GNU General Public License.  You are welcome to
modify and redistribute this program under the terms of the license.
To read the license, open "license.txt", which should have been
included in the distribution of this program, or click here:<br>

<center><wxp class="wxButton" width="150" height="30">
<param name="label" value="License Details">
<param name="id" value=20000></wxp></center>
<hr>"""

Start_Text += '<img src=' + os.path.join(main_directory, 'rex.gif>')
Start_Text += """<font size=+2>Rex</font>

<p>A Smash interpreter and game development program<br>
Version 1.3 -- 7/04/05<br>
Written with Python 2.3, wxPython 2.4<br>
Look and feel based on <i>Adventure Games Live</i> (http://www.rinkworks.com/adventure) by Samuel Stoddard

<p>Rex was written to help develop and playtest games written in Smash, which is largely an
extension of the Adventure Games Live Language (AGLL), both written by Samuel
Stoddard.  To use this program, you'll need the complete set of Smash source
files for a game.  This includes "game.txt", "start.sma", and all ".sma" files corresponding
to location files.  Optionally you may have: "copts.sma", "repeatpre.sma",
"repeatpost.sma", "objects.txt", "eobjects.txt", "objectactions.sma",
"eobjectactions.sma", "win.txt", "lose.txt".  Make sure all of the files
needed for a game -- including all .sma files, the .txt files listed above,
and all image files required for the game -- are all in a common directory.

<p>The author welcomes bug reports or other comments sent to: awalters3@yahoo.com

<p><font size=+1><b>Rex Menu Commands:</b></font> (note that many commands
have keyboard macros enabled)

<p>
	<ul>
	<li><b>Game</b>
		<ul>
		<li><b>Set game:</b> This is where to begin if you're using Rex for the
		first time.  Select this menu command, then browse for a directory containing
		the files for the Smash game you want to test.  Select the "start.sma" file
		and the game will commence.  First the tags and object names of "objects.txt"
		and "eobjects.txt" will be read, and then parsing will begin in "copts.sma",
		or "start.sma" if there is no constant options file.
		<li><b>Reset game:</b> If a Smash game is currently being tested, choosing
		this command will reset the game from the start.  The object names and tags
		will also be refreshed from "objects.txt" and "eobjects.txt".
		<li><b>Re-import inventory (Ctrl - I):</b> This command forces Rex to
		re-import the tags and object names from "objects.txt" and "eobjects.txt".  You
		should use this if you make changes to either of those two files and want
		those changes to take effect in Rex immediately, rather than
		waiting until the next time Rex is started or the current game is reset.
		Bear in mind that anything in the inventory that no longer corresponds to a
		valid tag will be dropped (you will be notified of any cases).
		<li><b>Re-import game.txt (Ctrl - G):</b> This command forces Rex to
		re-import the relevant information in the "game.txt" file in the current game's
		directory -- namely, the game's name and the field data for the game's Hall
		of Fame.  Note that if changes are made to "game.txt" after data is written to
		a Hall of Fame file, the fields may be incompatible the next time data from a
		"w" or "W" command is posted to that Hall of Fame file.  If you want to keep
		your Hall of Fame in one place (rather than having it start a new file), you'll
		need to make sure the changes to "game.txt" are also manually made to the Hall
		of Fame file.
		<li><b>Quit (Ctrl - Q):</b> This command exits Rex.  Note that if a game is in progress,
		the state will automatically be saved in "quick_save.ags" -- in Rex's
		directory, not that of the particular game.  This special save file, if
		detected when Rex is started up, will be automatically loaded.  If Rex is shut down
		in any other way, the quick-save operation will not take place.
		</ul>
	<li><b>State</b>
		<ul>
		<li><b>Save state (Alt - S):</b> Saves your current state to an ".ags" file
		of your choosing.  ".ags" files are plain text, so they can be edited if you
		take care not to break the formatting.  (See <u>External Editing</u>.)  An exception to this is if
		you try to save the state when you have just made a "losing" move or if a parsing error
		has just been encountered.  In these situations, you are given an "Undo" button, and if you
		elect to save, it actually saves the state you were in just before making the losing move or
		encountering the parsing error.
		<li><b>Load state (Alt - L):</b> Loads a previously saved state,
		even if it's for a Smash game other than the one currently being played.  Loading
		a saved state will also re-import the inventory as described above.
		<li><b>Quick save (Ctrl - S):</b> Saves your current state to the special
		"quick_save.ags" file located in Rex's directory.  The advantage to this over
		choosing the normal "Save state" command is that it's faster, as you don't need
		to go through the step of browsing for a file name/directory.
		<li><b>Quick load (Ctrl - L):</b> Loads the state saved in the special
		"quick_save.ags" file located in Rex's directory.  As with "Quick save", it's
		faster, as you don't need to browse for a saved state file.  As with "Load state",
		you can even switch from one Smash game to another immediately.
		<li><b>Back up one move (Ctrl - B):</b> Changes the state to the last previous state.  If you
		accidentally make a move or you didn't intend, or if you want to test another option,
		you can back up and try again.  Rex keeps track of 10 states before your current
		state, so in general you can backtrack to try a different course of action if it
		branched 10 or fewer moves ago.  Note that any states more than 10 states previous
		are lost forever unless you explicitly saved them.  So, for example, if you
		backtrack 7 times, then make 2 moves from there, you are now in a state from
		which you can backtrack only 5 additional times.  Trying to backtrack further than
		this will have no effect.  A couple other notes of interest:
			<ul>
			<li>Editing your current state does not "move forward" to a new state from which
			you can backtrack to immediately before the edit.  Instead, if you make an editing
			change and then backtrack, you are left in the same state you would have been in
			had you backtracked <i>without</i> making the editing change.
			<li>Backtracking works even if some of the 10 previous state changes involved
			loading saved states (even of different Smash games), setting new Smash games, or
			resetting the current game.
			</ul>
		<li><b>Re-enter location (Ctrl - R):</b> Re-parses the code for the current location.  Note
		that only this step of the normal sequence of events on a move takes place.  "repeatpre.sma"
		and "repeatpost.sma" are not parsed, the move number is not incremented, and timers are not
		altered.  This goes along with the idea that "re-entering" a location is not an actual move
		in the game or a new game state -- it is mainly a way of refreshing area text and options
		which you've just changed in the location file.  You can also use this command to complete
		the effect of some editing commands, such as jumping to a new location, to see updated
		text and options.  (Also see <u>Status Panels</u>, <u>Edit State button</u>.)
		</ul>
	<li><b>Display</b>
		<ul>
		<li><b>Show action text only (F1):</b> Hides the "area" text (text appearing in the
		upper-right blue box) temporarily.  Rex will stay in this state until you change it again.
		The primary use of this command is to view the entirety of images which are too large to fit
		in the "action" text box.
		<li><b>Show area text only (F2):</b> Hides the "action" text (text appearing in the
		upper-left pink box) temporarily.  Rex will stay in this state until you change it again.
		The primary use of this command is to view the entirety of images which
		are too large to fit in the "area" text box.
		<li><b>Show both (default) (F3):</b> Restores Rex to showing both the "action" text
		and the "area" text.
		<li><b>Generate AGL-style HTML (F4):</b> Allows you to create a file with HTML code
		corresponding to the current game state (the description text boxes, the options, the
		status line box, and the inventory box).  This can then be viewed separately in a browser
		to get a close approximation of how the game state would look in <i>Adventure Games
		Live</i>.  Normally this command is only needed when you want to check how such things as
		the layout of HTML tables will look in a game intended for AGL, but you could also use
		it to store "snapshots" of various states of the game without the need to use Rex to view them.
		<li><b>Purge log (Ctrl - P):</b> Erases all text from the log in Rex's bottom-left panel,
		if there is any.  This is necessary if you reach the limit of 30,000 characters
		in the log panel -- anything beyond this will not be logged until you purge the
		existing text.  (See <u>Logging</u> below for more information on logging.)
		</ul>
	<li><b>Location</b>
		<ul>
		<li><b>Compute current game's statistics (F7):</b> Makes Rex parse through each line of
		the current game's ".sma" files (excepting "admin.sma"), plus "win.txt" and
		"lose.txt" if either of these are present.  This procedure compiles a count of
		how many of each Smash command you've used and, should you elect to do so in the
		Options tabbed window, creates a file (with a default name of "wordstats.txt") which
		lists all the words used in description text visible to the player during play.
		<li><b>Compute current game's size (F8):</b> Counts the number and total file size of
		all Smash files for the current game, excluding "admin.sma", image files, and any
		Rex-created files such as ".ags" files.
		</ul>
	<li><b>Window</b>
		<ul>
		<li><b>Toggle left panels (F9):</b> Hides or reveals the two panels on the left side of
		the Rex interface.  Hiding these panels allows for more room in the main panel
		during gameplay.
		<li><b>Toggle bottom panels (F10):</b> Hides or reveals the two panels on the bottom of
		the Rex interface.  Hiding these panels allows for more room in the main panel
		during gameplay.
		<li><b>Toggle log/inv/both panels (F11):</b> Cycles through one of three settings: showing
		just the logging (bottom left) panel, showing just the inventory (bottom right) panel, or
		showing both.  Of course, neither is visible unless they are toggled collectively with F10.
		If both of the bottom panels are shown, you can adjust the screen area apportioned to each by
		dragging the dividing sash to the left or right.
		</ul>
	</ul>

<p><font size=+1><b>Main Panel Tabs:</b></font> The main panel, normally the one taking the most display area,
	is the panel in the upper right of Rex's interface.  You can click on one of its tabs to display various
	windows in the main panel.

	<p><ul>
	<li><b>Program Info:</b> This is the tabbed window which displays this information page in Rex.
	<li><b>Game Text:</b> This window is where you will interact with the game.  It looks similar to
	the interface of Adventure Games Live, except the yellow status box and the green
	inventory box have been given their own panels in Rex's interface.
	<li><b>Game State:</b> This window displays all the minute details of the current game state (if
	you've started a game).  Displayed on this page are the current location, the number
	of moves made, the "random number seed" (which is a function of the specific moves made in the game so
	far, and is only used in Smash's pseudo-random number generation algorithm), the state variables,
	the global variables, and the timers.
	<li><b>Game Description:</b> This window displays the data from the current game's "game.txt" file.
	At the top, the game's optional title graphic "graphic.gif" is displayed, if present.  If not, it
	shows the title in text.  Below that, in order, are shown the game's difficulty rating, the game's
	credits, and the game's full synopsis, for whichever are present.  All of these are optional.
	Finally is a description of the Hall of Fame fields' format, if a Hall of Fame has been designated
	for the game at all.
	<li><b>Options:</b> In this window you can select different settings to use in Rex.  Any time
	you make a change in this window, the settings are automatically saved in the appropriate
	configuration file.  This file will always be named "config.agc", but the directory in which
	it will be stored can vary.  If no game is currently in progress, any changes here will be
	considered "global" changes made to the "config.agc" file in Rex's directory.  If a game
	<i>is</i> in progress, the changes will be saved to "config.agc" in that game's directory
	(creating it, if there isn't one already).  When a different Smash game is started up (whether
	by initializing a new game, loading a saved state, or backing up), the corresponding "config.agc"
	file for that game will automatically be loaded in, so you can keep different configuration settings
	for different Smash games with ease.  Tip: if you want to make a new configuration file for a
	single Smash game and not lose the old one, you can handle this by copying the original
	"config.agc" to a file of a different name or path, temporarily, and copying it back to "config.agc"
	when you want to use it again.  And now, the options themselves:
		<ul>
		<li><b>Player Name:</b> This name will be used for entries to the Hall of Fame and no other
		purpose.  Click the "Change Name" button to choose a different name.  A name must have at least
		one non-whitespace character and not contain the "|" character, as "|" is used as the parameter
		separator in "w" and "W" commands.  "Player" is the default player name.
		<li><b>Hall of Fame file:</b> This is the file to which data from "w" and "W" commands will be
		posted, unless there is no Hall of Fame for the game according to the "game.txt" file.  If you
		have multiple Smash games in testing, make sure you've remembered to differentiate each one with
		its own Hall of Fame file, or they'll all compete in posting to one file (probably creating
		incompatibilities because of differing Hall of Fame columns).
		<li><b>Logging:</b> You can choose to have text be logged to either the bottom left (Logging)
		panel or an external file (default "logfile.txt" in Rex's directory), or both.  By default
		any changes to variables, inventory status, and location jumps are logged to the panel, though
		you will need to enable logging to the external file.  Other informative messages, as well as error
		warnings, are also logged.  If you choose to log to an external file, you will be able to browse
		for the directory and file name of the log file.  It is recommended, for clarity, that you use
		separate logging files for separate Smash games.  If you select the option to "Log all parsed lines",
		every single line of Smash code which Rex parses, in order, will show up in the logs (these lines
		will start with ">>>>>").  This can clutter things up quite rapidly, but it is often useful to
		enable this option when you need to track down an elusive bug with your Smash code.
		<li><b>Narration:</b> If you select either of the latter two items in the list box, the action
		text from any move made in a game will be logged to an external file (default "narrate.txt" in Rex's
		directory), following the move number, location, option type, and option text corresponding to the
		option chosen.  As a result, you can obtain a narrative transcript of a gameplay session.  You can
		choose to have the transcript formatted as plain text (replacing &lt;br&gt; tags by newline characters
		and &lt;p&gt; tags by two newline characters) or left with those HTML tags left in.  (It's left to
		you to choose the appropriate file extension on your own.)  As with logging, it is recommended
		that you use separate narration files for separate Smash games.  If "Narrate area text" is
		enabled, the area text on each turn will be logged as well.  If it's being narrated as HTML,
		this text will appear in an indented block at a smaller font, for easier distinction.  If
		it's being narrated as text, the area text will be enclosed in curly braces.
		<li><b>Development Mode:</b> By default, development mode is enabled.  In this mode, you can see
		details during testing which will not be available to players:
			<ul>
			<li>When examining items, you will see the objects' names and tags, as opposed to the names and
			order in the inventory.
			<li>An "Edit State" will be shown in the top left panel.  (See <u>Edit State button</u> below.)
			<li>A purple box listing variables, timers, and items which are "masked" from being logged is
			shown.  (See <u>Masking box</u> below.)
			<li>The current location will be displayed in a small blue box above the yellow status box in
			the bottom left panel.
			<li>Status lines' text will be preceded by the corresponding numbers which determine the lines'
			order.
			</ul>
		<li><b>Word Statistics:</b> If this box is checked, Rex will compile a list of all the words tallied up
		when you run the game statistics menu command and output it to a file of your choosing, listing each
		word encountered alongside the number of occurrences in your game.
		</ul>
	<li><b>Location File:</b> In this window you can see the code for the current location (if a game
	has been started).  It will refresh automatically when appropriate.
	</ul>

<p><font size=+1><b>Edit/Masking Panel:</b></font> The panel in the upper left of Rex's interface.

	<p><ul>
	<li><b>Edit State button:</b> If development mode is on, you will see this button at the top of the panel.
	When you select it, you will be able to input a line of Smash code to be parsed directly.
	In this way you can directly edit the value of a global or state variable, a timer, or the inventory.
	Other Smash commands which do not take succeeding code blocks (in other words, commands which have
	their full effect on one line) can be input here as well.  In addition, there are three Rex-only
	commands which can be input:
		<ul>
		<li><b>x</b> -- when followed by the name of a location, this command tells Rex to parse every line of
		code in that file.  A statistics report is then generated and logged, counting the number of each type
		of Smash command found, the total number of words of description text, and the total size of these words
		in bytes (for description text commands spanning multiple lines, a trailing space is counted at the end
		of all but the last line).  If any parsing errors occur, the process will stop part way through.  See
		the <u>Compute current game's statistics</u> command under <u>Rex Menu Commands, Location</u>, which will
		do a similar error-checking and statistics computation for the entire game at once.
			<ul>
			<li>Examples: "x townsquare" parses all of "townsquare.sma".  "x" by itself parses the current
			location's file.
			</ul>
		<li><b>m</b> -- this command allows you to "mask" certain variables, timers, or inventory objects.  If
		something is masked, changes involving it will not be logged, regardless of the logging options selected
		(see <u>Main Panel Tabs</u> in <u>Options</u>).  If anything is currently being masked, it will be listed
		in the purple box under the Edit State button, as long as development mode is on.  The syntax for this
		command is as shown below:
			<ul>
			<li>Global variables: "m v variable_name"
			<li>State variables: "m s location_name.variable_name"
			<li>Timers: "m t timer_name"
			<li>Regular Inventory: "m a object_tag"
			<li>Enumerated Inventory: "m e eobject_tag"
			</ul>
		<li><b>M</b> -- this command "unmasks" anything which has been masked with the "m" command.  The syntax
		is the same.  When either the "m" or the "M" command is used, any changes are saved to the configuration
		file ("config.agc" in Rex's directory), which is at this point created if not already present.
		</ul>
	<li><b>Masking box:</b> If present, this purple box lists all variables, timers, or inventory objects currently
	being masked (see above).  Any changes to this are automatically saved to "config.agc" in Rex's directory, and
	they are unchanged even if a different Smash game is started.
	</ul>

<p><font size=+1><b>Status Panel:</b></font> The panel in the middle left of Rex's interface.

	<p><ul>
	<li><b>Location box:</b> If development mode is on, there will be a small blue box at the top of the bottom left
	panel.  This box will list the current location, if any.
	<li><b>Status box:</b> The status box is the yellow box in the bottom left panel.  It works just as in the
	<i>Adventure Games Live</i> engine -- only the position in the interface is different.  If development mode is on,
	you will see the status line numbers in addition to the actual text of the status lines.
	</ul>

<p><font size=+1><b>Logging Panel:</b></font> The panel in the bottom left of Rex's interface.

	<blockquote>This is where changes to the game state are logged, if you've enabled the option to do so in the
	Options tab of the main panel.  (See <u>Logging</u> in <u>Options</u> in <u>Main Panel Tabs</u> above.)
	</blockquote>

<p><font size=+1><b>Inventory Panel:</b></font> The panel in the bottom right of Rex's interface.

	<blockquote>This is where your inventory is shown if you have a game in progress.  It will update automatically
	whenever you set a new game, reset your game, load a state, back up, or make a move in which your inventory
	is changed.  Despite all these possibilities, this window does not need to refresh on most moves, for most Smash
	games.  Note that the inventory panel does not have individual buttons for examining inventory items
	(this may be added into a future version of Rex).  Instead, invoking inventory items is handled through
	the button at the top of the green box.  This button will pop up a list box showing the items you have.
	If development mode is on, you will be able to see the items' tags as well.  Selecting an item from the list
	box will begin a Smash move consisting of invoking that item.
	</blockquote>

<p><font size=+1><b>External Editing:</b></font>

	<blockquote>Saved state files are given the ".ags" extension, but are plain text.  As such, you can edit these
	files in a text editor if you find it more convenient than making edits one line at a time from within Rex.
	You must be careful, however, not to ruin the precise file format in the process.  When a saved state file is
	loaded, Rex reads it in one line at a time, using lines starting with "---" as demarcators.  These lines should
	not be removed or reordered.  You can change either description text string, change either current image file,
	or add/remove options, inventory items, status lines, variables, or timers, but be sure to use the same format
	and the same number of lines as similar entries.  If the file format is ruined, Rex will not be able to properly
	load that state.
	
	<p>If you make any option or masking changes within Rex, a configuration file "config.agc" is created/altered in
	Rex's directory or the current game's directory, as appropriate.  (See <u>Options</u> in <u>Main Panel
	Tabs</u> above.)  Like saved state files, it uses a precise format which
	must be preserved in order to be loaded properly.  You can freely add or remove items to be masked here, but
	do not alter the "---" lines.  The thirteen lines above the final "------" are data pertaining to options
	which are set within Rex, and no lines should be added, removed, or reordered here.

	<p>Finally, there's the syntax of the Hall of Fame files.  The first few lines, up to but not including the first
	line with "|" by itself, present the data for the fields (columns) in the Hall of Fame.  The string before the
	first "|" of a line is the name of the column (for example, "Name").  Then comes a number from 0 to 4, the meaning
	of each being:
		<ul>
		<li>0: this column sorts numerical entries from largest to smallest (corresponding to "n&lt;" or a blank,
		in "game.txt" syntax) -- the default sorting method for a field.
		<li>1: this column sorts numerical entries from smallest to largest (corresponding to "n&gt;").
		<li>2: this column sorts alphabetic entries from largest to smallest (corresponding to "a&lt;").
		<li>3: this column sorts alphabetic entries from smallest to largest (corresponding to "a&gt;").
		<li>4: this column sorts customized entries in the order presented.
		</ul>
	If the number is a 4, there will be a further "|" and then a list of the valid entries for this column,
	separated by "|", in the order which determines their ranking when the Hall of Fame is sorted by this column.
	</blockquote>

<p><font size=+1><b>Troubleshooting / Compatibility with <i>Adventure Games Live</i>:</b></font>

	<blockquote>Although a lot of effort has been put into ensuring Rex will behave as closely as possible to the
	Smash engine behind <i>Adventure Games Live</i> (http://www.rinkworks.com/adventure), there may be obscure
	situations in which things differ.  A game author who ultimately intends his game to appear on that site would be
	advised to test at least one complete run-through of the game in the AGL Smash engine to make sure everything
	works properly, especially if some fancy, non-standard coding tricks have been used.  The following is a list
	of known differences, none of which should pose a problem to a game
	developer.  Anything else that <i>doesn't</i> work in one Smash engine/interpreter but works in the other
	should be considered a bug.<p>
		<ul>
		<li>Images extending beyond the boundaries of the action or area text boxes: as this is just a display
		problem with HTML tables in Rex, it's nothing to worry about from a development perspective.  You can use
		the menu commands under "Display" to temporarily allocate the entire width to one text box or the other
		to be able to see an image in its entirety.
		<li>&lt; signs and succeeding text disappearing from the display: this is another problem with the
		HTML display in Rex.  Internally, the &lt; is preserved properly.  To work around the display issue,
		you can replace any "&lt;" not meant as part of an HTML tag, but honestly meant as a "less than" sign,
		by "lt;" preceded by an ampersand in your Smash code, and it will then display correctly.
		</ul>
	</blockquote>
"""


LicenseText = """		    GNU GENERAL PUBLIC LICENSE
		       Version 2, June 1991

 Copyright (C) 1989, 1991 Free Software Foundation, Inc.
                       59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 Everyone is permitted to copy and distribute verbatim copies
 of this license document, but changing it is not allowed.

			    Preamble

  The licenses for most software are designed to take away your
freedom to share and change it.  By contrast, the GNU General Public
License is intended to guarantee your freedom to share and change free
software--to make sure the software is free for all its users.  This
General Public License applies to most of the Free Software
Foundation's software and to any other program whose authors commit to
using it.  (Some other Free Software Foundation software is covered by
the GNU Library General Public License instead.)  You can apply it to
your programs, too.

  When we speak of free software, we are referring to freedom, not
price.  Our General Public Licenses are designed to make sure that you
have the freedom to distribute copies of free software (and charge for
this service if you wish), that you receive source code or can get it
if you want it, that you can change the software or use pieces of it
in new free programs; and that you know you can do these things.

  To protect your rights, we need to make restrictions that forbid
anyone to deny you these rights or to ask you to surrender the rights.
These restrictions translate to certain responsibilities for you if you
distribute copies of the software, or if you modify it.

  For example, if you distribute copies of such a program, whether
gratis or for a fee, you must give the recipients all the rights that
you have.  You must make sure that they, too, receive or can get the
source code.  And you must show them these terms so they know their
rights.

  We protect your rights with two steps: (1) copyright the software, and
(2) offer you this license which gives you legal permission to copy,
distribute and/or modify the software.

  Also, for each author's protection and ours, we want to make certain
that everyone understands that there is no warranty for this free
software.  If the software is modified by someone else and passed on, we
want its recipients to know that what they have is not the original, so
that any problems introduced by others will not reflect on the original
authors' reputations.

  Finally, any free program is threatened constantly by software
patents.  We wish to avoid the danger that redistributors of a free
program will individually obtain patent licenses, in effect making the
program proprietary.  To prevent this, we have made it clear that any
patent must be licensed for everyone's free use or not licensed at all.

  The precise terms and conditions for copying, distribution and
modification follow.

		    GNU GENERAL PUBLIC LICENSE
   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION

  0. This License applies to any program or other work which contains
a notice placed by the copyright holder saying it may be distributed
under the terms of this General Public License.  The "Program", below,
refers to any such program or work, and a "work based on the Program"
means either the Program or any derivative work under copyright law:
that is to say, a work containing the Program or a portion of it,
either verbatim or with modifications and/or translated into another
language.  (Hereinafter, translation is included without limitation in
the term "modification".)  Each licensee is addressed as "you".

Activities other than copying, distribution and modification are not
covered by this License; they are outside its scope.  The act of
running the Program is not restricted, and the output from the Program
is covered only if its contents constitute a work based on the
Program (independent of having been made by running the Program).
Whether that is true depends on what the Program does.

  1. You may copy and distribute verbatim copies of the Program's
source code as you receive it, in any medium, provided that you
conspicuously and appropriately publish on each copy an appropriate
copyright notice and disclaimer of warranty; keep intact all the
notices that refer to this License and to the absence of any warranty;
and give any other recipients of the Program a copy of this License
along with the Program.

You may charge a fee for the physical act of transferring a copy, and
you may at your option offer warranty protection in exchange for a fee.

  2. You may modify your copy or copies of the Program or any portion
of it, thus forming a work based on the Program, and copy and
distribute such modifications or work under the terms of Section 1
above, provided that you also meet all of these conditions:

    a) You must cause the modified files to carry prominent notices
    stating that you changed the files and the date of any change.

    b) You must cause any work that you distribute or publish, that in
    whole or in part contains or is derived from the Program or any
    part thereof, to be licensed as a whole at no charge to all third
    parties under the terms of this License.

    c) If the modified program normally reads commands interactively
    when run, you must cause it, when started running for such
    interactive use in the most ordinary way, to print or display an
    announcement including an appropriate copyright notice and a
    notice that there is no warranty (or else, saying that you provide
    a warranty) and that users may redistribute the program under
    these conditions, and telling the user how to view a copy of this
    License.  (Exception: if the Program itself is interactive but
    does not normally print such an announcement, your work based on
    the Program is not required to print an announcement.)

These requirements apply to the modified work as a whole.  If
identifiable sections of that work are not derived from the Program,
and can be reasonably considered independent and separate works in
themselves, then this License, and its terms, do not apply to those
sections when you distribute them as separate works.  But when you
distribute the same sections as part of a whole which is a work based
on the Program, the distribution of the whole must be on the terms of
this License, whose permissions for other licensees extend to the
entire whole, and thus to each and every part regardless of who wrote it.

Thus, it is not the intent of this section to claim rights or contest
your rights to work written entirely by you; rather, the intent is to
exercise the right to control the distribution of derivative or
collective works based on the Program.

In addition, mere aggregation of another work not based on the Program
with the Program (or with a work based on the Program) on a volume of
a storage or distribution medium does not bring the other work under
the scope of this License.

  3. You may copy and distribute the Program (or a work based on it,
under Section 2) in object code or executable form under the terms of
Sections 1 and 2 above provided that you also do one of the following:

    a) Accompany it with the complete corresponding machine-readable
    source code, which must be distributed under the terms of Sections
    1 and 2 above on a medium customarily used for software interchange; or,

    b) Accompany it with a written offer, valid for at least three
    years, to give any third party, for a charge no more than your
    cost of physically performing source distribution, a complete
    machine-readable copy of the corresponding source code, to be
    distributed under the terms of Sections 1 and 2 above on a medium
    customarily used for software interchange; or,

    c) Accompany it with the information you received as to the offer
    to distribute corresponding source code.  (This alternative is
    allowed only for noncommercial distribution and only if you
    received the program in object code or executable form with such
    an offer, in accord with Subsection b above.)

The source code for a work means the preferred form of the work for
making modifications to it.  For an executable work, complete source
code means all the source code for all modules it contains, plus any
associated interface definition files, plus the scripts used to
control compilation and installation of the executable.  However, as a
special exception, the source code distributed need not include
anything that is normally distributed (in either source or binary
form) with the major components (compiler, kernel, and so on) of the
operating system on which the executable runs, unless that component
itself accompanies the executable.

If distribution of executable or object code is made by offering
access to copy from a designated place, then offering equivalent
access to copy the source code from the same place counts as
distribution of the source code, even though third parties are not
compelled to copy the source along with the object code.

  4. You may not copy, modify, sublicense, or distribute the Program
except as expressly provided under this License.  Any attempt
otherwise to copy, modify, sublicense or distribute the Program is
void, and will automatically terminate your rights under this License.
However, parties who have received copies, or rights, from you under
this License will not have their licenses terminated so long as such
parties remain in full compliance.

  5. You are not required to accept this License, since you have not
signed it.  However, nothing else grants you permission to modify or
distribute the Program or its derivative works.  These actions are
prohibited by law if you do not accept this License.  Therefore, by
modifying or distributing the Program (or any work based on the
Program), you indicate your acceptance of this License to do so, and
all its terms and conditions for copying, distributing or modifying
the Program or works based on it.

  6. Each time you redistribute the Program (or any work based on the
Program), the recipient automatically receives a license from the
original licensor to copy, distribute or modify the Program subject to
these terms and conditions.  You may not impose any further
restrictions on the recipients' exercise of the rights granted herein.
You are not responsible for enforcing compliance by third parties to
this License.

  7. If, as a consequence of a court judgment or allegation of patent
infringement or for any other reason (not limited to patent issues),
conditions are imposed on you (whether by court order, agreement or
otherwise) that contradict the conditions of this License, they do not
excuse you from the conditions of this License.  If you cannot
distribute so as to satisfy simultaneously your obligations under this
License and any other pertinent obligations, then as a consequence you
may not distribute the Program at all.  For example, if a patent
license would not permit royalty-free redistribution of the Program by
all those who receive copies directly or indirectly through you, then
the only way you could satisfy both it and this License would be to
refrain entirely from distribution of the Program.

If any portion of this section is held invalid or unenforceable under
any particular circumstance, the balance of the section is intended to
apply and the section as a whole is intended to apply in other
circumstances.

It is not the purpose of this section to induce you to infringe any
patents or other property right claims or to contest validity of any
such claims; this section has the sole purpose of protecting the
integrity of the free software distribution system, which is
implemented by public license practices.  Many people have made
generous contributions to the wide range of software distributed
through that system in reliance on consistent application of that
system; it is up to the author/donor to decide if he or she is willing
to distribute software through any other system and a licensee cannot
impose that choice.

This section is intended to make thoroughly clear what is believed to
be a consequence of the rest of this License.

  8. If the distribution and/or use of the Program is restricted in
certain countries either by patents or by copyrighted interfaces, the
original copyright holder who places the Program under this License
may add an explicit geographical distribution limitation excluding
those countries, so that distribution is permitted only in or among
countries not thus excluded.  In such case, this License incorporates
the limitation as if written in the body of this License.

  9. The Free Software Foundation may publish revised and/or new versions
of the General Public License from time to time.  Such new versions will
be similar in spirit to the present version, but may differ in detail to
address new problems or concerns.

Each version is given a distinguishing version number.  If the Program
specifies a version number of this License which applies to it and "any
later version", you have the option of following the terms and conditions
either of that version or of any later version published by the Free
Software Foundation.  If the Program does not specify a version number of
this License, you may choose any version ever published by the Free Software
Foundation.

  10. If you wish to incorporate parts of the Program into other free
programs whose distribution conditions are different, write to the author
to ask for permission.  For software which is copyrighted by the Free
Software Foundation, write to the Free Software Foundation; we sometimes
make exceptions for this.  Our decision will be guided by the two goals
of preserving the free status of all derivatives of our free software and
of promoting the sharing and reuse of software generally.

			    NO WARRANTY

  11. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED
OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS
TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE
PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,
REPAIR OR CORRECTION.

  12. IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,
INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING
OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED
TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY
YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER
PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGES.

		     END OF TERMS AND CONDITIONS

	    How to Apply These Terms to Your New Programs

  If you develop a new program, and you want it to be of the greatest
possible use to the public, the best way to achieve this is to make it
free software which everyone can redistribute and change under these terms.

  To do so, attach the following notices to the program.  It is safest
to attach them to the start of each source file to most effectively
convey the exclusion of warranty; and each file should have at least
the "copyright" line and a pointer to where the full notice is found.

    <one line to give the program's name and a brief idea of what it does.>
    Copyright (C) <year>  <name of author>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


Also add information on how to contact you by electronic and paper mail.

If the program is interactive, make it output a short notice like this
when it starts in an interactive mode:

    Gnomovision version 69, Copyright (C) year name of author
    Gnomovision comes with ABSOLUTELY NO WARRANTY; for details type `show w'.
    This is free software, and you are welcome to redistribute it
    under certain conditions; type `show c' for details.

The hypothetical commands `show w' and `show c' should show the appropriate
parts of the General Public License.  Of course, the commands you use may
be called something other than `show w' and `show c'; they could even be
mouse-clicks or menu items--whatever suits your program.

You should also get your employer (if you work as a programmer) or your
school, if any, to sign a "copyright disclaimer" for the program, if
necessary.  Here is a sample; alter the names:

  Yoyodyne, Inc., hereby disclaims all copyright interest in the program
  `Gnomovision' (which makes passes at compilers) written by James Hacker.

  <signature of Ty Coon>, 1 April 1989
  Ty Coon, President of Vice

This General Public License does not permit incorporating your program into
proprietary programs.  If your program is a subroutine library, you may
consider it more useful to permit linking proprietary applications with the
library.  If this is what you want to do, use the GNU Library General
Public License instead of this License.
"""


#----------------------------------------------------------------------------
#                              	GLOBAL SETTINGS
#----------------------------------------------------------------------------


""" Create some Game_State instances to handle multiple-move backups (up
	to 10 in a row, arbitrarily) during testing, plus a temporary state
	for backup purposes during some routines. """

cur_state = Game_State()
last_state = Game_State()
last2_state = Game_State()
last3_state = Game_State()
last4_state = Game_State()
last5_state = Game_State()
last6_state = Game_State()
last7_state = Game_State()
last8_state = Game_State()
last9_state = Game_State()
last10_state = Game_State()
temp_state = Game_State()

# Initialize some assorted main global variables.
next_jump = 'start'
cur_parse_file = 'start'
expression = ''
just_jumped = 0
reg_names = {}
enum_names = {}
reg_namekeys = []
enum_namekeys = []
com_count = {}
left_shown = 1
bot_shown = 1
loginv_shown = 0
left_sash = 125
bot_sash = 500
loginv_sash = 500
act_list = []
area_list = []
block_text_lines = []
moved_back = 0
wordcount_dictionary = {}
parse_context = 0
hide = 0

""" For the inventory display.  min_col_width is the minimum number of pixels
	in width a column of items can be forced to be before the inventory is
	displayed with one fewer column (down to one column minimum, naturally).
	inv_change is set during parsing when the inventory changes, and at the
	end of parsing, if it's set, UpdateInvWindow is called. """
min_col_width = 120
inv_change = 0

# System values which don't get stored in a saved state.
S1 = 0
S2 = 0
S3 = 0
S4 = 0
S5 = 0

# Options
playername = 'Player'

# Logging options (can toggle whether it logs to <log_filename> or the bottom panel, separately).
panellog = 1
textlog = 1
log_filename = os.path.join(main_directory, 'logfile.txt')
log_lines = 0

# Hall of Fame recording option:
hof_track = 1
hall_filename = os.path.join(main_directory, 'hall_of_fame.txt')

# Only for the case of renaming an existing Hall of Fame automatically to preserve incompatible data:
new_hallfile = ''

# Narration option (logs move number, text, and resulting action description text to <narrate_filename>).
narrate = 0
nar_area = 0
narrate_filename = os.path.join(main_directory, 'narrate.txt')

# Word statistics compilation option (can toggle whether it compiles results to external file <ws_filename>).
wordstats = 1
ws_filename = os.path.join(main_directory, 'wordstats.txt')

""" If develop=1, you get a button to edit the game state, you can
	see the current location in the bottom left panel, you can
	see the status line numbers, you can see the tags of items
	when trying to examine them, and a list of all masked items,
	variables, and timers is shown in the top left panel. """
develop = 1

""" Dictionaries for looking up a global variable,
	state variable, or object or timer tag, upon a change,
	to see whether it's masked and shouldn't have a change logged. """
gvarmasked = {}
svarmasked = {}
robjmasked = {}
eobjmasked = {}
timermasked = {}

""" Dictionary of characters composing words (for word counting).
	Value "2" characters can both start and continue words.
	Value "1" characters can continue words, but can't start them.
	E.g. in parsing "a 'b c- -d-", the words recorded are "a", "b", "c-", and "d-".
	The inclusion of some characters and the exclusion of others is up to one's
	judgement.

	Naturally a few words, such as the second word
	of "I'm 'fraid not," will not get their apostrophes included.  This way, though,
	doesn't count the lead quotes of singly quoted phrases as part of the first word. """

word_pieces = {}
for c in string.digits + "'":
	word_pieces[c] = 1
for i in [35, 36, 37, 64, 137, 178, 179, 185, 188, 189, 190]:
	word_pieces[chr(i)] = 1
for c in string.letters:
	word_pieces[c] = 2
for i in [138, 140, 142, 154, 156, 158, 159, 162, 163, 165]:
	word_pieces[chr(i)] = 2
for i in range(192, 256):
	word_pieces[chr(i)] = 2

# Dictionary of operators (op) which can be used with the syntax "v var (op) (expression)"
opdict = {}
for op in ["+", "-", "*", "/", "-/", "+/", "%", "-%", "+%", "**", "=", "!=", "!", "<", ">", "<=", ">=", ",", "|", ".&", ".|", ".^", "~", "<<", ">>", "++"]:
	opdict[op] = 1

# Dictionary for checking if a string consists solely of legal "tag" characters...
TagChars = {}
for c in string.letters + string.digits + "_":
	TagChars[c] = 1

# ...and some related regexes:
tag_regex = re.compile("^[a-zA-Z0-9_]+$")
tag_regex2 = re.compile("^[a-zA-Z0-9_-]+$")
tag_regex3 = re.compile("^[a-z0-9_]+$")

""" Stats_compute is set to 1 when the routine for computing
	one (or all) location's statistics is running and set back to
	0 after.  When stats computation is on, some things differ:
	e.g. variable changes are never logged, and some errors such
	as "g loc{var}" having "loc{var}" evaluated as a non-location
	are ignored (since during *run-time*, it may never have an error flagged). """
stats_compute = 0

# Data according to game.txt:
game_title = ''
credits = ""
difficulty = -1		# unless it's explicitly set to 0 or above, it won't be displayed
synopsis_short = ""
synopsis_long = ""

""" Note: these are the fields according to game.txt.  They might not be
	compatible with those in the actual Hall of Fame file.  If so, the discrepancy
	will be resolved when a w/W command comes about and data needs to be posted
	to the Hall of Fame. """
hof_fields = []


# Constants for the Mersenne Twister algorithm used in pseudo-random number generation.
N = 624
mt = range(N+1)
mti = N + 1
	
# And finally, the starting GUI.
app = MyApp(wx.Platform == "__WXMAC__")
app.MainLoop()
